// FDGeneral.cpp : implementation file
//

#include "stdafx.h"
#include "FDGeneral.h"
#include "file_ver.h"
#include "..\otherfunctions.h"
#include "..\emule.h"
#include "..\CommentDialogLst.h"

#pragma comment(lib, "version.lib") 

// CFDGeneral dialog

IMPLEMENT_DYNCREATE(CFDGeneral, CPropertyPage)

CFDGeneral::CFDGeneral() : CPropertyPage(CFDGeneral::IDD)
{
	//{{AFX_DATA_INIT(CFDGeneral)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT

	m_pFile = NULL;
	m_hFileIcon = NULL;
}

CFDGeneral::~CFDGeneral()
{
	if(m_hFileIcon != NULL)
	{
		::DestroyIcon(m_hFileIcon);
	}
}

void CFDGeneral::DoDataExchange(CDataExchange* pDX)
{
	CPropertyPage::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CFDGeneral)
	DDX_Control(pDX, IDC_PARTFILESTATUS_VAL, m_ctrlPartfileStatus);
	DDX_Control(pDX, IDC_METFILE_VAL, m_ctrlMetFile);
	DDX_Control(pDX, IDC_HASH_VAL, m_ctrlHash);
	DDX_Control(pDX, IDC_FOLDER_VAL, m_ctrlFolder);
	DDX_Control(pDX, IDC_FILETYPE_VAL, m_ctrlFiletype);
	DDX_Control(pDX, IDC_FILESIZE_VAL, m_ctrlFilesize);
	DDX_Control(pDX, IDC_FILEICON, m_ctrlFileIcon);
	DDX_Control(pDX, IDC_DESCRIPTION_VAL, m_ctrlDescription);
	//}}AFX_DATA_MAP
	DDX_Control(pDX, IDC_FDG_FILENAME, m_ctrlFilename);
	DDX_Control(pDX, IDC_CATEGORY_VAL, m_ctrlCategory);
}


BEGIN_MESSAGE_MAP(CFDGeneral, CPropertyPage)
	//{{AFX_MSG_MAP(CFDGeneral)
	//}}AFX_MSG_MAP
	ON_BN_CLICKED(IDC_COMMENTS_BTN, OnBnClickedCommentsBtn)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CFDGeneral message handlers

BOOL IsExecuteable(CString strFile)
{
	int iDot = strFile.ReverseFind('.');
	if(iDot == -1)
	{		
		return FALSE;		// no file extension found
	}
	
	CString strExt = strFile.Right(strFile.GetLength() - iDot);
	if(strExt.IsEmpty())
	{
		return FALSE;		// no file extension found
	}

	strExt.MakeUpper();
	if(strExt == ".EXE" || strExt == ".DLL")
	{
		return TRUE;		// executeable
	}

	return FALSE;			// not executeable
}

void CFDGeneral::Update()
{
	EMULE_TRY

	if(m_pFile == NULL)
		return;

	if(!::IsWindow(GetSafeHwnd()))
		return;

	CString strFilename = m_pFile->GetFileName();

	SHFILEINFO shfi;
	memset2(&shfi,0,sizeof(shfi));
	SHGetFileInfo(strFilename, FILE_ATTRIBUTE_NORMAL, &shfi, sizeof(shfi),
					SHGFI_ICON|SHGFI_USEFILEATTRIBUTES|SHGFI_TYPENAME|SHGFI_DISPLAYNAME);
	
	if(m_hFileIcon != NULL)
	{
		::DestroyIcon(m_hFileIcon);
	}
	m_hFileIcon = shfi.hIcon;
	m_ctrlFileIcon.SetIcon(m_hFileIcon);

	m_ctrlFilename.SetWindowText(MakeStringEscaped(strFilename));
	m_ctrlFiletype.SetWindowText(shfi.szTypeName);

	if(IsExecuteable(strFilename))
	{
		GetDlgItem(IDC_DESCRIPTION_LBL)->SetWindowText(::GetResString(IDS_FD_DESCRIPTION));
		
		CFileVersionInfo fvInfo;
		fvInfo.ReadVersionInfo(strFilename);
		if(fvInfo.IsValid() )
		{
			if(fvInfo.IsVersionInfoAvailable(SFI_FILEDESCRIPTION))
			{
				m_ctrlDescription.SetWindowText(fvInfo.GetVersionInfo(SFI_FILEDESCRIPTION));
			}
			else
			{
				m_ctrlDescription.SetWindowText(shfi.szDisplayName);
			}
		}
		else
		{
			m_ctrlDescription.SetWindowText(shfi.szDisplayName);
		}
	}
	else
	{
		GetDlgItem(IDC_DESCRIPTION_LBL)->SetWindowText(::GetResString(IDS_FD_OPENSWITH));
		// ToDo: better solution needed	
		CString strTempFile = g_eMuleApp.m_pGlobPrefs->GetTempDir() + "\\" + strFilename;
		CFile cfTemp(strTempFile, CFile::modeCreate);
		cfTemp.Close();
		char szExecuteable[MAX_PATH];
		FindExecutable(strTempFile, "", szExecuteable);
		CFile::Remove(strTempFile);

		CFileVersionInfo fvInfo;
		fvInfo.ReadVersionInfo(szExecuteable);
		if(fvInfo.IsValid() )
		{
			if(fvInfo.IsVersionInfoAvailable(SFI_FILEDESCRIPTION))
			{				
				m_ctrlDescription.SetWindowText(fvInfo.GetVersionInfo(SFI_FILEDESCRIPTION));
			}
			else
			{
				m_ctrlDescription.SetWindowText(szExecuteable);
			}
		}
		else
		{
			m_ctrlDescription.SetWindowText(szExecuteable);
		}
	}

	CCat		*pCat = CCat::GetCatByID(m_pFile->GetCatID());
	CString		strCategory = (pCat != NULL) ? pCat->GetTitle() : GetResString(IDS_CAT_UNCATEGORIZED);
	m_ctrlCategory.SetWindowText(strCategory);

	CString strOldDir;
	CString strNewDir = m_pFile->GetOutputDir() + CString("\\");
	m_ctrlFolder.GetWindowText(strOldDir);
	if(strOldDir != strNewDir)
		m_ctrlFolder.SetWindowText(strNewDir);
	m_ctrlFilesize.SetWindowText(CastItoXBytes(m_pFile->GetFileSize()));
	m_ctrlMetFile.SetWindowText(m_pFile->GetFullName());
	m_ctrlHash.SetWindowText(EncodeBase16(m_pFile->GetFileHash(), 16));
	m_ctrlPartfileStatus.SetWindowText(m_pFile->GetPartfileStatus());

	EMULE_CATCH
}

void CFDGeneral::OnBnClickedCommentsBtn()
{
	CCommentDialogLst dialog(m_pFile); 
	dialog.DoModal();
}

BOOL CFDGeneral::OnInitDialog()
{
	CPropertyPage::OnInitDialog();

	Localize();
	Update();
	
	return TRUE;  // return TRUE unless you set the focus to a control
	// EXCEPTION: OCX Property Pages should return FALSE
}

void CFDGeneral::Localize(void)
{
	EMULE_TRY

	if(GetSafeHwnd())
	{
		CString strBuffer;

		GetDlgItem(IDC_FILETYPE_LBL)->SetWindowText(GetResString(IDS_FILETYPE_LBL));
		strBuffer = m_pFile->GetFileName();
		if(IsExecuteable(strBuffer))
		{	
			GetDlgItem(IDC_DESCRIPTION_LBL)->SetWindowText(GetResString(IDS_FD_DESCRIPTION));
		}
		else
		{	
			GetDlgItem(IDC_DESCRIPTION_LBL)->SetWindowText(::GetResString(IDS_FD_OPENSWITH));
		}
		strBuffer.Format("%s:", GetResString(IDS_CAT));
		GetDlgItem(IDC_CATEGORY_LBL)->SetWindowText(strBuffer);
		GetDlgItem(IDC_FOLDER_LBL)->SetWindowText(GetResString(IDS_FD_OUTPUT));
		GetDlgItem(IDC_FILESIZE_LBL)->SetWindowText(GetResString(IDS_FD_SIZE));
		GetDlgItem(IDC_METFILE_LBL)->SetWindowText(GetResString(IDS_FD_MET));
		GetDlgItem(IDC_HASH_LBL)->SetWindowText(GetResString(IDS_FD_HASH));
		strBuffer.Format("%s:", GetResString(IDS_STATUS));
		GetDlgItem(IDC_PARTFILESTATUS_LBL)->SetWindowText(strBuffer);
		GetDlgItem(IDC_COMMENTS_BTN)->SetWindowText(GetResString(IDS_CMT_SHOWALL));
	}

	EMULE_CATCH
}