// Tasks.cpp: implementation of the CTasks class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#include "TasksSockets.h"
#include "../EmEngine.h"
#include "TaskProcessorSockets.h"
#include "OpCode.h"
#include "../Data/Server.h"

//////////////////////////////////////////////////////////////////////
// global and sockets task functions
CTask::~CTask()
{
}

bool CTask::OnException(CEmEngine&)
{
	// delete this buggy task
	return true;
}

bool CTask_Tcp_Err::Post(CEmEngine& stEngine, SOCKET hSocket, int nErrorCode)
{
	CTask_Tcp_Err* pTask = new CTask_Tcp_Err;
	if (pTask)
	{
		pTask->m_hSocket = hSocket;
		pTask->m_nError = nErrorCode;
		stEngine.PushToSocketsTP(pTask);
	} 
	else
	{
		ASSERT(FALSE);
		AddDebugLogLine(false, "no memory");
	}

	return NULL != pTask;
}

bool CTask_Tcp_Err::Process(CEmEngine& stEngine)
{
	// check if we still have this client
	CEmClient* pClient = stEngine.m_stSocketsTP.Lookup(m_hSocket);
	if (pClient)
	{
		ASSERT(pClient->m_hSocket == m_hSocket);
		// here we can write to the log that this client has disconnected from us

		stEngine.m_stSocketsTP.KillClient(pClient);
	}

	return true;
}

bool CTask_Tcp_Accepted::Process(CEmEngine& stEngine)
{
	CEmClient* pClient = stEngine.m_stSocketsTP.AllocClient(m_eType);
	ASSERT(pClient);
	if (pClient)
	{
		pClient->m_bFromOutside = true;
		pClient->m_hSocket = m_hSocket;
		pClient->m_bIsConnected = true;
		stEngine.m_stSocketsTP.m_mapClients[m_hSocket] = pClient;

		pClient->OnAccepted(stEngine);

	} else
	{
		ASSERT(FALSE);
		AddDebugLogLine(false, "no memory");
		if (closesocket(m_hSocket))
			AddDebugLogLine(false, "closesocket failed");
	}

	return true;
}

bool CTask_Tcp_Connected::Process(CEmEngine& stEngine)
{
	CEmClient* pClient = stEngine.m_stSocketsTP.Lookup(m_hSocket);
	if (pClient)
	{
		ASSERT(!pClient->m_bIsConnected);
		pClient->m_bIsConnected = true;

		pClient->OnConnected(stEngine);
	}
	return true;
}

bool CTask_Tcp_Web::Process(CEmEngine& stEngine)
{
	CEmClient_Web* pClient = (CEmClient_Web*) stEngine.m_stSocketsTP.Lookup(m_hSocket);
	if (pClient)
	{
		ASSERT(T_CLIENT_WEB == pClient->GetType());

		  CString strTxt;
		  strTxt.Format("Currently I have %d clients", stEngine.m_stSocketsTP.m_mapClients.size());
		  stEngine.m_stTcp.AllocSend(m_hSocket, (LPCTSTR) strTxt, strTxt.GetLength());
		  stEngine.m_stTcp.AllocDisconnect(m_hSocket);
	}

	return true;
}
// end of global and socket task functions
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////
// connect
CTask_Connect::CTask_Connect(ULONG nAddr, USHORT nPort, T_CLIENT_TYPE eType)
	: m_nAddr(nAddr), m_uPort(nPort), m_eType(eType)
{
}

//////////////////////////////////////////////////////////////////////
CTask_Connect::CTask_Connect(LPCTSTR sAddr, USHORT nPort, T_CLIENT_TYPE eType) 
	: m_nAddr(inet_addr(sAddr)), m_uPort(nPort), m_eType(eType)
{
}

//////////////////////////////////////////////////////////////////////
bool CTask_Connect::Process(CEmEngine& stEngine)
{
	if(m_nAddr && m_uPort)
		stEngine.m_stSocketsTP.AllocTcpConnect(m_nAddr, m_uPort, m_eType);
	else
	{
		// Connect to any server
	}

	return true;
}

//////////////////////////////////////////////////////////////////////
// disconnect
CTask_KillClient::CTask_KillClient(CEmClient* pClient)
	: m_pClient(pClient)
{
}

//////////////////////////////////////////////////////////////////////
bool CTask_KillClient::Process(CEmEngine& stEngine)
{
	stEngine.KillClient(m_pClient);

	return true;
}

//////////////////////////////////////////////////////////////////////
// send list of shared files
CTask_SendSharedList::CTask_SendSharedList()
{
	m_bProcessedUI = false;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendSharedList::Process(CEmEngine& stEngine)
{
	if(!stEngine.IsConnected())
		return true;

	// first collect shared files info
/*	if(!m_bProcessedUI)
	{
		stEngine.PushToUI(this);
		return false;
	}
*/
	// afterwards, send it
/*	COpCode_OFFERFILES stMsg;
	for(int i = 0; i < m_Files.GetCount(); i++)
	{
		COpCode_OFFERFILES::CStruct_Files file;
		md4cpy(file._Hash, m_Files[i]._Hash);
		file._Client._Addr = 0;
		file._Client._Port = 0;
		file._FileName.m_strValue = m_Files[i]._FileName;
		file._FileName.m_bValid = true;
		file._FileSize.m_dwValue = m_Files[i]._FileSize;
		file._FileSize.m_bValid = true;
		stMsg._Files.Add(file);
	}
	stEngine.m_stTcp.AllocSend(stEngine.m_stData.GetConnectedServer(), stMsg);
*/
	return true;
}

//////////////////////////////////////////////////////////////////////
/*bool CTask_SendSharedList::ProcessForUI(CEmEngine& stEngine)
{
	CKnownFile	*pKnownFile;
	CCKey		bufKey;
	for (POSITION pos = g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetStartPosition(); pos != 0; )
	{
		g_eMuleApp.m_pSharedFilesList->m_mapSharedFiles.GetNextAssoc(pos, bufKey, pKnownFile);
		if(pKnownFile)
		{
			OFFEREDFILE File;

			md4cpy(File._Hash, pKnownFile->GetFileHash());
			File._Addr = 0;
			File._Port = 0;
			File._FileName = pKnownFile->GetFileName();
			File._FileSize = pKnownFile->GetFileSize();

			m_Files.Add(File);

			pKnownFile->SetOfferedToServer(true);
		}
	}

	m_bProcessedUI = true;

	stEngine.m_stSocketsTP.Push(this);

	return false;
}
*/