//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#pragma once
#include "otherstructs.h"
#include "packets.h"
#include "types.h"
#include <afxcoll.h>
#include <afxtempl.h>
#include <afxcmn.h>
#include "securedvars.h"
#include "PastComment.h"

//--- xrmb:partprio ---
#include "opcodes.h"
//--- :xrmb ---
#include "Loggable.h"

#if JUMPSTART
//#include "Jumpstarter.h"
class CJumpstarter;
#endif

class CFileStatistic : public CLoggable{
	friend class CKnownFile;
	//--- xrmb:parttraffic ---
	//--- i need more friends, for member-access/speed reasons
	friend class CSharedFilesCtrl;
	friend class CKnownFileList;
	//--- :xrmb ---

public:
	//--- xrmb:parttraffic ---
	//--- the constructor is now in the .cpp file ---
	//CFileStatistic()					{requested = transferred = accepted = alltimerequested= alltimetransferred = alltimeaccepted = 0;}
	CFileStatistic();
	~CFileStatistic();

	double	GetCompleteReleases() { return completeReleases; }
	double	GetCompletePartReleases(uint16 part);
	void	AddTraffic(uint32 start, uint32 bytes); 

	void	resetStats(bool all=false);
	void	resetPartTraffic(uint16	part, bool all=false);
	//--- :xrmb ---

	void	AddRequest();
	void	AddAccepted();
	void	AddTransferred(uint32 bytes);
	uint16	GetRequests()				{return m_iNumRequested;}
	uint16	GetAccepts()				{return m_iNumAccepted;}
	uint64	GetTransferred()				{return m_qwNumTransferred;}
	uint16	GetAllTimeRequests()			{return alltimerequested;}
	uint16	GetAllTimeAccepts()			{return alltimeaccepted;}
	uint64	GetAllTimeTransferred()			{return alltimetransferred;}
	//--- xrmb:keepPartFileStats ---
	void	AddAll(uint32 _alltimerequested, uint32 _alltimeaccepted, uint64 _alltimetransferred);
	//--- :xrmb ---

	//--- xrmb:allyaknow ---
	bool	merge(CFileStatistic *m);
	//--- :xrmb ---

	//--- xrmb:parttraffic ---
	//--- i need them const, because i use const references, anyway, they const will not hurt anyone ---
	//--- and i let GetAllTimeRequests()/GetAllTimeAccepts() return uint32 ---
	uint16	GetRequests() const			{return m_iNumRequested;}
	uint16	GetAccepts() const			{return m_iNumAccepted;}
	uint64	GetTransferred()	const			{return m_qwNumTransferred;}
	uint32	GetAllTimeRequests() const		{return alltimerequested;}
	uint32	GetAllTimeAccepts() const		{return alltimeaccepted;}
	uint64	GetAllTimeTransferred() const 		{return alltimetransferred;}

	uint32	GetPartAccepted(uint16 part, bool session) const;
	//--- :xrmb ---
	CKnownFile* fileParent;

private:
	uint16 m_iNumRequested;
	uint64 m_qwNumTransferred;
	uint16 m_iNumAccepted;
	uint32 alltimerequested;
	uint64 alltimetransferred;
	uint32 alltimeaccepted;

	//--- xrmb:parttraffic ---
	uint32	*partAccepted;				// all time accepted requests on part level
	uint32	*partAcceptedSession;		// session accepted requests on part level

	uint32	*partTraffic;				// all time traffic on part level
	uint32	*partTrafficSession;		// session traffic on part level
	
	uint32	*blockTraffic;				// all time traffic on block level
	uint32	*blockTrafficSession;			// session traffic on block level
	
	double  completeReleases;			// how often the complete file was uploaded (basing on all time block-traffic)

	void	initTraffic();				// create the traffic arrays
	void	recalcCompleteReleases();		// recalculate the complete release count, don't do this to often
	//--- :xrmb ---
};

/*
					   CPartFile
					 /
		  CKnownFile
		/
CAbstractFile 
		\ 
		  CSearchFile
*/
class CAbstractFile : public CLoggable
{
public:
	CAbstractFile()
	{	
		m_lFileSize = 0;
		m_strFileName = "";
	}
	virtual ~CAbstractFile()
	{		
	}
	CString	GetFileName() const		{return m_strFileName;}
	void		SetFileName(const CString& NewName);
	const uchar*	GetFileHash() const			{return m_fileHash;}
	uint32	GetFileSize() const		{return m_lFileSize;}
	CString	GetFileType(bool bLocalized = false);
	virtual CString CreateED2kLink() const;
	virtual CString CreateED2kSourceLink() const;
	virtual CString CreateHTMLED2kLink() const;
protected:
	CString		m_strFileName;
	uchar		m_fileHash[16];
	uint32		m_lFileSize;
	CString m_strComment;
	EnumPartFileRating	m_eRating;
};

class CUpDownClient;

class CKnownFile : public CAbstractFile
{
public:
	CKnownFile();
	~CKnownFile();
	virtual bool	CreateFromFile(const CString& directory,const CString& filename); // create date, hashset and tags from a file
	CString	GetPath() const						{return directory;}
	void	SetPath(const CString& path) { directory = path; }
	const CString& GetFilePath() const { return m_strFilePath; }
	void	SetFilePath(LPCTSTR pszFilePath);

	virtual	bool	IsPartFile()			{return false;}
	virtual bool	LoadFromFile(CFile& file);	//load date, hashset and tags from a .met file
	bool	WriteToFile(FILE* file);
	uint32	GetFileDate()					{return m_timetLastWriteDate;}
	double	GetFileRatio(void) ; 			// Added by Tarod
	uint16	GetHashCount()					{ CSingleLock Lock(&m_mutexHashList,TRUE); return m_partHashArray.GetCount();}
	uchar*	GetPartHash(uint16 part);
	uint16	GetPartCount();
	uint32	m_timetLastWriteDate;
	// file upload priority
	uint8	m_iPriority;
	uint8	GetPriority(void)				{return m_iPriority;};
	void	SetPriority(uint8 iNewPriority) {m_iPriority = iNewPriority;};

	// Added by Tarod for AutoPriority
	void    UpdateUploadAutoPriority(void);	// Added by Tarod for UAP
	bool	IsAutoPrioritized() 					{return m_bAutoPriority;}
	void	SetAutoPriority(bool newAutoPriority) 	{m_bAutoPriority=newAutoPriority;}

	// DonGato - Moview Preview Mode
	uint8	m_iMoviePreviewMode;
	uint8	GetMovieMode()						{return m_iMoviePreviewMode;}
	void	SetMovieMode(uint8 iNewMode) 		{m_iMoviePreviewMode = iNewMode;}

	// shared file view permissions (all, only friends, no one)
	uint8	m_iPermissions;
	uint8	GetPermissions(void)			{return m_iPermissions;};
	void	SetPermissions(uint8 iNewPermissions) {m_iPermissions = iNewPermissions;};

	bool	LoadHashsetFromFile(CFile& file, bool checkhash);

	CFileStatistic statistic;
	//comment 
	CString	GetFileComment()         {if (!m_bCommentLoaded) LoadComment(); return m_strComment;} 
	void	SetFileComment(CString strNewComment);
	void	SetFileRating(EnumPartFileRating eNewRating); 
	EnumPartFileRating	GetFileRating()         	{if (!m_bCommentLoaded) LoadComment(); return m_eRating;}
	void	RemoveFileCommentAndRating();

	uint16	GetCompleteSourcesCount()		{return m_nCompleteSourcesCount;}		//netwolf: complete sources (zegzav) 13.05.03
	void	GetCompleteSourcesRange(uint16 *lo, uint16 *hi)	{*lo= m_nCompleteSourcesCountLo; *hi= m_nCompleteSourcesCountHi;}
	//--- xrmb:parttraffic ---
	uint32	GetTrafficBlock(uint32 start, uint32 end, bool session=false);
	uint32	GetTrafficPart(uint32 start, uint32 end, bool session=false);
	uint32	GetPartTraffic(uint16 part, bool session=false);
	uint32	GetBlockTraffic(uint16 block, bool session=false);
	bool	LoadFromFileTraffic(FILE* file, uint8 version);
	bool	SaveToFileTraffic(FILE* file);	

	uint32	GetPartSize(uint16 part);
	uint32	GetBlockSize(uint32 block);

	uint32	GetBlockCount();
	//--- :xrmb ---

	//--- xrmb:allyaknow ---
	void	SetSharedFile(bool b) { m_sharedFile=b; }
	bool	GetSharedFile() { return m_sharedFile; }
	//--- :xrmb ---

	//--- xrmb:partprio ---
	void	SetPartStatus(uint16 part, uint8 s)	{ m_partStatus[part]=s; }
	uint8	GetPartStatus(uint16 part);
	bool	HasHiddenParts();
    bool    HasParts(); // DbT:StartUploadReq
	void	WritePartStatus(CFile* file);
	//void	SetPartStatus(uint16 part, bool s)	{ m_partStatus[part]=s; }
	//bool	GetPartStatus(uint16 part)			{ bool r=true; m_partStatus.Lookup(part, r); return r; }
	//--- :xrmb ---

	// file sharing
	virtual	Packet*	CreateSrcInfoPacket(CUpDownClient* forClient);
		void	CalculateCompleteSources();	//netwolf: complete sources (zegzav) 13.05.03

#if JUMPSTART
    bool GetJumpstartEnabled(void) {return m_Jumpstarter!=NULL;};
    void SetJumpstartEnabled(bool enabled);
    bool IsJsComplete();
    void AddSentBlock(CUpDownClient* client, uint32 start_offset, uint32 togo);
    void WriteJumpstartPartStatus(CUpDownClient* client, CMemFile* data);
    bool AllowChunkForClient(uint32 partNo, CUpDownClient* client);
#endif

	void SetOfferedToServer(bool bOffered)	{m_bOfferedToServer= bOffered;}
	bool GetOfferedToServer()				{return m_bOfferedToServer;}

protected:
	CMutex	m_mutexHashList;

	bool	LoadTagsFromFile(CFile& file);
	bool	LoadDateFromFile(CFile& file);
	void	CreateHashFromFile(FILE* file, int Length, uchar* Output)	{CreateHashFromInput(file,0,Length,Output,0);}
	void	CreateHashFromFile(CFile* file, int Length, uchar* Output)	{CreateHashFromInput(0,file,Length,Output,0);}
	void	CreateHashFromString(uchar* in_string, int Length, uchar* Output)	{CreateHashFromInput(0,0,Length,Output,in_string);}
	void	LoadComment();//comment

	CArray<uchar*,uchar*>		m_partHashArray;
	CArray<CFileTag*,CFileTag*>	m_tagArray;
	CString						directory;
	CString						m_strFilePath;

	//--- xrmb:allyaknow ---
	bool	m_sharedFile;
	//--- :xrmb ---

	//--- xrmb:partprio ---
	CMap<uint16, uint16, uint8, uint8> m_partStatus;
	//CMap<uint16, uint16, bool, bool> m_partStatus;
	//--- :xrmb ---

private:
	void	CreateHashFromInput(FILE* file,CFile* file2, int Length, uchar* Output, uchar* = 0);
	bool	m_bAutoPriority ; // Tarod
	bool	m_bCommentLoaded;
	uint16	m_nCompleteSourcesCount;	// netwolf: complete sources (zegzav) 13.05.03
	uint16	m_nCompleteSourcesCountLo;
	uint16	m_nCompleteSourcesCountHi;
	uint32	m_nCompleteSourcesTime;	
#if JUMPSTART
    CJumpstarter* m_Jumpstarter;
#endif
	bool	m_bOfferedToServer;

public:
//	UI updates
	void	UpdateSharedFileDisplay();
};

// permission values for shared files
#define PERM_ALL		0
#define PERM_FRIENDS	1
#define PERM_NOONE		2

// constants for MD4Transform
#define S11 3
#define S12 7
#define S13 11
#define S14 19
#define S21 3
#define S22 5
#define S23 9
#define S24 13
#define S31 3
#define S32 9
#define S33 11
#define S34 15

// basic MD4 functions
#define MD4_F(x, y, z) ((((y) ^ (z)) & (x)) ^ (z))
#define MD4_G(x, y, z) (((z) & ((x) ^ (y))) | ((x) & (y)))
#define MD4_H(x, y, z) ((x) ^ (y) ^ (z))

// rotates x left n bits
// 15-April-2003 Sony: use MSVC intrinsic to save a few cycles
#ifdef _MSC_VER
#pragma intrinsic(_rotl)
#define MD4_ROTATE_LEFT(x, n) _rotl((x), (n))
#else
#define MD4_ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))
#endif

// partial transformations
#define MD4_FF(a, b, c, d, x, s) \
{ \
  (a) += MD4_F((b), (c), (d)) + (x); \
  (a) = MD4_ROTATE_LEFT((a), (s)); \
}

#define MD4_GG(a, b, c, d, x, s) \
{ \
  (a) += MD4_G((b), (c), (d)) + (x) + (uint32)0x5A827999; \
  (a) = MD4_ROTATE_LEFT((a), (s)); \
}

#define MD4_HH(a, b, c, d, x, s) \
{ \
  (a) += MD4_H((b), (c), (d)) + (x) + (uint32)0x6ED9EBA1; \
  (a) = MD4_ROTATE_LEFT((a), (s)); \
}

static void MD4Transform(uint32 Hash[4], uint32 x[16]);

