//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "partfile.h"
#include "emule.h"
#include "updownclient.h"
#include <math.h>
#include "ED2KLink.h"
#include "Preview.h"
#include "ini2.h"
#include "ArchiveRecovery.h"
#include <sys/stat.h>
#include <io.h>
#include "Smtp.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif


//	Barry - use this constant for both places
#define PROGRESS_HEIGHT 4	//netwolf 06.05.03 (was 3)
/*	Maella - removed mem leak
CBarShader CPartFile::s_LoadBar(PROGRESS_HEIGHT); // Barry - was 5
CBarShader CPartFile::s_ChunkBar(16);
*/
CMutex CPartFile::s_CompletingMutex;

CPartFile::CPartFile()
{
	Init();
}

CPartFile::CPartFile(CSearchFile* searchresult)
{
	Init();

	EMULE_TRY

	md4cpy(m_fileHash, searchresult->GetFileHash());
	for (int i = 0; i < searchresult->m_tagArray.GetCount();i++)
	{
		switch (searchresult->m_tagArray[i]->GetSpecialTag())
		{
			case FT_FILENAME:
			{
				SetFileName(searchresult->m_tagArray[i]->GetStringValue());
				break;
			}
			case FT_FILESIZE:
			{
				m_lFileSize = searchresult->m_tagArray[i]->GetIntValue();
				break;
			}
			default:
			CFileTag* newtag = new CFileTag(*searchresult->m_tagArray[i]);
			m_tagArray.Add(newtag);
		}
	}
//	Don't ask for a hashset, if there is none.
	if (m_lFileSize < PARTSIZE)
		m_bHashSetNeeded = false;
	CreatePartFile();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CPartFile::CPartFile(CString edonkeylink)
{
	Init();

	EMULE_TRY

	CED2KLink* pLink = 0;
	try
	{
		pLink = CED2KLink::CreateLinkFromUrl(edonkeylink);
		_ASSERT(pLink != 0);
		CED2KFileLink* pFileLink = pLink->GetFileLink();
		if (pFileLink == 0)
			throw GetResString(IDS_ERR_NOTAFILELINK);
		InitializeFromLink(pFileLink);
	}
	catch (CString error)
	{
		OUTPUT_DEBUG_TRACE();
		CString buffer;
		buffer.Format(GetResString(IDS_ERR_INVALIDLINK), error.GetBuffer());
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_LINKERROR), buffer);
		SetStatus(PS_ERROR);
	}
	delete pLink;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::InitializeFromLink(CED2KFileLink* fileLink)
{
	Init();

	EMULE_TRY

	try
	{
		//m_strFileName = CString(fileLink->GetName());
		SetFileName(CString(fileLink->GetName()));
		m_lFileSize = fileLink->GetSize();
	//	Ultras: prevent to donload an empty file
		if (!m_lFileSize)
		{
			AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_SKIPZEROLENGTHFILE), m_strFileName);
			SetStatus(PS_ERROR);
			return ;
		}
	//	Don't ask for a hashset, if there is none.
		if (m_lFileSize < PARTSIZE)
			m_bHashSetNeeded = false;
		memcpy2(m_fileHash, fileLink->GetHashKey(), sizeof(m_fileHash));
		if (!g_eMuleApp.m_pDownloadQueue->FileExists(m_fileHash))
			CreatePartFile();
		else
			SetStatus(PS_ERROR);
	}
	catch (CString error)
	{
		OUTPUT_DEBUG_TRACE();
		CString buffer;
		buffer.Format(GetResString(IDS_ERR_INVALIDLINK), error.GetBuffer());
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_LINKERROR), buffer);
		SetStatus(PS_ERROR);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CPartFile::CPartFile(CED2KFileLink* fileLink)
{
	Init();
	InitializeFromLink(fileLink);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::Init()
{
	EMULE_TRY

	m_strTempDir = "";
	newdate = true;
	m_dwLastFileSourcesRequestTime = 0;
	lastpurgetime = ::GetTickCount();
	m_bPaused = false;
	m_bStopped = false;
	SetStatus(PS_EMPTY);
	m_dwBytesTransferred = 0;
	priority = PR_NORMAL;
	m_iNumTransferringSrcs = 0;
	m_iSrcNNP = 0;
	m_iSrcOnQueue = 0;
	m_iSrcHighQR = 0;
	m_iSrcConnecting = 0;
	m_iSrcTooManyConns = 0;
	m_iSrcConnected = 0;
	m_iSrcTimeout = 0;
	m_iSrcConnViaServer = 0;
	m_iSrcUnknown = 0;
	m_iSrcRecHash = 0;
	m_iSrcLowToLow = 0;
	m_iSrcQueueFull = 0;
	m_iSrcNoUpload = 0;
	m_iSrcBanned = 0;
	m_iSrcError = 0;
	m_iSrcA4AF = 0;
	m_dataRate = 0;
	m_bHashSetNeeded = true;
	count = 0;
	m_dblPercentCompleted = 0.0;
	m_strPartMetFileName = "";
	completedsize = 0;

	m_bPreviewing = false;
	lastseencomplete = NULL;
	availablePartsCount = 0;
	m_ClientSrcAnswered = 0;
	m_LastNoNeededCheck = 0;
	m_eRating = PF_RATING_NONE;
	m_strComment = "";
	m_nTotalBufferData = 0;
	m_nLastBufferFlushTime = 0;
	m_bPercentUpdated = false;
	m_bRecoveringArchive = false;
	m_iGainDueToCompression = 0;
	m_iLostDueToCorruption = 0;
	m_iSessionGainDueToCompression = 0;
	m_iSessionLostDueToCorruption = 0;
	m_iTotalPacketsSavedDueToICH = 0;
	m_nSavedReduceDownload = 0;
	hasRating = false;
	hasComment = false;
	m_bDiscardSuperCompressed = false;

//	Tarod
	if (g_eMuleApp.m_pGlobPrefs->IsDAPEnabled())
		m_bAutoPriority = true;
	else
		m_bAutoPriority = false;

	m_LastRemovedTime = 0;
	m_nLastCompleteSrcCount = 0;
	m_bUpdateCompleteSrcCount = false;
	m_timeLastDownTransfer = CTime::GetCurrentTime();
	m_SessionStartTime = m_timeLastDownTransfer;
	m_SessionStartSize = 0;
	m_AvgDataRate = 0;
	m_StartTimeReset = false;

	m_eCategoryID = CAT_NONE;
	m_bIsBeingDeleted = false;
	m_bIsPreallocated = false;

//eklmn: it's need to initialize this variable to prevent extra screen refreshing
	m_dwLastRefreshedDLDisplayTime = ::GetTickCount();

//eklmn: create a Completion Event
//	m_hCompletionEvent = CreateEvent(NULL, TRUE, TRUE, _T("EPFileCompletionEvent"));
//	m_dwNumUploadingClients = 0;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CPartFile::~CPartFile()
{
	EMULE_TRY

//	Barry - Ensure all buffered data is written
	if (!m_bIsBeingDeleted)
		FlushBuffer();

	if (m_hPartFile.m_hFile != INVALID_HANDLE_VALUE && !m_bIsBeingDeleted)
	{
		m_hPartFile.Close();
	//	Update met file
		SavePartFile();
	}

	m_srcPartFrequencies.RemoveAll();
	m_ReqParts.RemoveAll();
	m_BlockedParts.RemoveAll();
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;)
		delete gaplist.GetNext(pos);

//	CloseHandle(m_hCompletionEvent);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::CreatePartFile()
{
	EMULE_TRY

	m_strTempDir = g_eMuleApp.m_pGlobPrefs->GetTempDir();	// added by InterCeptor (multiple tempdirs) 19.11.02
	if (m_strTempDir.Right(1) == '\\')
		m_strTempDir.Truncate(m_strTempDir.GetLength() - 1);	// added by InterCeptor (check for \\) 19.11.02
	directory = m_strTempDir;	// #zegzav:vdir4list (OP_ASKSHAREDFILESDIR)
	int i = 0;
	CString strFileName;
	do
	{
		i++;
		strFileName.Format(_T("%s\\%03i.part"), m_strTempDir, i);

	}
	while (PathFileExists(strFileName));

	m_strPartMetFileName.Format(_T("%03i.part.met"), i);
	m_strFullName.Format(_T("%s\\%s"), m_strTempDir, m_strPartMetFileName);
	CString buffer = m_strPartMetFileName.Left(m_strPartMetFileName.GetLength() - 4);
	CFileTag* partnametag = new CFileTag(FT_PARTFILENAME, buffer);
	m_tagArray.Add(partnametag);

	Gap_Struct* gap = new Gap_Struct;
	gap->m_dwStartOffset = 0;
	gap->m_dwEndOffset = m_lFileSize - 1;
	gaplist.AddTail(gap);

	CString partfull = m_strFullName.Left(m_strFullName.GetLength() - 4);
	SetFilePath(partfull);
	if (!m_hPartFile.Open(partfull, CFile::modeCreate | CFile::modeReadWrite | CFile::shareDenyWrite | CFile::osSequentialScan))
	{
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_CREATEPARTFILE));
		SetStatus(PS_ERROR);
	}

	const uint16 iPartCount = GetPartCount();
	m_srcPartFrequencies.SetSize(iPartCount);
	m_BlockedParts.SetSize(iPartCount);
	m_ReqParts.SetSize(iPartCount);
	for (uint32 i = 0; i < iPartCount;i++)
	{
		m_srcPartFrequencies[i] = 0;
		m_BlockedParts[i] = NULL;
		m_ReqParts[i] = 0;
	}
	m_bPaused = false;

//	Before we save file we need to stablish UAP by Tarod
	if (g_eMuleApp.m_pGlobPrefs->IsUAPEnabled())
	{
		CKnownFile::SetAutoPriority(true);
		CKnownFile::UpdateUploadAutoPriority();
		UpdateSharedFileDisplay();
	}
//	End Tarod

//	SLUGFILLER: hashlistRecheck
	if (m_lFileSize <= PARTSIZE)
		m_bHashSetNeeded = false;
//	SLUGFILLER: hashlistRecheck

//SyruS: cleanup
	if (g_eMuleApp.m_pGlobPrefs->GetAutoFilenameCleanup())
		SetFileName(CleanupFilename(GetFileName()).GetBuffer());

	SavePartFile();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::LoadPartFile(LPCTSTR in_directory, LPCTSTR in_filename)
{
	EMULE_TRY

	m_strTempDir = in_directory;	// added by InterCeptor (multiple tempdirs) 19.11.02
	CMap<uint16, uint16, Gap_Struct*, Gap_Struct*> gap_map; // safe Load Gaps [SlugFiller]
	m_dwBytesTransferred = 0;
	m_strPartMetFileName = in_filename;
	directory = in_directory;
	if (directory.Right(1) == '\\')
		directory.Truncate(directory.GetLength() - 1);	// added by InterCeptor (check for \\) 19.11.02

	m_strFullName.Format(_T("%s\\%s"), directory, m_strPartMetFileName);
	CSafeFile file;
	try
	{
	//	readfile data form part.met file
		if (!file.Open(m_strFullName, CFile::modeRead))
		{
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_OPENMET), m_strPartMetFileName, m_strFileName);
			return false;
		}
		uint8 version;
		file.Read(&version, 1);
		if (version != PARTFILE_VERSION)
		{
			if (file.m_hFile != INVALID_HANDLE_VALUE)
				file.Close();
			AddLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_BADMETVERSION), m_strPartMetFileName, m_strFileName);
			return false;
		}
		LoadDateFromFile(file);
		if (!LoadHashsetFromFile(file, false))
		{
			AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_HASHSETCORR), in_filename);
		}
		LoadSettingsFile();

		uint32 tagcount;
		file.Read(&tagcount, 4);
		for (uint32 j = 0; j != tagcount;j++)
		{
			CFileTag* newtag = new CFileTag(file);
			switch (newtag->GetSpecialTag())
			{
				case FT_FILENAME:
				{
					if (newtag->GetStringValue().IsEmpty())
					{
						AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_METCORRUPT), m_strPartMetFileName, m_strFileName);
						delete newtag;
						return false;
					}
					SetFileName(newtag->GetStringValue());
					delete newtag;
					break;
				}
				case FT_LASTSEENCOMPLETE:
				{
					lastseencomplete = newtag->GetIntValue();
					delete newtag;
					break;
				}
				case FT_FILESIZE:
				{
					m_lFileSize = newtag->GetIntValue();
					delete newtag;
					break;
				}
				case FT_TRANSFERRED:
				{
					m_dwBytesTransferred = newtag->GetIntValue();
					delete newtag;
					break;
				}
				case FT_PRIORITY:
				{
					priority = newtag->GetIntValue();
					delete newtag;

					if (priority == PR_AUTO)
						SetAutoPriority(true);
					else
						SetAutoPriority(false);
					break;
				}
				case FT_STATUS:
				{
					m_bPaused = newtag->GetIntValue();
					LoadSettingsFile();
					delete newtag;
					break;
				}
				case FT_ULPRIORITY:
				{
					if ((uint8)newtag->GetIntValue() == PR_AUTO)
					{
						CKnownFile::SetAutoPriority(true);
						CKnownFile::UpdateUploadAutoPriority();
					}
					else
					{
						CKnownFile::SetAutoPriority(false);
						CKnownFile::SetPriority((uint8)newtag->GetIntValue());
					}
					delete newtag;
					break;
				}
				case FT_CATEGORY:
				{
					m_eCategoryID = static_cast<_EnumCategories>(newtag->GetIntValue());
					delete newtag;
					break;
				}
				default:
				{
				//	Start Changes by Slugfiller for better exception handling
					if ( (!newtag->GetSpecialTag()) && ( newtag->GetTagName()[0] == FT_GAPSTART ||
					                                     newtag->GetTagName()[0] == FT_GAPEND ) )
					{
						Gap_Struct * gap;
						CString s = newtag->GetTagName();
						s = s.Right(s.GetLength() - 1);
						uint16 gapkey = _tstoi(s);
						if (!gap_map.Lookup(gapkey, gap))
						{
							gap = new Gap_Struct;
							gap_map.SetAt(gapkey, gap);
							gap->m_dwStartOffset = -1;
							gap->m_dwEndOffset = -1;
						}
						if (newtag->GetTagName()[0] == FT_GAPSTART)
							gap->m_dwStartOffset = newtag->GetIntValue();
						if (newtag->GetTagName()[0] == FT_GAPEND)
							gap->m_dwEndOffset = newtag->GetIntValue() - 1;
						delete newtag;
					//	End Changes by Slugfiller for better exception handling
					}
					else
						m_tagArray.Add(newtag);
				}

			//	START - enkeyDEV(Ottavio84) -ChangeDir-
				CString dfpath = CString(RemoveFileExtension(m_strFullName)) + CString(".dir");
				CStdioFile f;
				if (f.Open(dfpath, CFile::modeRead | CFile::typeText))
				{
					f.ReadString(m_strAlternativePath);
					f.Close();
				}
			//	END - enkeyDEV(Ottavio84) -ChangeDir-
			}

		}
		if (file.m_hFile != INVALID_HANDLE_VALUE)
			file.Close();
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
		if (error->m_cause == CFileException::endOfFile)
			AddLogLine(true, IDS_ERR_METCORRUPT, m_strPartMetFileName, m_strFileName);
		else
		{
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_FILEERROR), m_strPartMetFileName, m_strFileName, error);
		}
		error->Delete();
		for (POSITION pos = gap_map.GetStartPosition(); pos != NULL;)
		{
			Gap_Struct* gap;
			uint16 gapkey;
			gap_map.GetNextAssoc(pos, gapkey, gap);
			delete gap;
		}
		return false;
	}

//	Now to flush the map into the list (Slugfiller)
	for (POSITION pos = gap_map.GetStartPosition(); pos != NULL;)
	{
		Gap_Struct* gap;
		uint16 gapkey;
		gap_map.GetNextAssoc(pos, gapkey, gap);
	//SyruS (0.29a) SLUGFILLER: SafeHash - revised code, and extra safety
		if (gap->m_dwStartOffset != -1 && gap->m_dwEndOffset != -1 && gap->m_dwStartOffset <= gap->m_dwEndOffset && gap->m_dwStartOffset < m_lFileSize)
		{
			if (gap->m_dwEndOffset >= m_lFileSize)
				gap->m_dwEndOffset = m_lFileSize - 1; // Clipping
			AddGap(gap->m_dwStartOffset, gap->m_dwEndOffset); // All tags accounted for, use safe adding
		}
		delete gap;
	//	SLUGFILLER: SafeHash
	}

	EMULE_TRY
//--- xrmb:keepPartFileStats ---
	LoadPartFileStats();
//--- :xrmb ---
	EMULE_CATCH

//check if this is a backup
	if (m_strFullName.Right(7).CompareNoCase(_T(".backup")) == 0)
	{
		m_strFullName.Truncate(m_strFullName.GetLength() - 7);
	}

//	open permanent handle
	CString searchpath = m_strFullName.Left(m_strFullName.GetLength() - 4);
	if (!m_hPartFile.Open(searchpath, CFile::modeReadWrite | CFile::shareDenyWrite | CFile::osSequentialScan))
	{
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_FILEOPEN), m_strFullName, m_strFileName);
		return false;
	}
	SetFilePath(searchpath);

//SyruS (0.29a) SLUGFILLER: SafeHash - final safety, make sure any missing part of the file is gap
	if (m_hPartFile.GetLength() < m_lFileSize)
		AddGap(m_hPartFile.GetLength(), m_lFileSize - 1);
//	Goes both ways - Partfile should never be too large
	if (m_hPartFile.GetLength() > m_lFileSize)
	{
		TRACE("Partfile \"%s\" is too large! Truncating %I64u bytes.\n", GetFileName(), m_hPartFile.GetLength() - m_lFileSize);
		m_hPartFile.SetLength(m_lFileSize);
	}
//	SLUGFILLER: SafeHash

	const uint16 iPartCount = GetPartCount();
	m_srcPartFrequencies.SetSize(iPartCount);
	m_BlockedParts.SetSize(iPartCount);	//eklmn: ICR2.9+
	m_ReqParts.SetSize(iPartCount); //eklmn: ICR 2.10
	for (uint32 i = 0; i < iPartCount; i++)
	{
		m_srcPartFrequencies[i] = 0;
		m_BlockedParts[i] = NULL;
		m_ReqParts[i] = 0;		// ICR 2.10
	}
	SetStatus(PS_READY);
//	check hashcount, filesatus etc
	CSingleLock Lock(&m_mutexHashList, TRUE);
	INT_PTR iSize = m_partHashArray.GetCount();
	Lock.Unlock();
	if (iSize < GetPartCount() && GetFileSize() >= PARTSIZE)
	{
		SetStatus(PS_EMPTY);
		m_bHashSetNeeded = true;
		return true;
	}
	else
	{
		m_bHashSetNeeded = false;
		bool empty = true;
		for (int i = 0; i < iSize; i++)
		{
			if (IsComplete(i * PARTSIZE, ((i + 1) * PARTSIZE) - 1))
				empty = false;
		}
		if (empty)
			SetStatus(PS_EMPTY);
	}

	if (gaplist.IsEmpty())
	{	// is this file complete already?
		CompleteFile(false);
		return true;
	}

//	check date of .part file - if its wrong, rehash file
	CFileStatus filestatus;
	m_hPartFile.GetStatus(filestatus); // this; "...returns m_attribute without high-order flags" indicates a known MFC bug, wonder how many unknown there are... :)
	m_timeLastDownTransfer = filestatus.m_mtime;
	if (m_timetLastWriteDate != mktime(filestatus.m_mtime.GetLocalTm()))
	{
		AddLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_REHASH), m_strFileName);
	//	rehash
		SetStatus(PS_WAITINGFORHASH);
		CHashFileThread* addfilethread = (CHashFileThread*) AfxBeginThread(RUNTIME_CLASS(CHashFileThread), THREAD_PRIORITY_NORMAL + g_eMuleApp.m_pGlobPrefs->GetMainProcessPriority(), 0, CREATE_SUSPENDED);
		addfilethread->SetValues(0, directory, m_hPartFile.GetFileName().GetBuffer(), this);
		addfilethread->ResumeThread();
	}

	UpdateCompletedInfos();
	if (completedsize > m_dwBytesTransferred)
		m_iGainDueToCompression = completedsize - m_dwBytesTransferred;
	else if (completedsize != m_dwBytesTransferred)
		m_iLostDueToCorruption = m_dwBytesTransferred - completedsize;
//bond006 (smoothed "remaining time" calculation)
	m_SessionStartSize = completedsize;
	if (m_bPaused)
		SetStartTimeReset(true);

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::SavePartFile()
{
	EMULE_TRY

	switch (m_eStatus)
	{
		case PS_WAITINGFORHASH:
		case PS_HASHING:
		case PS_COMPLETING:
		case PS_COMPLETE:
		return false;
	}

	CSingleLock sLock(&m_FileCompleteMutex, TRUE); // will be unlocked on exit

	//get filedate
	CFileFind ff;
	CString strSearchPath = m_strFullName.Left(m_strFullName.GetLength() - 4);
	bool bEnd = !ff.FindFile(strSearchPath, 0);
	if (!bEnd)
		ff.FindNextFile();
	if (bEnd || ff.IsDirectory())
	{
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_SAVEMET), GetResString(IDS_ERR_PART_FNF), m_strPartMetFileName, m_strFileName);
		return false;
	}

	// create a backup of ".met"
	CString strBakFile = CString(m_strFullName) + CString(".bak");
	::CopyFile(m_strFullName, strBakFile, FALSE);

	CTime lwtime;
	ff.GetLastWriteTime(lwtime);
	m_timetLastWriteDate = mktime(lwtime.GetLocalTm());
	ff.Close();
	uint32 lsc = mktime(lastseencomplete.GetLocalTm());

	//create a temporary file
	CString strTempName(m_strFullName);

	strTempName.Append(_T(".tmp"));
	if (_tremove(strTempName) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strTempName);

	//CFile file;
	CStdioFile file;

	try
	{
	//open the file
		file.Open(strTempName, CFile::modeWrite|CFile::modeCreate|CFile::typeBinary);
		setvbuf(file.m_pStream, NULL, _IOFBF, 16384);
	//version
		uint8 uVersion = PARTFILE_VERSION;
		file.Write(&uVersion, 1);
	//date
		file.Write(&m_timetLastWriteDate, 4);
	//hash
		file.Write(&m_fileHash, 16);
		CSingleLock Lock(&m_mutexHashList, TRUE);
		uint16 uParts = m_partHashArray.GetCount();
		file.Write(&uParts, 2);
		for (int x = 0; x != uParts; x++)
			file.Write(m_partHashArray[x], 16);
		Lock.Unlock();
	//tags
		uint32 dwTagCount = m_tagArray.GetCount() + 8 + (gaplist.GetCount() * 2);
		file.Write(&dwTagCount, 4);
		CFileTag nametag(FT_FILENAME, m_strFileName);
		nametag.WriteToFile(file);
		CFileTag sizetag(FT_FILESIZE, m_lFileSize);
		sizetag.WriteToFile(file);
		CFileTag transtag(FT_TRANSFERRED, m_dwBytesTransferred);
		transtag.WriteToFile(file);
		CFileTag statustag(FT_STATUS, (m_bPaused) ? 1 : 0);
		statustag.WriteToFile(file);

		uint8 priority_value;
		if (IsAutoPrioritized())
			priority_value = PR_AUTO;
		else
			priority_value = priority;
		CFileTag prioritytag(FT_PRIORITY, priority_value);
		prioritytag.WriteToFile(file);

		CFileTag lsctag(FT_LASTSEENCOMPLETE, lsc);
		lsctag.WriteToFile(file);

		uint8 ulpriority_value;
		if (CKnownFile::IsAutoPrioritized())
			ulpriority_value = PR_AUTO;
		else
			ulpriority_value = m_iPriority;
		CFileTag ulprioritytag(FT_ULPRIORITY, ulpriority_value);
		ulprioritytag.WriteToFile(file);

		CFileTag categorytag(FT_CATEGORY, static_cast<int>(m_eCategoryID));
		categorytag.WriteToFile(file);

		for (uint32 j = 0; j != m_tagArray.GetCount();j++)
			m_tagArray[j]->WriteToFile(file);
	//gaps
		TCHAR namebuffer[10];
		TCHAR *number = &namebuffer[1];
		uint16 i_pos = 0;
		for (POSITION pos = gaplist.GetHeadPosition();pos != 0;)
		{
			_itot(i_pos, number, 10);
			namebuffer[0] = FT_GAPSTART;
			CFileTag gapstarttag(namebuffer, gaplist.GetAt(pos)->m_dwStartOffset);
			gapstarttag.WriteToFile(file);
		//	gap start = first missing byte but gap ends = first non-missing byte in edonkey
		//	but I think its easier to user the real limits
			namebuffer[0] = FT_GAPEND;
			CFileTag gapendtag(namebuffer, (gaplist.GetNext(pos)->m_dwEndOffset) + 1);
			gapendtag.WriteToFile(file);
			i_pos++;
		}
	}
	catch(CFileException* error)
	{
		OUTPUT_DEBUG_TRACE();
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_SAVEMET), GetErrorMessage(error), m_strPartMetFileName, m_strFileName);
		error->Delete();
		file.Close();
		_tremove(strTempName); // remove the partially written or otherwise damaged temporary file
		return false;
	}

	_commit(_fileno(file.m_pStream));
	file.Close();

	// delete an original ".met", after successfully writing the temporary part.met file...
	if (_tremove(m_strFullName) != 0 && errno != ENOENT)
		AddDebugLogLine(false, RGB_LOG_ERROR + _T("Failed to remove \"%s\" - %s"), m_strFullName, strerror(errno));

	// rename a ".met.tmp" to ".met"
	if (_trename(strTempName, m_strFullName) != 0)
		AddDebugLogLine(false, RGB_LOG_ERROR + _T("Failed to move temporary part.met file \"%s\" to \"%s\" - %s"), strTempName, m_strFullName, strerror(errno));

	CString dfpath = CString(RemoveFileExtension(m_strFullName)) + CString(_T(".dir"));
	CStdioFile fDir;

	try
	{
	//	enkeyDEV(Ottavio84) -ChangeDir-
		if (!m_strAlternativePath.IsEmpty())
		{
			if (fDir.Open(dfpath, CFile::modeCreate | CFile::modeWrite | CFile::typeText))
			{
				fDir.WriteString((LPCTSTR)m_strAlternativePath);
				fDir.Close();
			}
		}
	}
	catch(CFileException* error)
	{
		OUTPUT_DEBUG_TRACE();
		AddLogLine(false, RGB_LOG_ERROR + _T("Error by saving %s - %s"), dfpath, GetErrorMessage(error));
		error->Delete();
		fDir.Close();
	}

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::PartFileHashFinished(CKnownFile* pKnownFile)
{
	EMULE_TRY

	newdate = true;

	bool	bErrorFound = false;
	CSingleLock Lock(&m_mutexHashList, TRUE);
	INT_PTR hashSize = m_partHashArray.GetSize();

	for (uint32 i = 0; i != hashSize; i++)
	{
		if (IsComplete(i * PARTSIZE, ((i + 1) * PARTSIZE) - 1))
		{
			if (!(pKnownFile->GetPartHash(i) && !md4cmp(pKnownFile->GetPartHash(i), GetPartHash(i))))
			{
				AddLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_FOUNDCORRUPTION), i + 1, m_strFileName);
				AddGap(i * PARTSIZE, ((((i + 1) * PARTSIZE) - 1) >= m_lFileSize) ? m_lFileSize - 1 : ((i + 1) * PARTSIZE) - 1);
				bErrorFound = true;
			}
		}
	}
	Lock.Unlock();
	delete pKnownFile;
	if (!bErrorFound)
	{
		if (m_eStatus == PS_COMPLETING)
		{
			CompleteFile(true);
			return;
		}
		else
			AddLogLine(false, IDS_HASHINGDONE, m_strFileName);
	}
	else
	{
		SetStatus(PS_READY);
		SavePartFile();
	//	If a file is "ready" it can't be paused
		m_bPaused = false;
		return ;
	}
	SetStatus(PS_READY);
	SavePartFile();
	g_eMuleApp.m_pSharedFilesList->SafeAddKnownFile(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddGap(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = gaplist.GetHeadPosition();(pos2 = pos1) != NULL;)
	{
		Gap_Struct* cur_gap = gaplist.GetNext(pos1);
		if (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwEndOffset <= dwEnd)
		{ // this gap is inside the new gap - delete
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwStartOffset <= dwEnd)
		{ // a part of this gap is in the new gap - extend limit and delete
			dwEnd = cur_gap->m_dwEndOffset;
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->m_dwEndOffset <= dwEnd && cur_gap->m_dwEndOffset >= start)
		{ // a part of this gap is in the new gap - extend limit and delete
			start = cur_gap->m_dwStartOffset;
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (start >= cur_gap->m_dwStartOffset && dwEnd <= cur_gap->m_dwEndOffset)
		{ // new gap is already inside this gap - return
			return ;
		}
	}
	Gap_Struct* new_gap = new Gap_Struct;
	new_gap->m_dwStartOffset = start;
	new_gap->m_dwEndOffset = dwEnd;
	gaplist.AddTail(new_gap);
	UpdateDisplayedInfo();
	newdate = true;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsComplete() returns false if there are gaps in the range 'start' to 'dwEnd' (inclusive).
bool CPartFile::IsComplete(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	if (dwEnd >= m_lFileSize)
		dwEnd = m_lFileSize - 1;
	for (POSITION pos = gaplist.GetHeadPosition(); pos != 0;)
	{
		Gap_Struct * cur_gap = gaplist.GetNext(pos);

		if ( (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwEndOffset <= dwEnd)	// gap is inside range (this check is unnecessary)
		     || (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwStartOffset <= dwEnd)	// gap starts inside range
		     || (cur_gap->m_dwEndOffset <= dwEnd && cur_gap->m_dwEndOffset >= start)		// gap ends inside range
		     || (start >= cur_gap->m_dwStartOffset && dwEnd <= cur_gap->m_dwEndOffset) )	// range is inside gap
		{
			return false;
		}
	}
	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsPureGap(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	if (dwEnd >= m_lFileSize)
		dwEnd = m_lFileSize - 1;
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;)
	{
		Gap_Struct * cur_gap = gaplist.GetNext(pos);
		if (start >= cur_gap->m_dwStartOffset && dwEnd <= cur_gap->m_dwEndOffset)
		{
			return true;
		}
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsAlreadyRequested(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	for (POSITION pos = requestedblocks_list.GetHeadPosition();pos != 0;)
	{
		Requested_Block_Struct * cur_block = requestedblocks_list.GetNext(pos);
		if ((start <= cur_block->m_dwEndOffset) && (dwEnd >= cur_block->m_dwStartOffset))
			return true;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::GetNextEmptyBlockInPart(uint16 uPartNumber, Requested_Block_Struct *pNextEmptyBlock)
{
	EMULE_TRY

	Gap_Struct	   *pFirstGap = NULL;
	Gap_Struct	   *pCurrentGap = NULL;
	uint32			dwEnd;
	uint32			dwBlockLimit;

//	Find start of this part
	uint32		partStart = (PARTSIZE * uPartNumber);
	uint32		start = partStart;

//	What is the end limit of this block, i.e. can't go outside part (or filesize)
	uint32 partEnd = (PARTSIZE * (uPartNumber + 1)) - 1;
	if (partEnd >= GetFileSize())
		partEnd = GetFileSize() - 1;

//	Loop until find a suitable gap and return true, or no more gaps and return false
	while (true)
	{
		pFirstGap = NULL;

	//	Find the first gap from the start position
		for (POSITION pos = gaplist.GetHeadPosition(); pos != 0;)
		{
			pCurrentGap = gaplist.GetNext(pos);
		//	Want gaps that overlap start<->partEnd
			if ((pCurrentGap->m_dwStartOffset <= partEnd) && (pCurrentGap->m_dwEndOffset >= start))
			{
			//	Is this the first gap?
				if ((pFirstGap == NULL) || (pCurrentGap->m_dwStartOffset < pFirstGap->m_dwStartOffset))
					pFirstGap = pCurrentGap;
			}
		}

	//	If no gaps after start, exit
		if (pFirstGap == NULL)
			return false;

	//	Update start position if gap starts after current pos
		if (start < pFirstGap->m_dwStartOffset)
			start = pFirstGap->m_dwStartOffset;

	//	If this is not within part, exit
		if (start > partEnd)
			return false;

	//	Find end, keeping within the max block size and the part limit
		dwEnd = pFirstGap->m_dwEndOffset;
		dwBlockLimit = partStart + (EMBLOCKSIZE * (((start - partStart) / EMBLOCKSIZE) + 1)) - 1;
		if (dwEnd > dwBlockLimit)
			dwEnd = dwBlockLimit;
		if (dwEnd > partEnd)
			dwEnd = partEnd;

	//	If this gap has not already been requested, we have found a valid entry
		if (!IsAlreadyRequested(start, dwEnd))
		{
		//	Was this block to be returned
			if (pNextEmptyBlock != NULL)
			{
				pNextEmptyBlock->m_dwStartOffset = start;
				pNextEmptyBlock->m_dwEndOffset = dwEnd;
				md4cpy(pNextEmptyBlock->m_fileHash, GetFileHash());
				pNextEmptyBlock->m_dwBytesTransferred = 0;
			}
			return true;
		}
		else
		{
		//	Reposition to end of that gap
			start = dwEnd + 1;
		}

	//	If tried all gaps then break out of the loop
		if (dwEnd == partEnd)
			break;
	}

	EMULE_CATCH

//	No suitable gap found
	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::GetGapsInPart(CUpDownClient* pSource, uint16 numPart, Requested_Block_Struct** newblocks, uint16* count)
{
	EMULE_TRY

//	Calculate offsets of the Part
	const uint32 uPartStart = numPart * PARTSIZE;
	const uint32 uPartEnd = ((GetFileSize() - 1) < (uPartStart + PARTSIZE - 1)) ? (GetFileSize() - 1) : (uPartStart + PARTSIZE - 1);

	int newblockcount = 0;

//	Retrieve Gap(s) matching the part
	for (POSITION pos = gaplist.GetHeadPosition(); pos != NULL;)
	{
		Gap_Struct *CurGap = gaplist.GetNext(pos);

	//	Check if Gap is inside the limit
		if (CurGap->m_dwStartOffset > uPartEnd || CurGap->m_dwEndOffset < uPartStart)
			continue;

	//	Calculate offsets of the Gap within the part
		uint32 uGapStart = (CurGap->m_dwStartOffset < uPartStart) ? uPartStart : CurGap->m_dwStartOffset;
		const uint32 uGapEnd = (CurGap->m_dwEndOffset > uPartEnd) ? uPartEnd : CurGap->m_dwEndOffset;

	//	Create n request block(s)
		while (uGapEnd >= uGapStart && newblockcount < *count)
		{
		//	Calculate offsets of the block
			const uint32 startOffset = uGapStart;
			const uint32 endOffset = ((uGapStart + EMBLOCKSIZE - 1) > uGapEnd) ? uGapEnd : ((uGapStart + EMBLOCKSIZE - 1) - (uGapStart % EMBLOCKSIZE));

		//	Prepare offset of next block
			uGapStart = endOffset + 1;

		//	Check if block has already been requested
			if (IsAlreadyRequested(startOffset, endOffset) == false)
			{
			//	Create 1 request block
				Requested_Block_Struct * block = new Requested_Block_Struct;
				block->m_dwStartOffset = startOffset;
				block->m_dwEndOffset = endOffset;
			//	block->m_dwPackedSize = 0;
				block->m_dwBytesTransferred = 0;
				md4cpy(block->m_fileHash, GetFileHash());

			//	Flag the block as 'requested'
				requestedblocks_list.AddTail(block);

			//	Return the block to the source
				newblocks[newblockcount] = block;
				newblockcount++;
			}
		}
	}
	*count = newblockcount;

	return ;

	EMULE_CATCH

	*count = 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::FillGap(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = gaplist.GetHeadPosition();(pos2 = pos1) != NULL;)
	{
		Gap_Struct* cur_gap = gaplist.GetNext(pos1);
		if (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwEndOffset <= dwEnd)
		{ // our part fills this gap completly
			gaplist.RemoveAt(pos2);
			delete cur_gap;
			continue;
		}
		else if (cur_gap->m_dwStartOffset >= start && cur_gap->m_dwStartOffset <= dwEnd)
		{ // a part of this gap is in the part - set limit
			cur_gap->m_dwStartOffset = dwEnd + 1;
		}
		else if (cur_gap->m_dwEndOffset <= dwEnd && cur_gap->m_dwEndOffset >= start)
		{ // a part of this gap is in the part - set limit
			cur_gap->m_dwEndOffset = start - 1;
		}
		else if (start >= cur_gap->m_dwStartOffset && dwEnd <= cur_gap->m_dwEndOffset)
		{
			uint32 buffer = cur_gap->m_dwEndOffset;
			cur_gap->m_dwEndOffset = start - 1;
			cur_gap = new Gap_Struct;
			cur_gap->m_dwStartOffset = dwEnd + 1;
			cur_gap->m_dwEndOffset = buffer;
			gaplist.InsertAfter(pos1, cur_gap);
			break;
		}
	}

	UpdateCompletedInfos();
	UpdateDisplayedInfo();
	newdate = true;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::UpdateCompletedInfos()
{
	EMULE_TRY

	uint32 allgaps = 0;

	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;)
	{
		Gap_Struct * cur_gap = gaplist.GetNext(pos);
		allgaps += cur_gap->m_dwEndOffset - cur_gap->m_dwStartOffset;
	}

	if ((gaplist.GetCount() || requestedblocks_list.GetCount()))
	{
		m_dblPercentCompleted = ((1.0 - static_cast<double>(allgaps + 1) / m_lFileSize)) * 100;
		completedsize = (uint32)((m_lFileSize - (allgaps + 1)));
	}
	else
	{
		m_dblPercentCompleted = 100.0;
		completedsize = m_lFileSize;
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::DrawStatusBar(CDC* dc, RECT* rect, bool bFlat)
{
	EMULE_TRY

	const COLORREF crProgress = (bFlat) ? RGB(0, 150, 0) : RGB(0, 192, 0);
	COLORREF crHave;
	COLORREF crPending;
	COLORREF crMissing;

//	SLUGFILLER: grayPause - Colors by status
	if (GetStatus() == PS_EMPTY || GetStatus() == PS_READY)
	{
		crHave = (bFlat) ? RGB(0, 0, 0) : RGB(95, 95, 95);
		crPending = RGB(255, 208, 0);
		crMissing = RGB(255, 0, 0);
	}
	else
	{
		crHave = (bFlat) ? RGB(105, 105, 105) : RGB(142, 142, 142);
		crPending = RGB(255, 240, 142);
		crMissing = RGB(255, 142, 142);
	}
//	SLUGFILLER

	CBarShader s_ChunkBar(rect->bottom - rect->top, rect->right - rect->left);
	s_ChunkBar.SetFileSize(m_lFileSize);
	s_ChunkBar.Fill(crHave);

	uint32 allgaps = 0;

	if (m_eStatus == PS_COMPLETE || m_eStatus == PS_COMPLETING)
	{
		s_ChunkBar.FillRange(0, m_lFileSize, crProgress);
		s_ChunkBar.Draw(dc, rect->left, rect->top, bFlat);
		m_dblPercentCompleted = 100.0;
		completedsize = m_lFileSize;
		return ;
	}

//	red gaps
	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;)
	{
		Gap_Struct * cur_gap = gaplist.GetNext(pos);
		allgaps += cur_gap->m_dwEndOffset - cur_gap->m_dwStartOffset;
		bool gapdone = false;
		uint32 gapstart = cur_gap->m_dwStartOffset;
		uint32 gapend = cur_gap->m_dwEndOffset;
		for (uint32 i = 0; i < GetPartCount(); i++)
		{
			if (gapstart >= i * PARTSIZE && gapstart <= (i + 1) * PARTSIZE)
			{ // is in this part?
				if (gapend <= (i + 1) * PARTSIZE)
					gapdone = true;
				else
				{
					gapend = (i + 1) * PARTSIZE; // and next  part
				}
			//	paint
				COLORREF color;

				if (m_srcPartFrequencies.GetCount() >= (INT_PTR)i && m_srcPartFrequencies[static_cast<uint16>(i)])  // frequency?
				//	SLUGFILLER: grayPause
				{
					if (m_eStatus == PS_EMPTY || m_eStatus == PS_READY)
						color = RGB( 0,
						             (210 - (22 * (m_srcPartFrequencies[(uint16)i] - 1)) < 0) ? 0 : 210 - (22 * (m_srcPartFrequencies[(uint16)i] - 1))
						             , 255 );
					else
						color = RGB( 100,
						             (210 - (22 * (m_srcPartFrequencies[(uint16)i] - 1)) < 0) ? 0 : 255 - (22 * (m_srcPartFrequencies[(uint16)i] - 1))
						             , 255 );
				}
			//	SLUGFILLER: grayPause
				else
					color = crMissing;

				s_ChunkBar.FillRange(gapstart, gapend + 1, color);

				if (gapdone) // finished?
					break;
				else
				{
					gapstart = gapend;
					gapend = cur_gap->m_dwEndOffset;
				}
			}
		}
	}

//	yellow pending parts
	for (POSITION pos = requestedblocks_list.GetHeadPosition();pos != 0;)
	{
		Requested_Block_Struct * block = requestedblocks_list.GetNext(pos);
		s_ChunkBar.FillRange((block->m_dwStartOffset + block->m_dwBytesTransferred), block->m_dwEndOffset, crPending);
	}

	s_ChunkBar.Draw(dc, rect->left, rect->top, bFlat);

//--- xrmb:confirmeddownload ---
	if ((gaplist.GetCount() || requestedblocks_list.GetCount()))
	{
	//	all this here should be done in the Process, not in the drawing!!! its a waste of cpu time, or maybe not :)
		uint32	completedParts = 0;
		confirmedsize = 0;
		for (uint32 i = 0; i < GetPartCount(); i++)
		{
			uint32	dwEnd = (i + 1) * PARTSIZE - 1;
			bool	lastChunk = false;
		//--- last part? ---
			if (dwEnd > m_lFileSize)
			{
				dwEnd = m_lFileSize;
				lastChunk = true;
			}

			if (IsComplete(i * PARTSIZE, dwEnd))
			{
				completedParts++;

				if (lastChunk == false)
					confirmedsize += PARTSIZE;
				else
					confirmedsize += m_lFileSize % PARTSIZE;
			}
		}

		completedsize = (uint32)((m_lFileSize - (allgaps + 1)));

		m_dblPercentCompleted = static_cast<double>(completedsize) / m_lFileSize * 100.0;
		m_dblPercentConfirmed = static_cast<double>(confirmedsize) / m_lFileSize * 100.0;
	}
	else
	{
		m_dblPercentCompleted = 100.0;
		m_dblPercentConfirmed = 100.0;
		completedsize = m_lFileSize;
	}

//	green progress
	RECT gaprect;
	gaprect.top = rect->top;
	gaprect.bottom = gaprect.top + PROGRESS_HEIGHT; // Barry - was 4
	gaprect.left = rect->left;

	uint32	w = rect->right - rect->left + 1;
	uint32	wc = (uint32)(m_dblPercentConfirmed / 100.0 * w + 0.5);
	uint32	wp = (uint32)(m_dblPercentCompleted / 100.0 * w + 0.5);

	if (!bFlat)
	{
		CBarShader s_LoadBar(PROGRESS_HEIGHT); // No use of static element
		s_LoadBar.SetWidth(wp);
		s_LoadBar.SetFileSize(completedsize);
		s_LoadBar.Fill(crPending);
		s_LoadBar.FillRange(0, confirmedsize, crProgress);
		s_LoadBar.Draw(dc, gaprect.left, gaprect.top, false);
	}
	else
	{
		int	left = rect->left;
		CBrush* pOldBrush;

	//	green
		gaprect.right = left + wc;
		CBrush pr_Brush(crProgress);
		pOldBrush = dc->SelectObject(&pr_Brush);
		dc->FillRect(&gaprect, &pr_Brush);

	//	yellow
		gaprect.left = left + wc;
		gaprect.right = left + wp;
		CBrush pe_Brush(crPending);
		dc->SelectObject(&pe_Brush);
		dc->FillRect(&gaprect, &pe_Brush);

	//draw gray progress only if flat
		gaprect.left = left + wp;
		gaprect.right = left + w;
		CBrush rgb_Brush(RGB(224, 224, 224));
		dc->SelectObject(&rgb_Brush);
		dc->FillRect(&gaprect, &rgb_Brush);

		dc->SelectObject(pOldBrush);
	}
//--- :xrmb ---

	if ((gaplist.GetCount() || requestedblocks_list.GetCount()))
	{
		m_dblPercentCompleted = ((1.0 - static_cast<double>(allgaps + 1) / m_lFileSize)) * 100;
		completedsize = (uint32)((m_lFileSize - (allgaps + 1)));
	}
	else
	{
		m_dblPercentCompleted = 100.0;
		completedsize = m_lFileSize;
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::WritePartStatus(CFile* file)
{
	EMULE_TRY

	uint16 parts = GetPartCount();
	file->Write(&parts, 2);
	uint16 done = 0;
	while (done != parts)
	{
		uint8 towrite = 0;
		for (uint32 i = 0; i < 8; i++)
		{
		//--- xrmb:partprio ---
			if (GetPartStatus(done) == PR_PART_ON && IsComplete(done * PARTSIZE, ((done + 1) * PARTSIZE) - 1))
			//--- :xrmb ---
				towrite |= (1 << i);

			done++;
			if (done == parts)
				break;
		}
		file->Write(&towrite, 1);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::HasParts()
{
	EMULE_TRY

	uint16 parts = GetPartCount();
	for (uint16 i = 0; i < parts; i++)
	{
		if (GetPartStatus(i) != PR_PART_OFF && IsComplete(i * PARTSIZE, ((i + 1) * PARTSIZE) - 1))
			return true;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	netwolf: complete sources (zegzav) 03.05.03
void CPartFile::WriteCompleteSourcesCount(CFile* file)
{
	uint16 completecount = GetCompleteSourcesCount();
	file->Write(&completecount, 2);
}

int CPartFile::GetValidSourcesCount()
{
	EMULE_TRY

	int counter = 0;
	POSITION pos;
	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		if (!m_srcLists[sl].IsEmpty())
			for (pos = m_srcLists[sl].GetHeadPosition();pos != NULL;)
			{
				CUpDownClient* pSource = m_srcLists[sl].GetNext(pos);
				if ( pSource->GetDownloadState() != DS_ONQUEUE && pSource->GetDownloadState() != DS_DOWNLOADING &&
				     pSource->GetDownloadState() != DS_NONEEDEDPARTS )
					counter++;
			}
	return counter;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16 CPartFile::GetNotCurrentSourcesCount()
{
	EMULE_TRY

	uint16 counter = 0;

	POSITION pos;
	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		if (!m_srcLists[sl].IsEmpty())
			for (pos = m_srcLists[sl].GetHeadPosition();pos != NULL;)
			{
				CUpDownClient* pSource = m_srcLists[sl].GetNext(pos);
				if (pSource->GetDownloadState() != DS_ONQUEUE && pSource->GetDownloadState() != DS_DOWNLOADING)
					counter++;
			}
	return counter;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	netwolf: complete sources (zegzav)
uint16	CPartFile::GetCompleteSourcesCount()
{
	if (!m_bUpdateCompleteSrcCount)
		return m_nLastCompleteSrcCount;

	uint16 counter = 0;

	POSITION pos;
	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		if (!m_srcLists[sl].IsEmpty())
			for (pos = m_srcLists[sl].GetHeadPosition();pos != NULL;)
			{
				CUpDownClient* pSource = m_srcLists[sl].GetNext(pos);
				if (pSource->IsCompleteSource())
					counter++;
			}
	m_nLastCompleteSrcCount = counter;
	m_bUpdateCompleteSrcCount = false;
	return counter;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
EnumPartFileStatuses CPartFile::GetStatus(bool bIgnorePause)
{
	EMULE_TRY

	if (!m_bPaused || m_eStatus == PS_STOPPED || m_eStatus == PS_ERROR || m_eStatus == PS_COMPLETING || m_eStatus == PS_COMPLETE || bIgnorePause)
		return m_eStatus;
	else
	{
		if (m_bStopped)
			return PS_STOPPED;
		else
			return PS_PAUSED;
	}

	EMULE_CATCH

	return PS_READY;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Inserted by Vorlost for load rarest parts first, code by Denominator
uint16 CPartFile::GetRemainingBlocksInPart(uint16 partnumber)
{
	EMULE_TRY

	uint16		uNumRemainingBlocks = 0;
	for (uint32 i = 0; i < ceil((double)PARTSIZE / EMBLOCKSIZE);i++)
	{
		uint32 start = (PARTSIZE * partnumber) + i * EMBLOCKSIZE;
		uint32 dwEnd = (PARTSIZE * partnumber) + ((i + 1) * EMBLOCKSIZE) - 1;
		if (dwEnd >= PARTSIZE * ((uint32)partnumber + 1))
			dwEnd = (PARTSIZE * (partnumber + 1)) - 1;
		if (start >= GetFileSize())
			break;
		if (dwEnd >= GetFileSize())
			dwEnd = GetFileSize() - 1;
		if (!IsComplete(start, dwEnd) && !IsAlreadyRequested(start, dwEnd))
			uNumRemainingBlocks++;
	}
	return uNumRemainingBlocks;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Remove no needed sources from download queue
//	[In] number_sources_to_remove: Number of sources to be removed from list or 0xffffffff if we want all removed
void CPartFile::RemoveNoNeededSources(uint32 number_sources_to_remove)
{
	EMULE_TRY

	POSITION pos;
	uint32 removed = 0;
	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		if (!m_srcLists[sl].IsEmpty())
			for (pos = m_srcLists[sl].GetHeadPosition(); pos != NULL && removed < number_sources_to_remove;)
			{
				CUpDownClient* pSource = m_srcLists[sl].GetNext(pos);
				if ((pSource->GetDownloadState() == DS_NONEEDEDPARTS) || (pSource->IsRemoteQueueFull()))  //netwolf: katsyonak change
				{
					g_eMuleApp.m_pDownloadQueue->RemoveSource(pSource);
					removed ++; // Count removed source
				}
			}

	EMULE_CATCH

	lastpurgetime = ::GetTickCount();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CPartFile::Process(uint32 dwReduceDownload /*in percent*/, uint32 dwIteration)
{
	EMULE_TRY

	uint16			nOldTransSourceCount = GetTransferringSrcCount();
	uint16			iOldSrcOnQueueCount = GetOnQueueSrcCount();
	DWORD			dwCurTick = ::GetTickCount();
	CUpDownClient  *pSource = NULL;
	bool			bSocketsUsageIsHigh = false;
	bool			bIsClientConnected = false;
	bool			bReaskDownload = false;
	POSITION		pos1, pos2;
	uint32			dwCurrentDataRate = 0;
	uint32			dwLastNoNeededCheckTime = 0;

	m_dataRate = 0;

#ifdef LATER
#ifdef _DEBUG
	if (dwReduceDownload > 0)
	{
		AddLogLine(true, _T("Process: Throttling download to %ld%%"), dwReduceDownload);
	}
#endif
#endif LATER

//	If buffer size exceeds limit, or if not written within time limit, flush data
	if ( (m_nTotalBufferData > ((uint32)g_eMuleApp.m_pGlobPrefs->GetFileBufferSize()) * 30000)
	  || (dwCurTick > (m_nLastBufferFlushTime + BUFFER_TIME_LIMIT)) ) //netwolf: increased FileBufferSize
	{
	//	Avoid flushing while copying preview file
		if (!m_bPreviewing)
			FlushBuffer();
	}

#ifdef OLD_SOCKETS_ENABLED
	bSocketsUsageIsHigh = g_eMuleApp.m_pListenSocket->GetNumOpenSockets() > (g_eMuleApp.m_pGlobPrefs->GetMaxConnections() * 0.75);
#endif //OLD_SOCKETS_ENABLED

	m_iNumTransferringSrcs = 0;

//	calculate data rate, set limit etc.
//	9 iterations for download & 1 iteration for other files
	if (dwIteration < 10 && !m_downloadingSourceList.IsEmpty())
	{
		for (pos1 = m_downloadingSourceList.GetHeadPosition(); (pos2 = pos1) != NULL;)
		{
			pSource = m_downloadingSourceList.GetNext(pos1);
			if (pSource)
			{
				//eklmn: if new client has non DL state, then remove him
				if (pSource->GetDownloadState()!= DS_DOWNLOADING)
				{
					m_downloadingSourceList.RemoveAt(pos2);
					continue;
				}
				dwCurrentDataRate = pSource->CalculateDownloadRate();
				//eklmn: since client can be droped due timeout it needs to check the state
				if (pSource->GetDownloadState() != DS_DOWNLOADING)
				{
					continue;
				}
				m_iNumTransferringSrcs++;
				m_dataRate += dwCurrentDataRate;
				if (dwReduceDownload)
				{
					uint32 limit = dwReduceDownload * dwCurrentDataRate / 1000;
					if (limit < 1000 && dwReduceDownload == 200)
						limit += 1000;
					else if (limit < 1)
						limit = 1;
#ifdef OLD_SOCKETS_ENABLED

					pSource->m_pRequestSocket->SetDownloadLimit(limit);
#endif OLD_SOCKETS_ENABLED
				}
				else if (m_nSavedReduceDownload != dwReduceDownload)
#ifdef OLD_SOCKETS_ENABLED
				{
					pSource->m_pRequestSocket->DisableDownloadLimit();
				}
#endif OLD_SOCKETS_ENABLED

				m_nSavedReduceDownload = dwReduceDownload;
			}
		}
	}
	else
	{
	//	Initialize temp variables only if we need it
		bool	bAlreadyCounted = false;
		uint16	uNNPSrcCount = 0;
		uint16	uOnQueueSrcCount = 0;
		uint16	uHighQRSrcCount = 0;
		uint16	uConnectingSrcCount = 0;
		uint16	uTooManyConnsSrcCount = 0;
		uint16	uConnectedSrcCount = 0;
		uint16	uTimeoutSrcCount = 0;
		uint16	uConnViaServerSrcCount = 0;
		uint16	uUnknownSrcCount = 0;
		uint16	uReqHashSrcCount = 0;
		uint16	uLowToLowIDSrcCount = 0;
		uint16	uQueueFullSrcCount = 0;
		uint16	uNoUploadSrcCount = 0;
		uint16	uBannedSrcCount = 0;
		uint16	uErrorSrcCount = 0;

		for (uint32 dwSlotIndex = 0; dwSlotIndex < SOURCESSLOTS; dwSlotIndex++)
		{
			if (!m_srcLists[dwSlotIndex].IsEmpty())
			{
				for (pos1 = m_srcLists[dwSlotIndex].GetHeadPosition();pos1 != NULL;)
				{
					pSource = m_srcLists[dwSlotIndex].GetNext(pos1);

					if (!pSource)
						continue;

					bAlreadyCounted = false;
					bIsClientConnected = true;
					bReaskDownload = false;
					uint32 usethisfilereasktime = FILEREASKTIME;

					switch (pSource->GetDownloadState())
					{
						case DS_DOWNLOADING:
						{
							dwCurrentDataRate = pSource->CalculateDownloadRate();
						// eklmn: since client can be droped due timeout it needs to check the state
							if (pSource->GetDownloadState() != DS_DOWNLOADING)
								break;
						// DonGato: needed to update the source count
						// GetTransferringSrcCount() isn't updated inside this function.
							m_iNumTransferringSrcs++;
							m_dataRate += dwCurrentDataRate;
							if (dwReduceDownload)
							{
								uint32 limit = dwReduceDownload * dwCurrentDataRate / 1000;
								if (limit < 1000 && dwReduceDownload == 200)
									limit += 1000;
								else if (limit < 1)
									limit = 1;
#ifdef OLD_SOCKETS_ENABLED

								pSource->m_pRequestSocket->SetDownloadLimit(limit);
#endif //OLD_SOCKETS_ENABLED

							}
#ifdef OLD_SOCKETS_ENABLED
							else if (m_nSavedReduceDownload != dwReduceDownload)
								pSource->m_pRequestSocket->DisableDownloadLimit();
#endif //OLD_SOCKETS_ENABLED

							m_nSavedReduceDownload = dwReduceDownload;
							break;
						}

						case DS_BANNED:
							if (!bAlreadyCounted)
							{
								uBannedSrcCount++;
								bAlreadyCounted = true;
							}
							break;

						case DS_ERROR:
							if (!bAlreadyCounted)
							{
								uErrorSrcCount++;
								bAlreadyCounted = true;
								pSource->Disconnected();
							}
							break;

						case DS_CONNECTED:
					//	if m_pRequestSocket doesn't exist or isn't connected
#ifdef OLD_SOCKETS_ENABLED

							if (!(pSource->m_pRequestSocket && pSource->m_pRequestSocket->IsConnected()))
							{
							//	DonGato: it was DS_NONE before but that keeps for some time problematic clients.
							//	There is no need to do that as there will be no change (from our experience).
								pSource->SetDownloadState(DS_ERROR);

								if (!bAlreadyCounted)
								{
									uUnknownSrcCount++;
									bAlreadyCounted = true;
								}
								break;
							}

#endif //OLD_SOCKETS_ENABLED

							if (dwCurTick - pSource->GetEnteredConnectedState() > CONNECTION_TIMEOUT + 20000)
							{
								if (!bAlreadyCounted)
								{
									uTimeoutSrcCount++;
									bAlreadyCounted = true;
								}
								g_eMuleApp.m_pDownloadQueue->RemoveSource(pSource);
								break;
							}
							if (!bAlreadyCounted)
							{
								uConnectedSrcCount++;
								bAlreadyCounted = true;
							}
							bReaskDownload = true;
							break;

						case DS_LOWTOLOWIP: 	// if we now have a high ip we can ask
							if (!bAlreadyCounted)
							{
								uLowToLowIDSrcCount++;
								bAlreadyCounted = true;
							}
							if ( ((dwCurTick - lastpurgetime) > PURGE_TIME)
								&& (GetSourceCount() >= (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() * .8)) )
							{
								g_eMuleApp.m_pDownloadQueue->RemoveSource(pSource);
								lastpurgetime = dwCurTick;
								break;
							}
#ifdef OLD_SOCKETS_ENABLED
							if (g_eMuleApp.m_pServerConnect->IsLowID())
								break;
#endif //OLD_SOCKETS_ENABLED

						case DS_NONEEDEDPARTS:
						{
							if (!bAlreadyCounted)
							{
								uNNPSrcCount++;
								bAlreadyCounted = true;
							}
							if ( ((dwCurTick - lastpurgetime) > 40000)
							     && (GetSourceCount() >= (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() * .8)) )
								if (!pSource->SwapToAnotherFile(NULL, true))
								{
									g_eMuleApp.m_pDownloadQueue->RemoveSource(pSource);
									lastpurgetime = dwCurTick;
									break;
								}

						//	doubled reasktime for no needed parts - save connections and traffic
						//	Vorlost, noneededparts-sources will be checked filereasktime*4 when 3/4 of max. sockets are open
							if (bSocketsUsageIsHigh)
								usethisfilereasktime = FILEREASKTIME * 2;

							if (!((!pSource->GetLastAskedTime()) || (dwCurTick - pSource->GetLastAskedTime()) > usethisfilereasktime * 2))
								break;
						}
						case DS_ONQUEUE:
						{
							if (pSource->IsRemoteQueueFull())
							{
								if (!bAlreadyCounted)
								{
									uQueueFullSrcCount++;
									bAlreadyCounted = true;
								}
								if ( ((dwCurTick - lastpurgetime) > 60000)
								     && (GetSourceCount() >= (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() * .8)) )
								{
									g_eMuleApp.m_pDownloadQueue->RemoveSource(pSource);
									lastpurgetime = dwCurTick;
									break;
								}
							}
						//	Check Remote queue position (HighQR)
							else if ( g_eMuleApp.m_pGlobPrefs->UseRemoteQueueLimit()
							          && (pSource->GetRemoteQueueRank() > g_eMuleApp.m_pGlobPrefs->GetRemoteQueueLimit()) )
							{
								if (!bAlreadyCounted)
								{
									uHighQRSrcCount++;
									bAlreadyCounted = true;
								}
								g_eMuleApp.m_pDownloadQueue->RemoveSource(pSource);
								break;
							}
							else if (!bAlreadyCounted)
							{
								uOnQueueSrcCount++;
								bAlreadyCounted = true;
							}

							if (pSource->GetRemoteQueueRank() && bSocketsUsageIsHigh)  // Vorlost QueueRank > 250
							{
								if (pSource->GetRemoteQueueRank() > 250)
									usethisfilereasktime = FILEREASKTIME * 2.5 ; // Vorlost 54:10 minutes
								else
									usethisfilereasktime = FILEREASKTIME * 2; // Vorlost (QR 1-250 = 43:20 minutes)
							}

						//	LANCAST (moosetea) - We reask much faster if lancast user, UDP happens in 10 seconds
							if (pSource->IsOnLAN())
							{
								if ( (!pSource->GetLastAskedTime())
								     || (::GetTickCount() - pSource->GetLastAskedTime()) > LANCASTFILEREASKTIMEUDP )
								{
									pSource->UDPReaskForDownload();
								}
							}
							else
							{
#ifdef OLD_SOCKETS_ENABLED
								if ( g_eMuleApp.m_pServerConnect->IsConnected()
								     && ( (!pSource->GetLastAskedTime())
								          || (dwCurTick - pSource->GetLastAskedTime()) > usethisfilereasktime - 20000 ) )
								{
									pSource->UDPReaskForDownload();
								}
#endif //OLD_SOCKETS_ENABLED

							}

						//	START enkeyDEV(th1) -L2HAC- highid side, "prepare" phase
						//	(emulates a "Connecting via server..." without m_pRequestSocket timeout)
#ifdef OLD_SOCKETS_ENABLED
							if ( g_eMuleApp.m_pServerConnect->IsConnected()
							     && (pSource->GetLastAskedTime())
							     && ((dwCurTick - pSource->GetLastAskedTime()) > (FILEREASKTIME - L2HAC_PREPARE_PRECEDE))
							     && pSource->HasLowID()
							     && pSource->GetL2HACTime()
							     && !g_eMuleApp.m_pServerConnect->IsLowID() )
							{
								pSource->SetDownloadState(DS_WAITCALLBACK);
							}
#endif //OLD_SOCKETS_ENABLED
						}
						case DS_CONNECTING:
							{
								if (!bAlreadyCounted)
								{
									uConnectingSrcCount++;
									bAlreadyCounted = true;
								}
							}
						case DS_TOOMANYCONNS:
							{
								if (!bAlreadyCounted)
								{
									uTooManyConnsSrcCount++;
									bAlreadyCounted = true;
								}
							}
						case DS_NONE:
							{
								if (!bAlreadyCounted)
								{
									uNoUploadSrcCount++;
									bAlreadyCounted = true;
								}
								bReaskDownload = true;
							}
						case DS_WAITCALLBACK:
							{
								if (!bAlreadyCounted)
								{
									uConnViaServerSrcCount++;
									bAlreadyCounted = true;
								}
								bReaskDownload = true;
							}
						break;

						case DS_REQHASHSET:
							{
								uReqHashSrcCount++;
								break;
							}

						default:
							{
								uUnknownSrcCount++;
							}
					} // switch (state)
#ifdef OLD_SOCKETS_ENABLED
					if (bReaskDownload && g_eMuleApp.m_pServerConnect->IsConnected())
					{
					//	LANCAST (moosetea) - Because reask much faster if lancast user ie we may reask as soon as we get disconnected
					//	, we only use TCP on a LAN, because emule may try to send a TCP request and UDP request at the same time!
						if (pSource->IsOnLAN())
							usethisfilereasktime = LANCASTFILEREASKTIMETCP;

						if (!pSource->GetLastAskedTime() || (dwCurTick - pSource->GetLastAskedTime()) > usethisfilereasktime)
							bIsClientConnected = pSource->AskForDownload();
					}
#endif //OLD_SOCKETS_ENABLED
					pSource = NULL;
				}
			}
		}

	//	-khaos--+++> Set the members using temp vars.  We have to use temporary vars so that the disp counts 
	//	don't get screwed up during processing.  We don't need to count this, cause we can get a transferring
	//	sources over m_downloadingSourceList.GetCount()
		m_iSrcNNP = uNNPSrcCount;
		m_iSrcOnQueue = uOnQueueSrcCount;
		m_iSrcHighQR = uHighQRSrcCount;
		m_iSrcConnecting = uConnectingSrcCount;
		m_iSrcTooManyConns = uTooManyConnsSrcCount;
		m_iSrcConnected = uConnectedSrcCount;
		m_iSrcTimeout = uTimeoutSrcCount;
		m_iSrcConnViaServer = uConnViaServerSrcCount;
		m_iSrcUnknown = uUnknownSrcCount;
		m_iSrcRecHash = uReqHashSrcCount;
		m_iSrcLowToLow = uLowToLowIDSrcCount;
		m_iSrcQueueFull = uQueueFullSrcCount;
		m_iSrcNoUpload = uNoUploadSrcCount;
		m_iSrcBanned = uBannedSrcCount;
		m_iSrcError = uErrorSrcCount;
		m_iSrcA4AF = m_A4AFsrcList.GetCount();

	//	swap a source with NNS status if possible
	//	v:- eklmn: A4AF management
	//	Modified by Tarod -now handles A4AF too keeping usage of CPU less and do not forcing files
	//	that has No Needed Parts to download letting them swap to another file
		bool bIsNNS;

		if (((!m_LastNoNeededCheck) || (dwCurTick - m_LastNoNeededCheck) > 10000))
		{
			m_LastNoNeededCheck = dwCurTick;
			for (int sl = 0;sl < SOURCESSLOTS;sl++)
			{
				if (!m_srcLists[sl].IsEmpty())
				{
					for (pos1 = m_srcLists[sl].GetHeadPosition(); pos1 != NULL;)
					{
						pSource = m_srcLists[sl].GetNext(pos1);
						bIsNNS = false;
					//	check if current source has a part for this file
						if (pSource->GetDownloadState() == DS_NONEEDEDPARTS)
						{
						//	check if A4AF Auto file is enable
							if (g_eMuleApp.m_pDownloadQueue->GetA4AFAutoFile() == NULL)
							//	A4AF auto is disable => switch to any file
								pSource->SwapToAnotherFile(NULL, true);
							else if (g_eMuleApp.m_pDownloadQueue->GetA4AFAutoFile() != this)
							{
							//	A4AF auto is enable & isn't current file => switch to A4AF auto file
								if (!pSource->m_otherNoNeededMap.IsEmpty())
									bIsNNS = pSource->m_otherNoNeededMap.Lookup(g_eMuleApp.m_pDownloadQueue->GetA4AFAutoFile(), dwLastNoNeededCheckTime);

								if (!bIsNNS || (dwCurTick - dwLastNoNeededCheckTime) > FILEREASKTIME)
								{
									pSource->SwapToAnotherFile(g_eMuleApp.m_pDownloadQueue->GetA4AFAutoFile());
								}
							}
						}
					}
				}
			}
		}

	//	if this "File" is A4AF auto & A4AF list isn't empty,
	//	then check A4AF list and swap source to "this", if souce isn't NNS
		if (g_eMuleApp.m_pDownloadQueue->GetA4AFAutoFile() == this && !m_A4AFsrcList.IsEmpty())
		{
		//eklmn: use a two-pointer loop to prevent exception due to swaping (delete)
			for (pos1 = m_A4AFsrcList.GetHeadPosition(); pos1 != NULL;)
			{
				pSource = m_A4AFsrcList.GetNext(pos1);
				bIsNNS = false;
				if (pSource != NULL)
				{
					if (!pSource->m_otherNoNeededMap.IsEmpty())
						bIsNNS = pSource->m_otherNoNeededMap.Lookup(this,dwLastNoNeededCheckTime);
					if (!bIsNNS || (dwCurTick - dwLastNoNeededCheckTime) > FILEREASKTIME)
					{
						pSource->SwapToAnotherFile(this);
					}
				}
			}
		}

	//	check if we want new sources from server
	//	1) connection state will be checked in DL-queue manager, so we don't need to do it here
	//	2) we can call SetRequiredSourcesRefresh() so many time as we want, cause this function
	//	    check existency of the file.
		if ( ( (!m_dwLastFileSourcesRequestTime) || (dwCurTick - m_dwLastFileSourcesRequestTime) > SERVERREASKTIME)
			&& g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileSoft() > GetSourceCount() )
		{
		//	local server
			m_dwLastFileSourcesRequestTime = dwCurTick;
			g_eMuleApp.m_pDownloadQueue->SetRequiredSourcesRefresh(this);
		}
	//	Call SaveSources() only one time (in case there isn't DL)
#ifdef OLD_SOCKETS_ENABLED
		if (dwIteration && g_eMuleApp.m_pGlobPrefs->SLSEnable() && m_downloadingSourceList.IsEmpty() && g_eMuleApp.m_pServerConnect->IsConnected())
		{
			m_sourcesaver.Process(this, g_eMuleApp.m_pGlobPrefs->SLSMaxSourcesPerFile());
		}
#endif //OLD_SOCKETS_ENABLED

	//	calculate data rate, set limit etc. (every 5sec)
		count++;
		if (count == 5)
		{
			count = 0;
			UpdateDisplayedInfo();
			if (m_bPercentUpdated == false)
				UpdateCompletedInfos();
			m_bPercentUpdated = false;
		}
	}

//	For status category filters
	if ( GetTransferringSrcCount() != nOldTransSourceCount
	     || (iOldSrcOnQueueCount == 0 && GetOnQueueSrcCount() != 0)
	     || (iOldSrcOnQueueCount != 0 && GetOnQueueSrcCount() == 0) )
	{
		EnumCategories	curCat = g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.GetCurTabCat();

	//	Changes in file status can affect whether the file belongs to a predefined category or not.
		if (curCat >= CAT_PREDEFINED)
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.ChangeCategoryByID(curCat);
		else
			UpdateDisplayedInfo(true);

	//	If we're displaying file counts in the tabs, update them to reflect status changes.
		if (g_eMuleApp.m_pGlobPrefs->ShowCatTabInfo())
			g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();
	}

	return m_dataRate;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddSources(CMemFile& sources, uint32 serverip, uint16 serverport)
{
	EMULE_TRY

//if this part file is paused or complete/completing no sources needed any more (Cax2 + SyruS check)
	if ( GetStatus() == PS_STOPPED
	     || GetStatus() == PS_COMPLETE
	     || GetStatus() == PS_COMPLETING )
		return ;

	uint8 byteCount;
	uint8 debug_lowiddropped = 0;
	uint8 debug_possiblesources = 0;

	sources.Read(&byteCount, 1);
	//	Added by Tarod - if there is needed place for more sources just remove NNS
	int source_count_plus_new = GetSourceCount() + byteCount;

	if ( source_count_plus_new > g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile()
	     && ((lastpurgetime == 0) || (::GetTickCount() - lastpurgetime) > PURGE_TIME) )
		RemoveNoNeededSources(source_count_plus_new - g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile());

	uint32 dwUserID;
	uint16 nPort;

	for (int i = 0; i < static_cast<int>(byteCount); i++)
	{
		sources.Read(&dwUserID,4);
		sources.Read(&nPort,2);
	//	check first if we are this source

	//	MOD Note: Do not change this part - Merkur
#ifdef OLD_SOCKETS_ENABLED
		if (g_eMuleApp.m_pServerConnect->GetClientID() < 0x1000000 && g_eMuleApp.m_pServerConnect->IsConnected())
		{
			if ((g_eMuleApp.m_pServerConnect->GetClientID() == dwUserID)
				&& inet_addr(g_eMuleApp.m_pServerConnect->GetCurrentServer()->GetFullIP()) == serverip)
				continue;
		}
		else if (g_eMuleApp.m_pServerConnect->GetClientID() == dwUserID)
			continue;
		else if (dwUserID < 0x1000000 && !g_eMuleApp.m_pServerConnect->IsLocalServer(serverip, serverport))
		{
			debug_lowiddropped++;
			continue;
		}
#endif //OLD_SOCKETS_ENABLED
	//	MOD Note - end

		if (g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() > GetSourceCount())
		{
			debug_possiblesources++;
			CUpDownClient* pNewSource = g_eMuleApp.m_pDownloadQueue->CheckAndAddSource(this,dwUserID,nPort,serverip,serverport);
			if (pNewSource)
				pNewSource->SetUserName(GetResString(IDS_SERVER_SOURCE));
		}
	}
//	AddDebugLogLine(false, "RCV: %i sources from server, %i low id dropped, %i possible sources",count,debug_lowiddropped,debug_possiblesources);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	v- eklmn: changes(00): NewSrcPartsInfo & UpdateAvailablePartsCount merged (CPU load)
void CPartFile::NewSrcPartsInfo()
{
	EMULE_TRY
//	Cache part count
	const int iPartCount = GetPartCount();
//	Calculate a median position
	const int idxMedian = (int)ceil(iPartCount / 2.0) - 1;	// <- eklmn: feature(00): ICR
	uint8 uAvailablePartsCounter = 0;
	uint8 uMaxAvailableParts = 0;

//	Increase size if necessary
	if (m_srcPartFrequencies.GetSize() < iPartCount)
	{
		m_srcPartFrequencies.SetSize(iPartCount);
	}

//	Reset part counters & median
	for (int i = 0; i < iPartCount; i++)
	{
		m_srcPartFrequencies[i] = 0;
	}
	m_uMedian = 0;     // <- eklmn: feature(00): ICR
	CUpDownClient* pSource;

	POSITION pos;
	for (int sl = 0; sl < SOURCESSLOTS; ++sl)
	{
		if (!m_srcLists[sl].IsEmpty())
		{
			for (pos = m_srcLists[sl].GetHeadPosition();pos != NULL;)
			{
				pSource = m_srcLists[sl].GetNext(pos);
				if (pSource)
				{
					uAvailablePartsCounter = 0;
					int j = 0; // <- eklmn: feature(00): ICR v2
					for (int i = 0; i < iPartCount; i++)
					{
						if (pSource->IsPartAvailable(i))
						{
							if (j == idxMedian)
								m_uMedian++;       // <- eklmn: feature(00): ICR
							j++;                            // <- eklmn: feature(00): ICR
							m_srcPartFrequencies[i] += 1;
							uAvailablePartsCounter++;
						} // end: IsPartAv...
					}
					if (uAvailablePartsCounter > uMaxAvailableParts)
						uMaxAvailableParts = uAvailablePartsCounter;
				}
			}
		}
	}

	m_MaxSrcPartFrequency = 0;	// <- eklmn: feature(00): ICR
	for (int i = 0; i < iPartCount; i++)
		if (m_srcPartFrequencies[i] > m_MaxSrcPartFrequency)
			m_MaxSrcPartFrequency = m_srcPartFrequencies[i];

	if (iPartCount == uMaxAvailableParts && availablePartsCount < iPartCount)
		lastseencomplete = CTime::GetCurrentTime();
	availablePartsCount = uMaxAvailableParts;
	m_bUpdateCompleteSrcCount = true;	// netwolf: zegzav:completesrc

	UpdateDisplayedInfo();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	v- eklmn: feature(00): Intelligent Chunk Request(ICR v2.11, 26.09.2003) for eMule+
uint16 CPartFile::GetNextRequestedBlock(CUpDownClient *pSource)
{
#ifdef OLD_SOCKETS_ENABLED
	EMULE_TRY

	if (pSource == NULL)
		return 0xFFFF;

//	Cache some values => speed improvement
	const int		iPartCount = GetPartCount();
	int16			nNumAvailableNeededParts = 0;
	int16			nNumCurBlockedParts = 0;
	const uint8		bytePreviewMode = GetMovieMode();
//	Mode 0=NoPreview,1=AVI(1F,1L),2=AVI(2F,1L),3=AVI(3F,1L)
//	Mode 4=MPG(1F),5=MPG(2F),6=MPG(3F),7=AVI/MPG(2F,2L) F=First,L=Last
	uint8		   *pbytePartStatuses = pSource->GetPartStatus();
	uint16			nGoodPart = pSource->GetLastDownPartAsked();

	_Bvector		gettingParts;

//	Find out which parts the source currently has pending requests for
	pSource->ShowDownloadingParts(gettingParts);
	if ((nGoodPart != 0xFFFF) && GetNextEmptyBlockInPart(nGoodPart, NULL))
	{
	//AddDebugLogLine(false, "Priority 0 (always choose currently downloaded part until it will be finished). File: %s, Part : %u", this->GetFileName(), nGoodPart);
		return pSource->GetLastDownPartAsked();
	}
	else
	{
		pSource->SetLastDownPartAsked(0xFFFF);
	}

//	Construct Lists of block sorted by priority. Use number of parts to create limited lists and
//	keep data in list in same memory area (speed improvements)
//	CList<int>		rarestReqPartsList(iPartCount);	// priority 1 (highest)
	CList<int>		uniquePartsList(iPartCount);	// priority 2
	CList<int>		rarestPartsList(iPartCount);	// priority 4
	CList<int>		moviePartsList(5);				// priority 5 => only first+last parts
//	CList<int>		requestedPartsList(iPartCount);	// priority 6
	CList<int>		randomPartsList(iPartCount);	// priority 7
//----------------------------------------------------
	CList<uint16, uint16>	curBlockedPartsList;
	uint32			dwLastPartSize = 0;
//
//	Scan for requested chunks in order to find how much they filled.
	for (int i = 0; i < iPartCount ; i++)
	{
		if (pSource->IsPartAvailable(i) == true)
		{
			m_ReqParts[i] = GetBytesInPart(i, true);
			if (m_ReqParts[i] != 0)
				nNumAvailableNeededParts++;
		}
		else
		{
			m_ReqParts[i] = 0xFFFFFFFF; // mark parts as not available to save CPU load
		}
	//	Fill the m_BlockedParts array with NULL
		m_BlockedParts[i] = NULL;
	}
//
//	Scan DownloadList for Blocked parts
	CUpDownClient*		pDownloadingSource;

	for (POSITION pos = m_downloadingSourceList.GetHeadPosition(); pos != NULL;)
	{
		pDownloadingSource = m_downloadingSourceList.GetNext(pos);
		if (pDownloadingSource && pDownloadingSource->GetLastDownPartAsked() < iPartCount)
		{
			m_BlockedParts[pDownloadingSource->GetLastDownPartAsked()] = pDownloadingSource;
			nNumCurBlockedParts++;
		}
	}

//	Define condition of rareness
	uint16		uRarestPart = static_cast<uint16>(floor(0.1 * m_uMedian)); // 10% from median

	if (uRarestPart < 3)
		uRarestPart = 3;

//	Build lists
	uint16		uMostCompletedUniqueReqPart;
	uint16		uMostCompletedRarestReqPart;
	uint16		uMostCompletedReqPart;
	uint32		dwMostCompletedUniqueReqPart = PARTSIZE;
	uint32		dwMostCompletedRarestReqPart = PARTSIZE;
	uint32		dwMostCompletedReqPart = PARTSIZE;

//	For each part in the part file...
	for (int i = 0; i < iPartCount ; i++)
	{
	//	If the part is available and it hasn't been completed...
		if (m_ReqParts[i] != 0xFFFFFFFF && m_ReqParts[i] != 0)
		{
		//	If the part is only available from one source, try to download
		//		it first in order to prevent file disappearing)
			if (m_srcPartFrequencies[i] == 1)
			{
			//	If the part is a full sized part...
				if (m_ReqParts[i] > (PARTSIZE - EMBLOCKSIZE)) // limit change to prevent consecutive download due overlaping
				{
					uniquePartsList.AddHead(i);
				}
				else
				{
				//	If we've completed more of this part (or rather require less to complete it)
				//		than any other one-source part seen so far...
					if (m_ReqParts[i] < dwMostCompletedUniqueReqPart)
					{
						uMostCompletedUniqueReqPart = i;
						dwMostCompletedUniqueReqPart = m_ReqParts[i];
					}
				}
			}
		//	------------------------ Rarest Parts -------------------------
			if (m_uMedian >= (uRarestPart + m_srcPartFrequencies[i]))
			{
				if (m_ReqParts[i] == PARTSIZE)
				{
					for (int j = 0; j < (m_uMedian - m_srcPartFrequencies[i] + 1); j++)
						rarestPartsList.AddHead(i);
				}
				else
				{
					if (m_ReqParts[i] < dwMostCompletedRarestReqPart)
					{
						uMostCompletedRarestReqPart = i;
						dwMostCompletedRarestReqPart = m_ReqParts[i];
					}
				}
			} // end rarest
		//	--------------------------- Movie or Archive ---------------------------
			switch (bytePreviewMode)
			{
				case 3:	//	3 = AVI(3F,1L)
				{
					if (iPartCount > 3 && (i == iPartCount - 1))
						moviePartsList.AddHead(i);
				}
				case 6:	//	6 = MPG(3F)
				{
					if (i == 2 && iPartCount > 2)
						moviePartsList.AddHead(2);
					if (i == 1 && iPartCount > 1)
						moviePartsList.AddHead(1);
					if (i == 0)
						moviePartsList.AddHead(0);
					break;
				}
				case 7:	//	7 = AVI/MPG(2F,2L)
				{
					if (iPartCount > 3 && (i == iPartCount - 2))
						moviePartsList.AddHead(i);
				}
				case 2:	//	2 = AVI(2F,1L)
				{
					if (iPartCount > 2 && (i == iPartCount - 1))
						moviePartsList.AddHead(i);
				}
				case 5:	//	5 = MPG(2F)
				{
					if (i == 1 && iPartCount > 1)
						moviePartsList.AddHead(1);
					if (i == 0)
						moviePartsList.AddHead(0);
					break;
				}
				case 1:	//	1 = AVI(1F,1L)
				{
					if (iPartCount > 1 && (i == iPartCount - 1))
						moviePartsList.AddHead(i);
				}
				case 4:	//	4 = MPG(1F)
				{
					if (i == 0)
						moviePartsList.AddHead(0);
				}
			}
		//	--------------------- widespreaded parts ---------------------
		//	Check blocked parts
			if (m_BlockedParts[i] != NULL && nNumAvailableNeededParts > nNumCurBlockedParts)
			{
				continue;
			}
			if (m_ReqParts[i] == PARTSIZE)
			{
			//	Unstarted
				for (int j = 0; j < (m_MaxSrcPartFrequency - m_srcPartFrequencies[i] + 1); j++)
					randomPartsList.AddHead(i);
			}
			else
			{
				if (m_ReqParts[i] < dwMostCompletedReqPart)
				{
					uMostCompletedReqPart	= i;
					dwMostCompletedReqPart = m_ReqParts[i];
				}
			} // end widespreaded
		} // end: if....
	} // end for

//	Select final result
//	------------------------------ Unique & Requested --------------------------------------------------
//	Try to complete a UniquePart and MostDownloaded as soon as possible to become a new source
//	it doesn't matter which one download, if there are parts same filled.
	if (dwMostCompletedUniqueReqPart != PARTSIZE && moviePartsList.IsEmpty())
	{
		nGoodPart = uMostCompletedUniqueReqPart;
	//	AddDebugLogLine(false, "Priority 1 (Unique unfinished part). File: %s, Part : %u", this->GetFileName(), nGoodPart);
	}
//	------------------------------ Unique & Unrequested ------------------------------------------------
	else if (!uniquePartsList.IsEmpty() && moviePartsList.IsEmpty())
	{
	//	If there is only one unique part, then take it, otherwise take it random.
		if (uniquePartsList.GetCount() == 1)
		{
			nGoodPart = uniquePartsList.GetHead();
		}
		else
		{
			const uint16		uRandomness = static_cast<uint16>(ROUND((static_cast<double>(rand()) / RAND_MAX) * (uniquePartsList.GetCount() - 1)));

			nGoodPart = uniquePartsList.GetAt(uniquePartsList.FindIndex(uRandomness));
		}
	//	AddDebugLogLine(false, "Priority 2 (Unique unrequested part). File: %s, Part : %u", this->GetFileName(), nGoodPart);
	}
//	------------------------------ Rarest & Requested --------------------------------------------------
//	Try to complete a RarestPart and MostDownloaded as soon as possible to become a new source
//	it doesn't matter which one download, if there are parts same filled.
	else if (dwMostCompletedRarestReqPart != PARTSIZE)
	{
		nGoodPart = uMostCompletedRarestReqPart;
	//	AddDebugLogLine(false, "Priority 3 (Rarest unfinished part). File: %s, Part : %u", this->GetFileName(), nGoodPart);
	}
//	------------------------------ Rarest & Unrequested ------------------------------------------------
	else if (rarestPartsList.IsEmpty() == false)
	{
	//	If there is only one rarest part, then take it, otherwise take it random.
		if (rarestPartsList.GetCount() == 1)
		{
			nGoodPart = rarestPartsList.GetHead();
		}
		else
		{
			const uint16		uRandomness = static_cast<uint16>(ROUND((static_cast<double>(rand()) / RAND_MAX) * (rarestPartsList.GetCount() - 1)));

			nGoodPart = rarestPartsList.GetAt(rarestPartsList.FindIndex(uRandomness));
		}
	//	AddDebugLogLine(false, "Priority 4 (Rarest unrequested part). File: %s, Part : %u", this->GetFileName(), nGoodPart);
	}
//	---------------------------------- Movies Part  -----------------------------------------------------
	else if (moviePartsList.IsEmpty() == false)
	{
		nGoodPart = moviePartsList.GetHead();
	//	AddDebugLogLine(false, "Priority 5 (Preview part). File: %s, Part : %u", this->GetFileName(), nGoodPart);
	}
//	--------------------------------------- Requested ----------------------------------------------------
//	Try to complete an UnfinishedPart and MostDownloaded as soon as possible to become a new source
//	it doesn't matter which one download, if there are parts same filled.
	else if (dwMostCompletedReqPart != PARTSIZE)
	{
		nGoodPart = uMostCompletedReqPart;
	//	m_BlockedParts[nGoodPart] = pSource;
	//	AddDebugLogLine(false, "Priority 6 (Widespread unfinished part). File: %s, Part : %u", this->GetFileName(), nGoodPart);
	}
//	-------------------------------- Unrequested & Random -------------------------------------------------
	else if (randomPartsList.IsEmpty() == false)
	{
		const uint16		uRandomness = static_cast<uint16>(ROUND((static_cast<double>(rand()) / RAND_MAX) * (randomPartsList.GetCount() - 1)));

		nGoodPart = randomPartsList.GetAt(randomPartsList.FindIndex(uRandomness));
	//	m_BlockedParts[nGoodPart] = pSource;
	//	AddDebugLogLine(false, "Priority 7 (Widespread unrequested part). File: %s, Part : %u", this->GetFileName(), nGoodPart);
	}
	else
	{
		nGoodPart = 0xFFFF;
	}

//	AddDebugLogLine(false, "ICR-Test: File: %s, A-parts : %u, B-Parts %u. Choosen Part: %u", this->GetFileName(), nNumAvailableNeededParts, curBlockedPartsList.GetCount(),nGoodPart);

//	Clear lists
	if (!uniquePartsList.IsEmpty())
	{
		rarestPartsList.RemoveAll();
	}
	if (!rarestPartsList.IsEmpty())
	{
		rarestPartsList.RemoveAll();
	}
	if (!moviePartsList.IsEmpty())
	{
		moviePartsList.RemoveAll();
	}
	if (!randomPartsList.IsEmpty())
	{
		randomPartsList.RemoveAll();
	}
	if (!curBlockedPartsList.IsEmpty())
	{
		curBlockedPartsList.RemoveAll();
	}

	pSource->SetLastDownPartAsked(nGoodPart);
	return nGoodPart;

	EMULE_CATCH
#endif //OLD_SOCKETS_ENABLED
	return 0xFFFF;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CPartFile::GetBytesInPart(uint16 partnumber, bool left2DL, bool dontRecalculate)
{
	EMULE_TRY

	if (partnumber >= GetPartCount())
		return (left2DL ? 0 : 0xFFFFFFFF);

	uint32 PartSize;
	uint32 left2DLSize;

	const uint32 uStart = partnumber * PARTSIZE;
	const uint32 uEnd = ((m_lFileSize - 1) < (uStart + PARTSIZE - 1)) ? (m_lFileSize - 1) : (uStart + PARTSIZE - 1);

	if (dontRecalculate)
	{
		PartSize = m_ReqParts[partnumber];
		left2DLSize = (uEnd - uStart + 1) - PartSize;
	}
	else
	{
	//	Set part size to fullfilled
		PartSize = left2DLSize = uEnd - uStart + 1;
		for (POSITION pos = gaplist.GetHeadPosition(); pos != 0;)
		{
			Gap_Struct * cur_gap = gaplist.GetNext(pos);
		//	Check if Gap is into the limit
			if (cur_gap->m_dwEndOffset < uStart || cur_gap->m_dwStartOffset > uEnd)
				continue;
			if (cur_gap->m_dwStartOffset <= uStart && cur_gap->m_dwEndOffset > uStart && cur_gap->m_dwEndOffset < uEnd)
			{
				PartSize -= cur_gap->m_dwEndOffset - uStart;
				continue;
			}
			if (cur_gap->m_dwStartOffset >= uStart && cur_gap->m_dwEndOffset <= uEnd)
			{
				PartSize -= cur_gap->m_dwEndOffset - cur_gap->m_dwStartOffset;
				continue;
			}
			if (cur_gap->m_dwStartOffset <= uStart && cur_gap->m_dwEndOffset >= uEnd)
			{
				PartSize = 0;
				continue;
			}
			if (cur_gap->m_dwStartOffset > uStart && cur_gap->m_dwStartOffset < uEnd && cur_gap->m_dwEndOffset >= uEnd)
			{
				PartSize -= uEnd - cur_gap->m_dwStartOffset;
			}
		}
	}

	left2DLSize -= PartSize;

	if (left2DL)
		return left2DLSize;
	else
		return PartSize;

	EMULE_CATCH

	return (left2DL ? 0 : 0xFFFFFFFF);

}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemoveBlockFromList(uint32 start, uint32 dwEnd)
{
	EMULE_TRY

	POSITION pos1, pos2;
	for (pos1 = requestedblocks_list.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		if (requestedblocks_list.GetNext(pos1)->m_dwStartOffset <= start && requestedblocks_list.GetAt(pos2)->m_dwEndOffset >= dwEnd)
			requestedblocks_list.RemoveAt(pos2);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemoveAllRequestedBlocks(void)
{
	EMULE_TRY

	requestedblocks_list.RemoveAll();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::CompleteFile(bool bIsHashingDone)
{
	EMULE_TRY

	StopFile();

//Show completing instead of paused
	SetStatus(PS_COMPLETING);

	if (g_eMuleApp.m_pGlobPrefs->ShowCatTabInfo())
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	UpdateDisplayedInfo(true);

//	If the part file isn't already hashed...
	if (!bIsHashingDone)
	{
	//	Create a thread to hash the part file in the background
		m_dataRate = 0;

		CString	partFileNameBase = m_strPartMetFileName.Left(m_strPartMetFileName.GetLength() - 4);
		CHashFileThread	*addFileThread = NULL;

		addFileThread = (CHashFileThread*)AfxBeginThread( RUNTIME_CLASS(CHashFileThread),
		                THREAD_PRIORITY_BELOW_NORMAL + g_eMuleApp.m_pGlobPrefs->GetMainProcessPriority(),
		                0,
		                CREATE_SUSPENDED );
		addFileThread->SetValues(0, m_strTempDir, partFileNameBase, this);
		addFileThread->ResumeThread();
		return ;
	}
	else
	{
		CWinThread *pThread = AfxBeginThread((AFX_THREADPROC)CompleteThreadProc, this, THREAD_PRIORITY_BELOW_NORMAL + g_eMuleApp.m_pGlobPrefs->GetMainProcessPriority(), 0, CREATE_SUSPENDED);

		if (pThread)
		{
			pThread->ResumeThread();
		}
		else
		{
			throw CString(_T("Error creating file complete thread"));
		}
	}
	GetAvgDataRate(true);

	if (g_eMuleApp.m_pGlobPrefs->ShowCatTabInfo())
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	UpdateDisplayedInfo(true);
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void GenerateNewFileName(const CString& outputDir, CString& newname)
{
	ASSERT(!newname.IsEmpty());

	int namecount = 0;

	TCHAR p_fname[_MAX_FNAME];
	TCHAR p_fext[_MAX_EXT];
	_tsplitpath(newname, NULL, NULL, p_fname, p_fext);

	CString fname(p_fname);
	CString fext(p_fext);
//search for matching ()s and check if it contains a number inside
	if (fname.Right(1) == ')')
	{
		int ll = fname.GetLength();
		int ob = fname.ReverseFind( '(' );
		if (ob != -1 && ob + 2 != ll)
		{ //check for ) existence or () situation
			bool found = true;
			for (int i = ob + 1; i < ll - 1; i++)
			{	//check for digits inside brackets
				if (!_istdigit(fname[i]))
				{
					found = false;
					break;
				}
			}
			if (found)
			{
				namecount = _ttoi(fname.Mid(ob + 1, ll - ob - 2));
				fname.Truncate(ob);
			}
		}
	}

	do
	{
		namecount++;
		newname.Format(_T("%s\\%s(%d)%s"), outputDir, fname, namecount, fext);
	}
	while (PathFileExists(newname));
}

//	Lord KiRon - using threads for file completion
BOOL CPartFile::PerformFileComplete(void)
{
	EMULE_TRY

//	Wait until currently downloading remote client finish
//	WaitForSingleObject(m_hCompletionEvent, INFINITE);

//	Get current category
	EnumCategories eCatID = GetCatID();

	CSingleLock sLock(&m_FileCompleteMutex, TRUE); // will be unlocked on exit
	CString partfilename = m_strFullName.Left(m_strFullName.GetLength() - 4);

	CString newname;
	newname.Format(_T("%s\\%s"), GetOutputDir(), g_eMuleApp.StripInvalidFilenameChars(GetFileName()));

	if (!PathFileExists(GetOutputDir()))
		::CreateDirectory(GetOutputDir(), 0);

//	close permanent handle
	try
	{
		if (m_hPartFile.m_hFile != INVALID_HANDLE_VALUE)
			m_hPartFile.Close();
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
		AddDebugLogLine(false, _T("Failed to close permanent handle %s - %s"), GetFileName(), GetErrorMessage(error));
		error->Delete();
	}

	bool renamed = false;
	if (PathFileExists(newname))
	{
		renamed = true;
		GenerateNewFileName(GetOutputDir(), newname);
	}

	BOOL bSlowComplete = g_eMuleApp.m_pGlobPrefs->SlowComplete();
	int BlockSize = g_eMuleApp.m_pGlobPrefs->SlowCompleteBlockSize() * 1024;

//	we do this only if SlowComplete enabled
//	AND the files are on the different HDs (paratitions) - check by the drive leter
//	AND any of path is not in URL/Network format (has : as second leter)
	if ( bSlowComplete && ( _toupper(partfilename[0]) != _toupper(newname[0]) && partfilename[1] == _T(':')
	                        && newname[1] == _T(':') ) )
	{
		CFile fInputFile;
		CFile fOutputFile;
		try
		{
			char *buffer = new char[BlockSize];
			DWORD dwRead = 0;
			CString strTempNew = CString(newname);
			strTempNew += _T(".$$$");

		//	We going to write everithing into the temporary file at same place the newname located
		//	and after we finished - rename , this will take care of syncronization problem with SharedFiles
			if ( !fInputFile.Open( partfilename, CFile::modeRead | CFile::shareDenyWrite | CFile::typeBinary
			                       | CFile::osSequentialScan ) )
			{
			//	let catch to care about closing everything
				throw CString(GetResString(IDS_ERR_SLOWREAD) + _T(" (") + partfilename + _T(")"));
			}
			if ( !fOutputFile.Open( strTempNew, CFile::modeWrite | CFile::shareDenyWrite | CFile::osSequentialScan
								   | CFile::modeCreate ) )
			{
				fInputFile.Close();
			//	let catch to care about closing everything
				throw CString(GetResString(IDS_ERR_SLOWWRITE) + _T(" (") + strTempNew + _T(")"));
			}

		//	Most of the work done here
		//	just copy the file data from one to another
			do
			{
				dwRead = fInputFile.Read(buffer, BlockSize);
				fOutputFile.Write(buffer, dwRead);
				Sleep(0); // release the CPU for other processes
			}
			while (dwRead > 0);
			delete []buffer;

			fInputFile.Close();
			fOutputFile.Close();

		//	Rename temporary file back to normal name
			if (_trename(strTempNew, newname))
			{
			//	let catch to care about closing everything
				throw CString(GetResString(IDS_ERR_RENFAIL));
			}
			CFile::Remove(partfilename);
		}
		catch (CMemoryException * error)	//check for exception that will be thow by "new"
		{
			error->Delete();
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_OUTMEM));
			m_bPaused = true;
			SetStatus(PS_ERROR);
			return FALSE;
		}
		catch (CString str)
		{
			AddLogLine(true, RGB_LOG_ERROR + str);
			m_bPaused = true;
			SetStatus(PS_ERROR);
			return FALSE;
		}
		catch (...)
		{
		//	Purity - Sending message for erroneous file
			CString MessageText;
			MessageText.Format(GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName());
			g_eMuleApp.m_pSMTPConnection->SendMail(MessageText, g_eMuleApp.m_pGlobPrefs->GetNotifierPopOnImportantError(), g_eMuleApp.m_pGlobPrefs->IsSMTPWarningEnabled());
			g_eMuleApp.m_pdlgEmule->ShowNotifier(MessageText, TBN_IMPORTANTEVENT, false, g_eMuleApp.m_pGlobPrefs->GetNotifierPopOnImportantError());
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName());
			m_bPaused = true;
			SetStatus(PS_ERROR);
			return FALSE;
		}
	}
	else
	{
		if (_trename(partfilename, newname))
		{
		//	Purity - Sending message for erroneous file
			CString MessageText;
			MessageText.Format(GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName());
			g_eMuleApp.m_pSMTPConnection->SendMail(MessageText, g_eMuleApp.m_pGlobPrefs->GetNotifierPopOnImportantError(), g_eMuleApp.m_pGlobPrefs->IsSMTPWarningEnabled());
			g_eMuleApp.m_pdlgEmule->ShowNotifier(MessageText, TBN_IMPORTANTEVENT, false, g_eMuleApp.m_pGlobPrefs->GetNotifierPopOnImportantError());
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_COMPLETIONFAILED), GetFileName(), strerror(errno));
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_COMPLETIONFAILED2));
			m_bPaused = true;
			SetStatus(PS_ERROR);
			return FALSE;
		}
	}

//	bluecow: to have the accurate date stored in known.met we have to update the 'date' of a just completed file.
//	if we don't update the file date here (after commiting the file and before adding the record to known.met),
//	that file will be rehashed at next startup and there would also be a duplicate entry (hash+size) in known.met because of different file date!
	ASSERT(m_hPartFile.m_hFile == INVALID_HANDLE_VALUE); // the file must be closed/commited!
	struct _stat st;
	if (_tstat(newname, &st) == 0)
		m_timetLastWriteDate = st.st_mtime;

	if (_tremove(m_strFullName))
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), m_strFullName);

//	enkeyDEV(Ottavio84) -ChangeDir-
	CString dirFile = CString(RemoveFileExtension(m_strFullName)) + CString(".dir");
	if (_tremove(dirFile) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), dirFile);

//	Tarod - erase .bak file
	CString bakFile = CString(m_strFullName) + CString(".bak");
	if (_tremove(bakFile) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), bakFile);

//	DonGato - remove .settings file
	CString strSettingsFile = CString(RemoveFileExtension(m_strFullName)) + CString(".settings");
	if (_tremove(strSettingsFile) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strSettingsFile);

//	netwolf - remove .txtscr file
	m_sourcesaver.DeleteFile(this);

//--- xrmb:keepPartFileStats ---
//--- remove the stats file ---
	RemovePartFileStats();
//--- :xrmb ---

//	Change AutoPriority to NORMAL when UAP is disabled
	if (CKnownFile::IsAutoPrioritized() && !g_eMuleApp.m_pGlobPrefs->IsUAPEnabled())
	{
		CKnownFile::SetAutoPriority(false);
		CKnownFile::SetPriority(PR_NORMAL);
	}

	m_strFullName = newname;
	directory = GetOutputDir();
	SetFilePath(m_strFullName);
	SetStatus(PS_COMPLETE);
	m_bPaused = false;

	UpdateDisplayedInfo(true);
	AddLogLine(true, RGB_LOG_SUCCESS + GetResString(IDS_DOWNLOADDONE), GetFileName());

//	Purity - Sending message for completed file
	CString MessageText;
	MessageText.Format(GetResString(IDS_TBN_DOWNLOADDONE) + _T(" ") + GetFileName());
	g_eMuleApp.m_pSMTPConnection->SendMail(MessageText, g_eMuleApp.m_pGlobPrefs->GetUseDownloadNotifier(), g_eMuleApp.m_pGlobPrefs->IsSMTPInfoEnabled());
	g_eMuleApp.m_pdlgEmule->ShowNotifier(MessageText, TBN_DLOAD, false, g_eMuleApp.m_pGlobPrefs->GetUseDownloadNotifier());
	if (renamed)
		AddLogLine(true, IDS_DOWNLOADRENAMED, _tcsrchr(newname, '\\') + 1);
	g_eMuleApp.m_pKnownFilesList->SafeAddKnownFile(this);
	g_eMuleApp.m_pDownloadQueue->RemoveFile(this);
	UpdateDisplayedInfo();

#ifdef OLD_SOCKETS_ENABLED
#ifdef MOBILE_MULE
	g_eMuleApp.m_pMMServer->AddFinishedFile(this);
#endif //MOBILE_MULE
#endif //OLD_SOCKETS_ENABLED

//	Added by Tarod (starts next paused file)
	if (g_eMuleApp.m_pGlobPrefs->DownloadPausedOnComplete())
	{
		g_eMuleApp.m_pDownloadQueue->StartNextFile(eCatID);
	}

//	-khaos--+++> Extended Statistics Added 2-10-03
	g_eMuleApp.m_pGlobPrefs->Add2DownCompletedFiles();		// Increments cumDownCompletedFiles in prefs struct
	g_eMuleApp.m_pGlobPrefs->Add2DownSessionCompletedFiles(); // Increments sesDownCompletedFiles in prefs struct
	g_eMuleApp.m_pGlobPrefs->SaveCompletedDownloadsStat();	// Saves cumDownCompletedFiles to INI

	if (directory != g_eMuleApp.m_pGlobPrefs->GetIncomingDir())
	{
		for (POSITION pos = g_eMuleApp.m_pGlobPrefs->m_sharedDirList.GetHeadPosition();pos != 0;)
			if (g_eMuleApp.m_pGlobPrefs->m_sharedDirList.GetNext(pos) == (directory + "\\"))
				return TRUE;
		g_eMuleApp.m_pGlobPrefs->m_sharedDirList.AddTail(directory + "\\");
		g_eMuleApp.m_pSharedFilesList->Reload();
	}

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	return TRUE;

	EMULE_CATCH

	return FALSE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemoveAllSources(bool bTryToSwap)
{
	EMULE_TRY

//	TODO transfer sources to other downloading files if possible
	CUpDownClient* pRemoveSource;

	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		while (!m_srcLists[sl].IsEmpty())
		{
			pRemoveSource = m_srcLists[sl].GetHead();
			if (bTryToSwap)
			{
				if (!pRemoveSource->SwapToAnotherFile(NULL))
					g_eMuleApp.m_pDownloadQueue->RemoveSource(pRemoveSource);
				else
				//eklmn: if switching was succesfull we need to remove the entry from graphical list
					g_eMuleApp.m_pDownloadList->RemoveSource(pRemoveSource, this);
			}
			else
				g_eMuleApp.m_pDownloadQueue->RemoveSource(pRemoveSource);
		}

//eklmn: remove A4AF entry from graphical list & anotherRequestList
	if (!m_A4AFsrcList.IsEmpty())
	{
		for (POSITION pos = m_A4AFsrcList.GetHeadPosition(); pos != NULL;)
		{
			pRemoveSource = m_A4AFsrcList.GetNext(pos);
			g_eMuleApp.m_pDownloadList->RemoveSource(pRemoveSource, this);
			pRemoveSource->RemoveRequestForAnotherFile(this);
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::DeleteFile()
{
	EMULE_TRY

	ASSERT (!m_bPreviewing);

	m_bIsBeingDeleted = true;

	StopFile();

//	m_hPartFile isn't the handle to the file...
	if (m_hPartFile.m_hFile != INVALID_HANDLE_VALUE)
		m_hPartFile.Close();

	if (_tremove(m_strFullName))
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), m_strFullName);

	CString partfilename = m_strFullName.Left(m_strFullName.GetLength() - 4);
	if (_tremove(partfilename))
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), partfilename);

//	enkeyDEV(Ottavio84) -ChangeDir-
	CString dirFile = CString(RemoveFileExtension(m_strFullName)) + CString(".dir");
	if (_tremove(dirFile) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), dirFile);

//	Tarod - erase .bak file
	CString bakFile = CString(m_strFullName) + CString(".bak");
	if (_tremove(bakFile) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), bakFile);

//	DonGato - remove .settings file
	CString strSettingsFile = CString(RemoveFileExtension(m_strFullName)) + CString(".settings");
	if (_tremove(strSettingsFile) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), strSettingsFile);

//	netwolf - remove .txtscr file
	m_sourcesaver.DeleteFile(this);

//--- xrmb:keepPartFileStats ---
//--- remove the stats file ---
	RemovePartFileStats();
//--- :xrmb ---

//Cax2 - file properties get deleted inside one of these calls now...
	g_eMuleApp.m_pSharedFilesList->RemoveFile(this);
	g_eMuleApp.m_pDownloadQueue->RemoveFile(this);
	g_eMuleApp.m_pDownloadList->RemoveFile(this);

	delete this;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::HashSinglePart(uint16 partnumber)
{
	EMULE_TRY

	if ((GetHashCount() <= partnumber) && (GetPartCount() > 1))
	{
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_HASHERRORWARNING), GetFileName());
		this->m_bHashSetNeeded = true;
		return true;
	}
	else if (!GetPartHash(partnumber) && GetPartCount() != 1)
	{
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_INCOMPLETEHASH), GetFileName());
		this->m_bHashSetNeeded = true;
		return true;
	}
	else
	{
		uchar hashresult[16];

		m_hPartFile.Seek((LONGLONG)PARTSIZE * partnumber, 0);

		uint32 length = PARTSIZE;

		if ((ULONGLONG)PARTSIZE * (partnumber + 1) > m_hPartFile.GetLength())
		{
			length = (m_hPartFile.GetLength() - ((ULONGLONG)PARTSIZE * partnumber));
		}
		ASSERT(length <= PARTSIZE);
		CreateHashFromFile(&m_hPartFile, length, hashresult);

		if (GetPartCount() > 1 || GetFileSize() == PARTSIZE)
		{
			if (md4cmp(hashresult, GetPartHash(partnumber)))
				return false;
			else
				return true;
		}
		else
		{
			if (md4cmp(hashresult, m_fileHash))
				return false;
			else
				return true;
		}
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsCorruptedPart(uint16 uPartIndex)
{
	EMULE_TRY

	return corrupted_list.Find(uPartIndex);

	EMULE_CATCH

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsMovie()
{
	CString fname = GetFileName();

	fname.MakeLower();

	bool bIsVCD = _tcsstr(fname, _T(".bin"));

	return (ED2KFT_VIDEO == GetED2KFileTypeID(GetFileName()) || bIsVCD);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsAviMovie()
{
	EMULE_TRY

	CString fname = GetFileName();
	fname.MakeLower();

	return (_tcsstr(fname, _T(".avi")) || _tcsstr(fname, _T(".ogm")) || _tcsstr(fname, _T(".divx")));

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsMpgMovie()
{
	EMULE_TRY

	CString fname = GetFileName();
	fname.MakeLower();

	return (_tcsstr(fname, _T(".mpg")) || _tcsstr(fname, _T(".mpe")) || _tcsstr(fname, _T(".bin")));

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsMovieChunk(uint8 mode, uint16 chunk)
{
	EMULE_TRY

	uint32 lastblock = ceil(static_cast<double>(PARTSIZE) / EMBLOCKSIZE), dwEnd;
	if (chunk == GetPartCount() - 2)
		dwEnd = (PARTSIZE * chunk) + (lastblock * EMBLOCKSIZE) - 1; //Endadress from 1chunk before the last chunk

	switch (mode)
	{
		case 0:  //Mode 0: No Preview
		return false;
		break;
		case 1:  //Mode 1: 1F,1L(2L) AVI
		if (chunk == 0)
			return true; //1F
		if (chunk == GetPartCount() - 1)
			return true; //1L is a moviechunk ?
		if (chunk == (GetPartCount() - 2) && (((GetFileSize() - 1) - MINLASTMOVIESIZE) <= dwEnd))
			return true; //2L is a Part of 1/8L at the end of the file
		break;
		case 2:  //Mode 2: 1F,2F,1L(2L) Long AVI
		if (chunk == 0)
			return true; //1F
		if (chunk == 1)
			return true; //2F
		if (chunk == GetPartCount() - 1)
			return true; //1L is a moviechunk ?
		if (chunk == (GetPartCount() - 2) && (((GetFileSize() - 1) - MINLASTMOVIESIZE) <= dwEnd))
			return true; //2L is a Part of 1/8L at the end of the file
		break;
		case 3:  //Mode 3: 1F,2F,3F,1L(2L) Longest AVI
		if (chunk == 0)
			return true; //1F
		if (chunk == 1)
			return true; //2F
		if (chunk == 2)
			return true; //3F
		if (chunk == GetPartCount() - 1)
			return true; //1L is a moviechunk
		if (chunk == (GetPartCount() - 2) && (((GetFileSize() - 1) - MINLASTMOVIESIZE) <= dwEnd))
			return true; //2L is a Part of 1/8L at the end of the file
		break;
		case 4:  //Mode 4: 1F MPG
		if (chunk == 0)
			return true; //1F
		break;
		case 5:  //Mode 5: 1F,2F Long MPG
		if (chunk == 0)
			return true; //1F
		if (chunk == 1)
			return true; //2F
		break;
		case 6:  //Mode 6: 1F,2F,3F Longest MPG
		if (chunk == 0)
			return true; //1F
		if (chunk == 1)
			return true; //2F
		if (chunk == 2)
			return true; //3F
		break;
		case 7:  //Mode 7: 1F,2F,1L,2L AVI/MPG normal
		if (chunk == 0)
			return true; //1F
		if (chunk == 1)
			return true; //2F
		if (chunk == GetPartCount() - 1)
			return true; //1L is a moviechunk ?
		if (chunk == GetPartCount() - 2)
			return true; //2L is a moviechunk ?
		break;
		default:
		return false;
		break;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::IsBetterMovieChunk(uint16 oldchunk, uint16 newchunk)
{
	EMULE_TRY

//The routine is based on that both old&new-chunk are moviechunks
	if (oldchunk == 0) //Chunk = 1F
	{
	//	1/8F & !1/8L & (newchunk=1L | 2L)
		if ( IsComplete(0, (MINFIRSTMOVIESIZE) - 1) //first 1/8 chunk present
		     && !IsComplete((GetFileSize() - 1) - (MINLASTMOVIESIZE) , GetFileSize() - 1) //last 1/8 chunk not present
		     && ( newchunk == GetPartCount() - 1  //better chunk could be 1L
		          || (newchunk == GetPartCount() - 2) ) ) //better chunk could be 2L
			return true;
	}
	else if (oldchunk == GetPartCount() - 1 || oldchunk == GetPartCount() - 2) //Chunk = 1L | 2L
	{
	//	(1/8L & !1/8F) | (1/8L & 1/8F)
		if ( ( ( !IsComplete(0, (MINFIRSTMOVIESIZE) - 1) //first 1/8 chunk not present
		         && IsComplete((GetFileSize() - 1) - (MINLASTMOVIESIZE) , GetFileSize() - 1) ) //last 1/8 chunk present
		       || ( IsComplete(0, (MINFIRSTMOVIESIZE) - 1) //first 1/8 chunk present
		            && IsComplete((GetFileSize() - 1) - (MINLASTMOVIESIZE) , GetFileSize() - 1) ) //last 1/8 chunk present
		       && newchunk == 0 ) )  //better chunk could be 1F
			return true;
	}
	else if (oldchunk == 1) //Chunk = 2F
	{
	//	1F | 1L
		if (newchunk == 0 || newchunk == GetPartCount() - 1)
			return true;
	}
	else if (oldchunk == 2) //Chunk = 3F
	{
	//	1F | 1L otherwise 2F
		if (newchunk == 0 || newchunk == GetPartCount() - 1)
			return true;
		else if (newchunk == 1)
			return true;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//Movie1 disables (movie --> normal) & (movie --> worst movie)
bool CPartFile::Movie1(uint8 mode, uint16 oldchunk, uint16 newchunk)
{
	EMULE_TRY

	if (IsMovie())
	{
		if (!IsMovieChunk(mode, newchunk) && IsMovieChunk(mode, oldchunk))
			return false;
		if (IsMovieChunk(mode, newchunk) && IsMovieChunk(mode, oldchunk))
		{
			if (!IsBetterMovieChunk(oldchunk, newchunk))
				return false;
		}
	}
	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//Movie2 enables (normal --> movie) & (movie --> better movie)
bool CPartFile::Movie2(uint8 mode, uint8 raremovie, bool forcepreview, uint16 oldchunk, uint16 newchunk)
{
	EMULE_TRY

	if ( IsMovie() && ( ( m_srcPartFrequencies[oldchunk] > raremovie
	                      && m_srcPartFrequencies[newchunk] > raremovie )      //old & new chunk are nonrare
	                    || (m_srcPartFrequencies[oldchunk] == m_srcPartFrequencies[newchunk]) //old & new chunk are same rare
	                    || forcepreview ) )            //Load preview chunks first no matter of rareness
	{
	//This Line just for debugging:
	//AddLogLine(true, CString("Movie2::Mode:%d old:%d new:%d rarem:%d - The new & old ones are nonrare,samerare or forced, now we decide if new is a better movie or could change to movie..."),mode,oldchunk,newchunk,raremovie);
		if (!IsMovieChunk(mode, oldchunk) && IsMovieChunk(mode, newchunk))
		{
			if (forcepreview)
				return true;
		}
		if (IsMovieChunk(mode, oldchunk) && IsMovieChunk(mode, newchunk))
		{
			if (IsBetterMovieChunk(oldchunk, newchunk))
			{
				if (forcepreview)
					return true;
			}
		}
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint8 CPartFile::PrepareComparePart(uint16 chunk)
{
	EMULE_TRY

	uint8 addchunk = 0;

//	Is this the last chunk ?
	if (chunk == GetPartCount() - 1)
	{
		uint32 lastblock = ceil(static_cast<double>(PARTSIZE) / EMBLOCKSIZE);
		uint32 lastblockinpart = lastblock;
		uint32 start;
	//	Get last possible blocknumber
		for (lastblockinpart = 0; lastblockinpart != lastblock;lastblockinpart++)
		{
			start = (PARTSIZE * chunk) + lastblockinpart * EMBLOCKSIZE;
			if (start >= GetFileSize())
				break;
		}
		addchunk = lastblock - lastblockinpart;
	}

	return GetRemainingBlocksInPart(chunk) + addchunk;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Barry - Also want to preview zip/rar files
bool CPartFile::IsArchive(bool onlyPreviewable)
{
	EMULE_TRY

	CString extension = CString(GetFileName()).Right(4);

	if (onlyPreviewable)
		return ((extension.CompareNoCase(".rar") == 0) || (extension.CompareNoCase(".zip") == 0));

	return (ED2KFT_ARCHIVE == GetED2KFileTypeID(GetFileName()));

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetPriority(uint8 np)
{
	EMULE_TRY

	priority = np;
	g_eMuleApp.m_pDownloadQueue->SortByPriority();
	UpdateDisplayedInfo(true);
	SavePartFile();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::StopFile()
{
	EMULE_TRY

	if (m_eStatus == PS_COMPLETE || m_eStatus == PS_COMPLETING)
		return ;

//	Barry - Need to tell any connected clients to stop sending the file
//	eklmn: send a downloading client a cancel
	CUpDownClient* pSource;
	Packet* packet = new Packet(OP_CANCELTRANSFER);

	for (POSITION pos = m_downloadingSourceList.GetHeadPosition();pos != NULL;)
	{
		pSource = m_downloadingSourceList.GetNext(pos);
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

		pSource->m_pRequestSocket->SendPacket(packet, false, true);
#endif //OLD_SOCKETS_ENABLED

		pSource->SetDownloadState(DS_ONQUEUE);
	}
	delete packet;

//eklmn: reset A4AF auto file
	if (g_eMuleApp.m_pDownloadQueue->GetA4AFAutoFile() == this)
		g_eMuleApp.m_pDownloadQueue->SetA4AFAutoFile(NULL);

	RemoveAllSources(true);

	m_bPaused = true;
	m_bStopped = true;
	m_dataRate = 0;
	m_iNumTransferringSrcs = 0;
	m_iSrcNNP = 0;
	m_iSrcOnQueue = 0;
	m_iSrcHighQR = 0;
	m_iSrcConnecting = 0;
	m_iSrcTooManyConns = 0;
	m_iSrcConnected = 0;
	m_iSrcTimeout = 0;
	m_iSrcConnViaServer = 0;
	m_iSrcUnknown = 0;
	m_iSrcRecHash = 0;
	m_iSrcLowToLow = 0;
	m_iSrcQueueFull = 0;
	m_iSrcNoUpload = 0;
	m_iSrcBanned = 0;
	m_iSrcError = 0;
	if (!m_bIsBeingDeleted)
		FlushBuffer();
	UpdateDisplayedInfo(true);
	if (!m_bIsBeingDeleted)
		SavePartFile();
	SaveSettingsFile();

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::PauseFile()
{
	EMULE_TRY

	if (m_eStatus == PS_COMPLETE || m_eStatus == PS_COMPLETING)
		return ;

	CUpDownClient* pSource;
	POSITION pos;
	Packet* packet = new Packet(OP_CANCELTRANSFER);

//eklmn: send a downloading client a cancel
	for (pos = m_downloadingSourceList.GetHeadPosition();pos != NULL;)
	{
		pSource = m_downloadingSourceList.GetNext(pos);
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(packet->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

		pSource->m_pRequestSocket->SendPacket(packet, false, true);
#endif //OLD_SOCKETS_ENABLED

		pSource->SetDownloadState(DS_ONQUEUE);
	}
//eklmn: at this point nobody is downloading, and we can try to swap sources
	for (int sl = 0;sl < SOURCESSLOTS;sl++)
	{
		if (!m_srcLists[sl].IsEmpty())
		{
			for (pos = m_srcLists[sl].GetHeadPosition(); pos != NULL;)
			{
				CUpDownClient* pSource = m_srcLists[sl].GetNext(pos);
			//	eklmn: swap source to any other file
				if (pSource->GetDownloadState() == DS_ONQUEUE)
					pSource->SwapToAnotherFile(NULL);
			}
		}
	}
	delete packet;
	m_bPaused = true;
	m_bStopped = false;
	m_dataRate = 0;
	m_iNumTransferringSrcs = 0;
	UpdateDisplayedInfo(true);
	if (!m_bIsBeingDeleted)
		SavePartFile();
	SaveSettingsFile();

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::ResumeFile()
{
	EMULE_TRY

	if (m_eStatus == PS_COMPLETE || m_eStatus == PS_COMPLETING)
		return ;
//eklmn: check if chunks are availible & set status
	SetStatus(PS_EMPTY);
	for (int i = 0; i < GetPartCount(); i++)
	{
		uint32 uStart = i * PARTSIZE;
		uint32 uEnd = ((m_lFileSize - 1) < (uStart + PARTSIZE - 1)) ? (m_lFileSize - 1) : (uStart + PARTSIZE - 1);
		if (IsComplete(uStart, uEnd) == true)
		{
			SetStatus(PS_READY);
			break;
		}
	}

	if (GetStartTimeReset())
	{	//SyruS needed for files which are paused after adding or loading
		m_SessionStartTime = CTime::GetCurrentTime();
		SetStartTimeReset(false);
	}

	m_bPaused = false;
	m_bStopped = false;
	m_dwLastFileSourcesRequestTime = 0;
	SavePartFile();
	SaveSettingsFile();
	UpdateDisplayedInfo(true);

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Added by Vorlost to load initial sources from Server after connect
void CPartFile::GetSourcesAfterServerConnect()
{
	EMULE_TRY

	if (m_dwLastFileSourcesRequestTime > 0)
	{
		m_dwLastFileSourcesRequestTime = 0;
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::GetPartfileStatus()
{
	EMULE_TRY

	CString mybuffer = _T("");

	if (GetTransferringSrcCount() > 0)
		mybuffer = GetResString(IDS_DOWNLOADING);
	else
	{
		if (!IsStalled())
			mybuffer = GetResString(IDS_WAITING);
		else
			mybuffer = GetResString(IDS_STALLED);
	}

	switch (GetStatus())
	{
		case PS_HASHING:
		case PS_WAITINGFORHASH:
		mybuffer = GetResString(IDS_HASHING);
		break;
		case PS_COMPLETING:
		mybuffer = GetResString(IDS_COMPLETING);
		break;
		case PS_COMPLETE:
		mybuffer = GetResString(IDS_COMPLETE);
		break;
		case PS_PAUSED:
		mybuffer = GetResString(IDS_PAUSED);
		break;
		case PS_ERROR:
		mybuffer = GetResString(IDS_ERRORLIKE);
		break;
		case PS_STOPPED:
		mybuffer = GetResString(IDS_STOPPED);
		break;
	}
	return mybuffer;

	EMULE_CATCH

	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CPartFile::GetPartfileStatusID()
{
	EMULE_TRY

	int	ipfStatus = 0;

	if (GetTransferringSrcCount() > 0)
		ipfStatus = PS_DOWNLOADING;
	else
	{
		if (!IsStalled())
			ipfStatus = PS_WAITINGFORSOURCE;
		else
			ipfStatus = PS_STALLED;
	}

	switch (GetStatus())
	{
		case PS_HASHING:
		case PS_WAITINGFORHASH:
		ipfStatus = PS_HASHING;
		break;
		case PS_COMPLETING:
		case PS_COMPLETE:
		case PS_PAUSED:
		case PS_ERROR:
		case PS_STOPPED:
		ipfStatus = GetStatus();
		break;
	}
	return ipfStatus;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CPartFile::getPartfileStatusRang()
{
	EMULE_TRY

//	Sort order is Complete - Completing - Downloading - Waiting - Stalled - Paused - Errors

	int status = 2;
	if (GetTransferringSrcCount() == 0)
	{
		if (!IsStalled())
			status = 3;
		else
			status = 4;
	}
	switch (GetStatus())
	{
		case PS_HASHING:
		case PS_WAITINGFORHASH:
		status = 6;
		break;
		case PS_COMPLETING:
		status = 1;
		break;
		case PS_COMPLETE:
		status = 0;
		break;
		case PS_PAUSED:
		status = 5;
		break;
		case PS_STOPPED:
		status = 7;
		break;
		case PS_ERROR:
		status = 8;
		break;
	}
	return status;

	EMULE_CATCH

	return 1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
sint32 CPartFile::GetTimeRemaining(bool bAverage /*=false*/)
{
	EMULE_TRY

	uint32	iDataRate = 0;
	sint32	iTimeRemaining = -1;

	if (bAverage)
		iDataRate = GetAvgDataRate();
	else
		iDataRate = GetDataRate();
	iTimeRemaining = (iDataRate > 0 ? (GetFileSize() - GetCompletedSize()) / iDataRate : -1);
	return ((iTimeRemaining > 0 && iTimeRemaining <= 8640000 /*100 days*/) ? iTimeRemaining : -1);

	EMULE_CATCH

	return -1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetAvgDataRate() recalculates the average bytes/second for the session. It stores the
//	average in 'm_AvgDataRate' as well as returning it.
uint32 CPartFile::GetAvgDataRate(bool bUpdated /*=false*/)
{
	EMULE_TRY

	if (bUpdated)
	{
		CTimeSpan	timeSpan;

	//	Calc on last FlushBuffer() to keep sessionstats for completed files
		if (m_eStatus == PS_COMPLETING || m_eStatus == PS_COMPLETE)
		{
			timeSpan = GetFlushTimeSpan();
		}
		else
		{
		//	TODO: Modify to use time since first data
			timeSpan = GetSessionTimeSpan();
		}

	//	Delay 1st calculation a bit & dont use m_dataRate for balancing (peak-stopper?)
		uint32	iSeconds = timeSpan.GetTotalSeconds();

		if (iSeconds > 10)
			m_AvgDataRate = ((GetSessionTransferred() /*+ GetDataRate()*/) / (iSeconds /*+ 1*/));
		else
			m_AvgDataRate = 0;
	}

	EMULE_CATCH
	return m_AvgDataRate;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::PreviewFile()
{
	EMULE_TRY

	if (IsArchive(true))
	{
		if ((!m_bRecoveringArchive) && (!m_bPreviewing))
			CArchiveRecovery::recover(this, true);
		return ;
	}

	if (!PreviewAvailable())
	{
		return ;
	}

	if (g_eMuleApp.m_pGlobPrefs->BackupPreview())
	{
		m_bPreviewing = true;
		CPreviewThread* pThread = (CPreviewThread*) AfxBeginThread(RUNTIME_CLASS(CPreviewThread), THREAD_PRIORITY_NORMAL + g_eMuleApp.m_pGlobPrefs->GetMainProcessPriority(), 0, CREATE_SUSPENDED);
		pThread->SetValues(this, g_eMuleApp.m_pGlobPrefs->GetVideoPlayer());
		pThread->ResumeThread();
	}
	else
	{
		CString strLine = this->GetFullName();
		CString player = g_eMuleApp.m_pGlobPrefs->GetVideoPlayer();
		strLine.Replace(_T(".met"), _T(""));

		// If path has spaces use quotes
		if (strLine.Find(_T(' ')) != -1)
		{
			strLine = _T('\"') + strLine + _T('\"');
		}

		if (player.IsEmpty())
		{
			ShellExecute(NULL, _T("open"), strLine, NULL, NULL, SW_SHOWNORMAL);
		}
		else
		{
			CString strPlayerPath = player;
			int i = strPlayerPath.ReverseFind('\\');
			strPlayerPath = (i >= 0) ? strPlayerPath.Left(i + 1) : _T("");
			ShellExecute(NULL, _T("open"), player, strLine, strPlayerPath, SW_SHOWNORMAL);
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::PreviewAvailable()
{
	EMULE_TRY

	uint64 space = GetFreeDiskSpaceX(m_strTempDir);

	if (GetStatus() == PS_COMPLETING || GetStatus() == PS_COMPLETE || GetStatus() == PS_HASHING)		// no preview! :)
		return false;

//	Barry - Allow preview of archives of any length > 1k
	if (IsArchive(true))
	{
		if (GetFileSize() > 1024 && GetCompletedSize() > 1024 && !m_bRecoveringArchive && ((space + 100000000) > (2 * GetFileSize())))
			return true;
		else
			return false;
	}

	if (!IsMovie())
		return false;

	if (g_eMuleApp.m_pGlobPrefs->BackupPreview())
		return !( (GetStatus() != PS_READY && GetStatus() != PS_PAUSED)
		          || m_bPreviewing || GetPartCount() < 5 || !IsMovie() || (space + 100000000) < GetFileSize()
		          || (!IsComplete(0, PARTSIZE - 1) || !IsComplete(PARTSIZE * (GetPartCount() - 1), GetFileSize() - 1)) );
	else
	{
		TCHAR szVideoPlayerFileName[_MAX_FNAME];
		_tsplitpath(g_eMuleApp.m_pGlobPrefs->GetVideoPlayer(), NULL, NULL, szVideoPlayerFileName, NULL);

	//	enable the preview command if the according option is specified 'PreviewSmallBlocks'
	//	or if VideoLAN client is specified
		if (g_eMuleApp.m_pGlobPrefs->GetPreviewSmallBlocks() || !_tcsicmp(szVideoPlayerFileName, _T("vlc")))
		{
			if (m_bPreviewing)
				return false;

			uint8 uState = GetStatus();
			if (!(uState == PS_READY || uState == PS_EMPTY || uState == PS_PAUSED))
				return false;

			EED2KFileType eFileType = GetED2KFileTypeID(GetFileName());
			if (!(eFileType == ED2KFT_VIDEO || eFileType == ED2KFT_AUDIO || eFileType == ED2KFT_CDIMAGE))
			{
				return false;
			}

		//	If it's an MPEG file, VLC is even capable of showing parts of the file if the beginning of the file is missing!
			bool bMPEG = false;
			LPCTSTR pszExt = _tcsrchr(GetFileName(), _T('.'));
			if (pszExt != NULL)
			{
				CString strExt(pszExt);
				strExt.MakeLower();
				bMPEG = (strExt == _T(".mpg") || strExt == _T(".mpeg") || strExt == _T(".mpe") || strExt == _T(".mp3") || strExt == _T(".mp2") || strExt == _T(".mpa"));
			}

			if (bMPEG)
			{
			//	TODO: search a block which is at least 16K (Audio) or 256K (Video)
				if (GetCompletedSize() < 16 * 1024)
					return false;
			}
			else
			{
			//	For AVI files it depends on the used codec..
				if (!IsComplete(0, 256 * 1024))
					return false;
			}

			return true;
		}
		else
		{
			return !( (GetStatus() != PS_READY && GetStatus() != PS_PAUSED)
			          || m_bPreviewing || GetPartCount() < 2 || !IsMovie() || !IsComplete(0, PARTSIZE - 1) );
		}
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Packet*	CPartFile::CreateSrcInfoPacket(CUpDownClient* forClient)
{
	EMULE_TRY

	if (forClient->m_pReqPartFile != this)
		return NULL;

	if (GetSourceCount() == 0)
		return NULL;

	CMemFile data;
	uint16 nCount = 0;
	CUpDownClient* pSource;
	bool bNeeded;
	POSITION pos;

	data.Write(m_fileHash, 16);
	data.Write(&nCount, 2);

	const uint16 uPartCount = GetPartCount();

	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		if (!m_srcLists[sl].IsEmpty())
			for (pos = m_srcLists[sl].GetHeadPosition();pos != NULL;)
			{
				bNeeded = false;
				pSource = m_srcLists[sl].GetNext(pos);

			//	eklmn: don't send a client to himself as a source
				if (forClient == pSource)
					continue;

			//	Modified by Tarod -sends sources with LowID that are in same sever
			//	that client is requesting sources
				if ( pSource->HasLowID()
				     && ( pSource->GetServerIP() != forClient->GetServerIP()
				          || pSource->GetServerPort() != forClient->GetServerPort() ) )
				{
					continue;
				}

			//	LANCAST (moosetea) - We don't share Lan clients sources as they are private ips/userids
				if (pSource->IsOnLAN())
					continue;

			//	Added by Tarod, do not send incomplete established conns. to avoid sending
			//	no valid sources
				switch (pSource->GetDownloadState())
				{
					case DS_NONE:
					case DS_CONNECTING:
					case DS_ERROR:
					case DS_TOOMANYCONNS:
					continue;
					break;
				}

			//	only send source which have needed parts for this client if possible
				uint8* pReqStatus = forClient->GetPartStatus();
				uint8* pSrcStatus = pSource->GetPartStatus();
				if (forClient->m_pReqPartFile == this && pReqStatus && pSrcStatus)
				{
					for (int x = 0; x < uPartCount; x++)
					{
						if (pSrcStatus[x] && !pReqStatus[x])
						{
							bNeeded = true;
							break;
						}
					}
				}
				else if (pSource->GetPartStatus())
				{
					uint8 * srcstatus = pSource->GetPartStatus();
					for (int x = 0; x < uPartCount; x++)
					{
						if (srcstatus[x])
						{
							bNeeded = true;
							break;
						}
					}
				}

				if (bNeeded)
				{
					nCount++;
					uint32 dwID = pSource->GetUserID();
					uint16 nPort = pSource->GetUserPort();
					uint32 dwServerIP = pSource->GetServerIP();
					uint16 nServerPort = pSource->GetServerPort();
					data.Write(&dwID, 4);
					data.Write(&nPort, 2);
					data.Write(&dwServerIP, 4);
					data.Write(&nServerPort, 2);
					if (nCount > 500)
						break;
				}
			}

	if (!nCount)
		return 0;
	data.Seek(16, 0);
	data.Write(&nCount, 2);

	Packet* pPacket = new Packet(&data, OP_EMULEPROT);
	pPacket->m_eOpcode = OP_ANSWERSOURCES;
	if (nCount > 28)
		pPacket->PackPacket();
//	AddDebugLogLine(false, "Send:Source User(%s) File(%s) Count(%i)", forClient->GetUserName(), GetFileName(), nCount);
	return pPacket;

	EMULE_CATCH

	return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//v- eklmn: support Source Exchange v2
void CPartFile::AddClientSources(CMemFile* sources, uint8 sourceexchangeversion)
{
	EMULE_TRY

//if this part file is paused or complete/completing no sources needed any more (Cax2 + SyruS check)
	if ( GetStatus() == PS_PAUSED
	     || GetStatus() == PS_STOPPED
	     || GetStatus() == PS_COMPLETE
	     || GetStatus() == PS_COMPLETING )
		return ;

	uint16 nCount;
	sources->Read(&nCount, 2);

//	Check if the data size matches the 'nCount' for v1 or v2 and eventually correct the source
//	exchange version while reading the packet data. Otherwise we could experience a higher
//	chance in dealing with wrong source data, userhashs and finally duplicate sources.
	UINT uDataSize = sources->GetLength() - sources->GetPosition();
	if ((UINT)nCount * (4 + 2 + 4 + 2) == uDataSize)
		sourceexchangeversion = 1;
	else if ((UINT)nCount * (4 + 2 + 4 + 2 + 16) == uDataSize)
		sourceexchangeversion = 2;
	else
	{
	//	If v3 inserts additional data (like v2), the above code will correctly filter those packets.
	//	If v3 appends additional data after <count>(<Sources>)[count], we are in trouble with the
	//	above code. Though a client which does not understand v3+ should never receive such a packet.
		g_eMuleApp.m_pdlgEmule->AddDebugLogLine(false, _T("Received invalid source exchange packet (v%u) of data size %u for %s\n"), sourceexchangeversion, uDataSize, GetFileName());
		return ;
	}

	for (int i = 0;i != nCount;i++)
	{
		CClientSource source;
		sources->Read(&source.sourceID, 4);
		sources->Read(&source.sourcePort, 2);
		sources->Read(&source.serverIP, 4);
		sources->Read(&source.serverPort, 2);
		if (sourceexchangeversion > 1)
			sources->Read(&source.achUserHash, 16);
		AddClientSource(&source, i, true, sourceexchangeversion);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::GetSizeToTransfer(uint32& pui32SizeToTransfer)
{
	EMULE_TRY

	for (POSITION pos = gaplist.GetHeadPosition();pos != 0;)
	{
		Gap_Struct * cur_gap = gaplist.GetNext(pos);
		pui32SizeToTransfer += cur_gap->m_dwEndOffset - cur_gap->m_dwStartOffset;
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Updates priority of file if autopriority is activated
void CPartFile::UpdateDownloadAutoPriority(void)
{
	EMULE_TRY

	if (IsAutoPrioritized())
	{
		int iValidSources = GetSourceCount() - GetNotCurrentSourcesCount();

		if (iValidSources < g_eMuleApp.m_pGlobPrefs->PriorityHigh() && priority != PR_HIGH)
			SetPriority(PR_HIGH);
		else if (iValidSources >= g_eMuleApp.m_pGlobPrefs->PriorityHigh() && iValidSources < g_eMuleApp.m_pGlobPrefs->PriorityLow() && priority != PR_NORMAL)
			SetPriority(PR_NORMAL);
		else if (iValidSources >= g_eMuleApp.m_pGlobPrefs->PriorityLow() && priority != PR_LOW)
			SetPriority(PR_LOW);
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	making this function return a higher when more sources have the extended
//	protocol will force you to ask a larger variety of people for sources
uint32 CPartFile::GetCommonFilePenalty()
{
//	TODO: implement, but never return less than MINCOMMONPENALTY!
	return MINCOMMONPENALTY;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::GetStatsFullPath() const
{
	return RemoveFileExtension(m_strFullName) + _T(".stats");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//--- xrmb:keepPartFileStats ---
bool CPartFile::LoadPartFileStats()
{
	EMULE_TRY

	FILE * fh = _tfopen(GetStatsFullPath(), _T("rb"));

//--- if we cant open this file... you lose the stats ---
	if (fh == NULL)
		return false;

//--- get version ---
	uint8	version = 0;
	if (fread(&version, sizeof(version), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

	uint32 alltimerequested = 0;
	uint32 alltimeaccepted = 0;
	uint64 alltimetransferred = 0;

	if (fread(&alltimerequested, sizeof(alltimerequested), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}
	if (fread(&alltimeaccepted, sizeof(alltimeaccepted), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}
	if (fread(&alltimetransferred, sizeof(alltimetransferred), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

	statistic.AddAll(alltimerequested, alltimeaccepted, alltimetransferred);

//--- this is for the parttraffic ---
//--- i thought about explicit adding the traffic like for the other stats above ---
//--- but it doesnt make sense, because you'll no transfer if you reach this ---
	if (LoadFromFileTraffic(fh, version) == false)
	{
		fclose(fh);
		return false;
	}

	fclose(fh);

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CPartFile::SavePartFileStats()
{
	EMULE_TRY

	if (!m_strFullName || IsBadStringPtr(m_strFullName, MAX_PATH))
	{
		return false;
	}
	FILE *fh = _tfopen(GetStatsFullPath(), _T("wb"));

//--- if we cant open this file... you lose the stats ---
	if (fh == NULL)
		return false;

//--- set version ---
	uint8	version = 2;
	if (fwrite(&version, sizeof(version), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

	uint32 alltimerequested = statistic.GetAllTimeRequests();
	uint32 alltimeaccepted = statistic.GetAllTimeAccepts();
	uint64 alltimetransferred = statistic.GetAllTimeTransferred();

	if (fwrite(&alltimerequested, sizeof(alltimerequested), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}
	if (fwrite(&alltimeaccepted, sizeof(alltimeaccepted), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}
	if (fwrite(&alltimetransferred, sizeof(alltimetransferred), 1, fh) != 1)
	{
		fclose(fh);
		return false;
	}

//--- this is for the parttraffic ---
	if (SaveToFileTraffic(fh) == false)
	{
		fclose(fh);
		return false;
	}

	fclose(fh);

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::RemovePartFileStats()
{
	EMULE_TRY

	if (_tremove(GetStatsFullPath()) && errno != ENOENT)
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_DELETEFAILED), GetStatsFullPath());

	EMULE_CATCH
}
//--- :xrmb ---
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Used as separate thread to complete file [Lord KiRon]
UINT CPartFile::CompleteThreadProc(CPartFile* pFile)
{
	EMULE_TRY

	if (g_eMuleApp.m_pGlobPrefs->IsNTBased())
	{
		SetThreadLocale(g_eMuleApp.m_pGlobPrefs->GetLanguageID());
	}

	if (!pFile)
	{
		return -1;
	}
	CSingleLock Lock(&s_CompletingMutex, TRUE); // Prevent of completing two files in same time
	pFile->PerformFileComplete();
	return 0;

	EMULE_CATCH

	return -1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::DtLocale(CTime dt)
{

	SYSTEMTIME st;
	dt.GetAsSystemTime(st);

	int nDateSize = GetDateFormat(LOCALE_USER_DEFAULT, 0, &st, NULL, NULL, 0);
	int nTimeSize = GetTimeFormat(LOCALE_USER_DEFAULT, 0, &st, NULL, NULL, 0);

	CString strDate;
	GetDateFormat(LOCALE_USER_DEFAULT, DATE_SHORTDATE, &st, NULL, strDate.GetBuffer(nDateSize), nDateSize);
	strDate.ReleaseBuffer();

	CString strTime;
	GetTimeFormat(LOCALE_USER_DEFAULT, TIME_NOSECONDS, &st, NULL, strTime.GetBuffer(nTimeSize), nTimeSize);
	strTime.ReleaseBuffer();

	CString strDay;
	switch (dt.GetDayOfWeek())
	{
		case 1:
		strDay = GetResString(IDS_SCH_EXCEPT_SUN);
		break;
		case 2:
		strDay = GetResString(IDS_SCH_EXCEPT_MON);
		break;
		case 3:
		strDay = GetResString(IDS_SCH_EXCEPT_TUE);
		break;
		case 4:
		strDay = GetResString(IDS_SCH_EXCEPT_WED);
		break;
		case 5:
		strDay = GetResString(IDS_SCH_EXCEPT_THU);
		break;
		case 6:
		strDay = GetResString(IDS_SCH_EXCEPT_FRI);
		break;
		case 7:
		strDay = GetResString(IDS_SCH_EXCEPT_SAT);
		break;
	}

	return strDay + _T(", ") + strDate + _T(", ") + strTime;

}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	LoadSettingsFile() loads settings for partfiles (Movie Preview Mode, Stopped status, Discard SCB, Preallocation)
void CPartFile::LoadSettingsFile()
{
	EMULE_TRY

	CString strIniFile = CString(RemoveFileExtension(m_strFullName)) + CString(".settings");

	if (PathFileExists(strIniFile))
	{
		CIni filesettings(strIniFile);

		filesettings.SetDefaultCategory(_T("Settings"));
		uint8 Mode = filesettings.GetInt(_T("MoviePreviewMode"), 0);
		CKnownFile::SetMovieMode(Mode);
		m_bDiscardSuperCompressed = filesettings.GetBool(_T("DiscardSuperCompressed"), false);
		m_bStopped = filesettings.GetBool(_T("StoppedStatus"), m_bStopped);
		m_bIsPreallocated = filesettings.GetBool(_T("Preallocated"), m_bIsPreallocated);

		filesettings.CloseWithoutSave();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	SaveSettingsFile() save settings for partfiles (Movie Preview Mode, Stopped status, Discard SCB, Preallocation)
void CPartFile::SaveSettingsFile()
{
	EMULE_TRY

	CString strIniFile = CString(RemoveFileExtension(m_strFullName)) + CString(".settings");

	CIni filesettings(strIniFile);

	filesettings.SetDefaultCategory(_T("Settings"));
	filesettings.SetInt(_T("MoviePreviewMode"), ((CKnownFile*)this)->GetMovieMode());
	filesettings.SetBool(_T("DiscardSuperCompressed"), m_bDiscardSuperCompressed);
	filesettings.SetBool(_T("StoppedStatus"), m_bStopped);
	filesettings.SetBool(_T("Preallocated"), m_bIsPreallocated);

	filesettings.SaveAndClose();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Barry - Replaces BlockReceived()
//
//         Originally this only wrote to disk when a full 180k block
//         had been received from a client, and only asked for data in
//         180k blocks.
//
//		   This meant that on average 90k was lost for every connection
//		   to a client data source. That is a lot of wasted data.
//
//		   To reduce the lost data, packets are now written to a buffer
//		   and flushed to disk regularly regardless of size downloaded.
//		   This includes compressed packets.
//
//		   Data is also requested only where gaps are, not in 180k blocks.
//		   The requests will still not exceed 180k, but may be smaller to
//		   fill a gap.
//
uint32 CPartFile::WriteToBuffer(uint32 transize, BYTE *data, uint32 start, uint32 dwEnd, Requested_Block_Struct *block)
{
	EMULE_TRY

//	Increment transferred bytes counter for this file
	m_dwBytesTransferred += transize;

//	This is needed a few times
	uint32 lenData = dwEnd - start + 1;

	if (lenData > transize)
	{
		m_iGainDueToCompression += lenData - transize;
		m_iSessionGainDueToCompression += lenData - transize;
	}

//	Occasionally packets are duplicated, no point writing it twice
	if (IsComplete(start, dwEnd))
	{
		AddDebugLogLine(false, _T("File '%s' has already been written from %lu to %lu"), GetFileName(), start, dwEnd);
		return 0;
	}

//	Create copy of data as new buffer
	BYTE *buffer = new BYTE[lenData];
	memcpy2(buffer, data, lenData);

//	Create a new buffered queue entry
	PartFileBufferedData *item = new PartFileBufferedData;
	item->m_pbyteBuffer = buffer;
	item->m_dwStartOffset = start;
	item->m_dwEndOffset = dwEnd;
	item->block = block;

//	Add to the queue in the correct position (most likely the end)
	PartFileBufferedData *queueItem;
	bool added = false;
	POSITION pos = m_BufferedData_list.GetTailPosition();
	while (pos != NULL)
	{
		queueItem = m_BufferedData_list.GetPrev(pos);
		if (item->m_dwEndOffset > queueItem->m_dwEndOffset)
		{
			added = true;
			m_BufferedData_list.InsertAfter(pos, item);
			break;
		}
	}
	if (!added)
		m_BufferedData_list.AddHead(item);

//	Increment buffer size marker
	m_nTotalBufferData += lenData;

//	Mark this small section of the file as filled
	FillGap(item->m_dwStartOffset, item->m_dwEndOffset);

//	Update the flushed mark on the requested block
//	The loop here is unfortunate but necessary to detect deleted blocks.
	pos = requestedblocks_list.GetHeadPosition();
	while (pos != NULL)
	{
		if (requestedblocks_list.GetNext(pos) == item->block)
			item->block->m_dwBytesTransferred += lenData;
	}

//	v- eklmn: bugfix(20): check last part in WriteToBuffer correct way
	uint16 partnumber = floor(start / static_cast<double>(PARTSIZE));
	const uint32 uStart = partnumber * PARTSIZE;
	const uint32 uEnd = ((m_lFileSize - 1) < (uStart + PARTSIZE - 1)) ? (m_lFileSize - 1) : (uStart + PARTSIZE - 1);
	if (IsComplete(uStart, uEnd))
	{
		FlushBuffer();
		m_ReqParts[partnumber] = PARTSIZE;
	}

	m_timeLastDownTransfer = CTime::GetCurrentTime();

//	Return the length of data written to the buffer
	return lenData;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	FlushBuffer() flushes buffered data from the source to the part file
void CPartFile::FlushBuffer(void)
{
	EMULE_TRY

	m_nLastBufferFlushTime = GetTickCount();

//	If there's no buffered data to flush, just return
	if (m_BufferedData_list.IsEmpty())
		return ;

	CSingleLock sLock(&m_FileCompleteMutex, TRUE); // will be unlocked on exit

//	Reset the Last Down Transfer timer
	m_timeLastDownTransfer = CTime::GetCurrentTime();
	GetAvgDataRate(true);	//SyruS do update on flushing

//	Calculate the number of parts in the part file
	uint32	iPartCount = GetPartCount();
	bool *changedPart = new bool[iPartCount];

//	AddDebugLogLine(false, "Flushing file %s - buffer size = %ld bytes (%ld queued items) transferred = %ld [time = %ld]\n", GetFileName(), m_nTotalBufferData, m_BufferedData_list.GetCount(), transferred, m_nLastBufferFlushTime);

	try
	{
	//
	//	Remember which parts need to be checked at the end of the flush
	//
		for (int uPartNumber = 0; (uint32)uPartNumber < iPartCount; uPartNumber++)
		{
			changedPart[uPartNumber] = false;
		}

	//	Ensure file is big enough to write data to (the last item will be the furthest from the start)
		PartFileBufferedData *item = m_BufferedData_list.GetTail();

	//	If the end of the buffered data is past the current end of the part file, extend it
		if (m_hPartFile.GetLength() <= item->m_dwEndOffset)
			m_hPartFile.SetLength(item->m_dwEndOffset + 1);	// netwolf: file should not be too large (taab)

	//	For each block of buffered data...
		for (int i = m_BufferedData_list.GetCount(); i > 0; i--)
		{
		//	Get top item
			item = m_BufferedData_list.GetHead();

		//	This is needed a few times
			uint32 lenData = item->m_dwEndOffset - item->m_dwStartOffset + 1;

		//SyruS (0.29a) SLUGFILLER: SafeHash - could be more than one part
			/*int curPart = item->m_dwStartOffset/PARTSIZE;
			changedPart[curPart] = true;*/

		//	For each part spanned by the buffered data, set the part's changed flag
			for (uint32 curPart = item->m_dwStartOffset / PARTSIZE; curPart <= item->m_dwEndOffset / PARTSIZE; curPart++)
				changedPart[curPart] = true;
		//	SLUGFILLER: SafeHash

		//	Go to the correct position in file and write block of data
			m_hPartFile.Seek(item->m_dwStartOffset, CFile::begin);
			m_hPartFile.Write(item->m_pbyteBuffer, lenData);

		//	Remove item from queue
			m_BufferedData_list.RemoveHead();

		//	Decrease buffer size
			m_nTotalBufferData -= lenData;

		//	Release memory used by this item
			delete [] item->m_pbyteBuffer;
			delete item;
		}

	//	If the size of the part file is greater than the size of the target file, set it to the
	//	size of the target file.
		if (m_hPartFile.GetLength() > m_lFileSize)
			m_hPartFile.SetLength(m_lFileSize);

	//	Flush to disk
		m_hPartFile.Flush();

	//	Check each part of the file
		uint32	partRange = (m_lFileSize % PARTSIZE) - 1;

	//	For each part in the part file (in reverse)...
		for (int uPartNumber = iPartCount - 1; uPartNumber >= 0; uPartNumber--)
		{
		//	If we didn't just write it, skip it
			if (changedPart[uPartNumber] == false)
			{
			//	Any parts other than last must be full size
				partRange = PARTSIZE - 1;
				continue;
			}

		//	Is this 9MB part complete
			if (IsComplete(PARTSIZE * uPartNumber, (PARTSIZE * (uPartNumber + 1)) - 1))
			{
			//	Is part corrupt
				if (!HashSinglePart(uPartNumber))
				{
					AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_PARTCORRUPT), uPartNumber, GetFileName());
					AddGap(PARTSIZE * uPartNumber, (PARTSIZE * uPartNumber + partRange));
					corrupted_list.AddTail(uPartNumber);
				//	Reduce transferred amount by corrupt amount
					this->m_iLostDueToCorruption += (partRange + 1);
					this->m_iSessionLostDueToCorruption += (partRange + 1);
				}
				else
				{
				//	Successfully completed part, make it available for sharing
					if (m_eStatus == PS_EMPTY)
					{
						SetStatus(PS_READY);
						g_eMuleApp.m_pSharedFilesList->SafeAddKnownFile(this);
					}
				}
			}
			else if (IsCorruptedPart(uPartNumber))
			{
			//	Try to recover with minimal loss
				if (HashSinglePart(uPartNumber))
				{
					m_iTotalPacketsSavedDueToICH++;
					FillGap(PARTSIZE * uPartNumber, (PARTSIZE * uPartNumber + partRange));
					RemoveBlockFromList(PARTSIZE * uPartNumber, (PARTSIZE * uPartNumber + partRange));
					AddLogLine(true, RGB_LOG_SUCCESS + GetResString(IDS_ICHWORKED), uPartNumber, GetFileName());
					//	Successfully completed part, make it available for sharing
					if (m_eStatus == PS_EMPTY)
					{
						SetStatus(PS_READY);
						g_eMuleApp.m_pSharedFilesList->SafeAddKnownFile(this);
					}
					g_eMuleApp.m_pGlobPrefs->Add2SessionPartsSavedByICH();
				}
			}
		//	Any parts other than last must be full size
			partRange = PARTSIZE - 1;
		}

	//	Update met file
		SavePartFile();

	//	Is this file finished?
		if (gaplist.IsEmpty() && !m_bHashSetNeeded) // SLUGFILLER: hashlistRecheck - can't complete without a hashset...
			CompleteFile(false);
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();

		if (g_eMuleApp.m_pGlobPrefs->IsErrorBeepEnabled())
			Beep(800, 200);

		if (error->m_cause == CFileException::diskFull)
		{
			CString strSpace = CastItoXBytes(GetFreeDiskSpaceX(m_strTempDir));
			CString strTemp;
			strTemp.Format("%s (" + m_strTempDir + ")", strSpace);
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_OUTOFSPACE), GetFileName());
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_DWTOT_FS), strTemp);
		//	Purity - Sending message for insufficient diskspace
			CString MessageText;
			MessageText.Format(GetResString(IDS_ERR_OUTOFSPACE), GetFileName());
			g_eMuleApp.m_pSMTPConnection->SendMail(MessageText, g_eMuleApp.m_pGlobPrefs->GetNotifierPopOnImportantError(), g_eMuleApp.m_pGlobPrefs->IsSMTPWarningEnabled());
			g_eMuleApp.m_pdlgEmule->ShowNotifier(MessageText, TBN_IMPORTANTEVENT, false, g_eMuleApp.m_pGlobPrefs->GetNotifierPopOnImportantError());
		}
		else
		{
			AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_WRITEERROR), GetFileName(), GetErrorMessage(error));
			SetStatus(PS_ERROR);
		}
		PauseFile();
		error->Delete();
	}
	catch (...)
	{
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_WRITEERROR), GetFileName(), GetResString(IDS_UNKNOWN));
		SetStatus(PS_ERROR);
		PauseFile();
	}
	delete[] changedPart;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Barry - This will invert the gap list, up to caller to delete gaps when done
//	'Gaps' returned are really the filled areas, and guaranteed to be in order
void CPartFile::GetFilledList(CTypedPtrList<CPtrList, Gap_Struct*> *filled)
{
	EMULE_TRY

	Gap_Struct * gap = NULL;
	Gap_Struct *best = NULL;
	POSITION pos;
	uint32 start = 0;
	uint32 bestEnd = 0;

//	Loop until done
	bool finished = false;
	while (!finished)
	{
		finished = true;
	//	Find first gap after current start pos
		bestEnd = m_lFileSize;
		pos = gaplist.GetHeadPosition();
		best = gaplist.GetHead();
		while (pos != NULL)
		{
			gap = gaplist.GetNext(pos);
			if ((gap->m_dwStartOffset > start) && (gap->m_dwEndOffset < bestEnd))
			{
				best = gap;
				bestEnd = best->m_dwEndOffset;
				finished = false;
			}
		}

		if (!finished)
		{
		//	Invert this gap
			gap = new Gap_Struct;
			gap->m_dwStartOffset = start;
			gap->m_dwEndOffset = best->m_dwStartOffset - 1;
			start = best->m_dwEndOffset + 1;
			filled->AddTail(gap);
		}
		else if (best->m_dwEndOffset < m_lFileSize)
		{
			gap = new Gap_Struct;
			gap->m_dwStartOffset = best->m_dwEndOffset + 1;
			gap->m_dwEndOffset = m_lFileSize;
			filled->AddTail(gap);
		}
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::UpdateFileRatingCommentAvail()
{
	EMULE_TRY

	if (!this)
		return ;

	bool prev = (hasComment || hasRating);

	hasComment = false;
	hasRating = false;

	POSITION pos;
	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		if (!m_srcLists[sl].IsEmpty())
			for (pos = m_srcLists[sl].GetHeadPosition();pos != NULL;)
			{
				CUpDownClient* pSource = m_srcLists[sl].GetNext(pos);
				if (pSource->GetFileComment().GetLength() > 0)
					hasComment = true;
				if (pSource->GetFileRating() != PF_RATING_NONE)
					hasRating = true;
				if (hasComment && hasRating)
					break;
			}

//	#zegzav:pastcomment
	if ((!hasComment) || (!hasRating))
	{
		for (pos = m_pastCommentList.GetHeadPosition(); pos != 0;)
		{
			CPastComment &pc = m_pastCommentList.GetNext(pos);
			if (pc.GetComment().GetLength() > 0)
				hasComment = true;
			if (pc.GetRating() > 0)
				hasRating = true;
			if (hasComment && hasRating)
				break;
		}
	}

	if (prev != (hasComment || hasRating))
		UpdateDisplayedInfo();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16 CPartFile::GetSourceCount()
{
	EMULE_TRY

	uint16 count = 0;
	for (int i = 0;i < SOURCESSLOTS;i++)
		count += m_srcLists[i].GetCount();
	return count;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
EnumPartFileRating CPartFile::GetRating()
{
	EMULE_TRY

	if (!hasRating)
		return PF_RATING_NONE;

	int num, tot, fRate;

	num = tot = 0;

	POSITION pos;

//	For each source slot...
	for (int sl = 0; sl < SOURCESSLOTS; sl++)
	{
	//	If the slot isn't empty...
		if (!m_srcLists[sl].IsEmpty())
		{
			for (pos = m_srcLists[sl].GetHeadPosition(); pos != NULL;)
			{
				fRate = m_srcLists[sl].GetNext(pos)->GetFileRating();
				if (fRate > 0)
				{
					num++;
				//	Cax2 - bugfix: for some $%#ing reason  fair=4 & good=3, breaking the progression from fake(1) to excellent(5)
					if (fRate == PF_RATING_GOOD || fRate == PF_RATING_FAIR)
						fRate = (fRate == PF_RATING_GOOD) ? 4 : 3;

					tot += fRate;
				}
			}
		}
	}

//	#zegzav:pastcomment
	for (pos = m_pastCommentList.GetHeadPosition(); pos != 0;)
	{
		fRate = m_pastCommentList.GetNext(pos).GetRating();

		if (fRate > 0)
		{
			num++;
		//Cax2 - bugfix: for some $%#ing reason  fair=4 & good=3, breaking the progression from fake(1) to excellent(5)
			if (fRate == PF_RATING_GOOD || fRate == PF_RATING_FAIR)
				fRate = (fRate == PF_RATING_GOOD) ? 4 : 3;

			tot += fRate;
		}
	}

	if (num > 0)
	{
		num = static_cast<double>(tot) / num + .5;	//Cax2 - get the average of all the ratings
	//	Cax2 - bugfix: for some $%#ing reason good=3 & fair=4, breaking the progression from fake(1) to excellent(5)
		if (num == 3 || num == 4)
			num = (num == 3) ? PF_RATING_FAIR : PF_RATING_GOOD;
	}
	return static_cast<_EnumPartFileRating>(num);	//Cax2 - if no ratings found, will return 0!

	EMULE_CATCH

	return PF_RATING_NONE;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	UpdateDisplayedInfo() causes the DownloadList ctrl to redisplay the item for this file if either
//	'force' is true or if the redisplay timer has expired.
void CPartFile::UpdateDisplayedInfo(boolean force)
{
	EMULE_TRY

	g_eMuleApp.m_pDownloadList->UpdateFile(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	enkeyDEV(kei-kun) -hostname sources in ED2K superLink-
//v- eklmn: support Source Exchange v2
void CPartFile::AddClientSource(CClientSource* source, int i_Source, bool exchanged, uint8 sourceexchangeversion)
{
	EMULE_TRY

//	check first if we are this source
#ifdef OLD_SOCKETS_ENABLED
	if (g_eMuleApp.m_pServerConnect->GetClientID() < 0x1000000 && g_eMuleApp.m_pServerConnect->IsConnected())
	{
		if ((g_eMuleApp.m_pServerConnect->GetClientID() == source->sourceID)
			&& g_eMuleApp.m_pServerConnect->GetCurrentServer()->GetIP() == source->serverIP)
			return ;
	//	although we are currently having a LowID, we could have had a HighID before, which
	//	that client is sending us now! seems unlikely ... it happend!
		if (source->sourceID == g_eMuleApp.m_pServerConnect->GetLocalIP())
			return ;
	}
	else if (g_eMuleApp.m_pServerConnect->GetClientID() == source->sourceID)
		return ;
//	Modified by Tarod -use LowID source only if source is in our same server
	else if (source->sourceID < 0x1000000 && !g_eMuleApp.m_pServerConnect->IsLocalServer(source->serverIP, source->serverPort))
		return ;
#endif //OLD_SOCKETS_ENABLED

	if ( g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFile() > GetSourceCount()
	     && i_Source < (int)g_eMuleApp.m_pGlobPrefs->GetMaxSourcePerFileSoft() ) // DonGato - Using Max Sources per File (soft)	//Cax2 - used new call
	{
		CUpDownClient* pNewSource = g_eMuleApp.m_pDownloadQueue->CheckAndAddSource(this,source->sourceID,source->sourcePort, source->serverIP,source->serverPort);
		if (exchanged && pNewSource)
		{
		//Cax2 - exchanged sources are named now
			pNewSource->SetUserName(GetResString(IDS_EXCHANGEDSOURCE));
		//	check exchange version
			if (sourceexchangeversion > 1)
				pNewSource->SetUserHash(source->achUserHash);
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::AddClientSources(CTypedPtrList<CPtrList, CClientSource*>* sources)
{
	EMULE_TRY

	POSITION pos;
	pos = sources->GetHeadPosition();
	while (pos != NULL)
	{
		CClientSource * source = (CClientSource*)sources->GetAt(pos);

		if (source->sourceType != ED2KLINK_SOURCE_IP)
		{
			sources->GetNext(pos);
			continue;
		}

		POSITION removePos = pos;
		sources->GetNext(pos);
		sources->RemoveAt(removePos);

		AddClientSource(source);
		delete source;
	}
	g_eMuleApp.m_pDownloadQueue->AddClientHostnameToResolve(sources);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	enkeyDEV(Ottavio84) -Create source ed2k from downloads-
CString CPartFile::CreateED2KSourceLink(uint32 expirein, int sourcenum, bool bLog /*=true*/)
{
	EMULE_TRY

	CTypedPtrList<CPtrList, CUpDownClient*> srcstoadd;
	POSITION pos1, pos2;
	for (int sl = 0;sl < SOURCESSLOTS;sl++)
		if (!m_srcLists[sl].IsEmpty())
			for (pos1 = m_srcLists[sl].GetHeadPosition();pos1 != NULL;)
			{
				CUpDownClient* pSource = m_srcLists[sl].GetNext(pos1);
				if (pSource->HasLowID())
					continue;
				if (srcstoadd.IsEmpty())
				{
					srcstoadd.AddHead(pSource);
					continue;
				}
				if ((sourcenum == 0) || (srcstoadd.GetCount() < sourcenum) || (pSource->GetAvailablePartCount() > srcstoadd.GetTail()->GetAvailablePartCount()))
				{
					if ((sourcenum != 0) && (srcstoadd.GetCount() == sourcenum))
						srcstoadd.RemoveTail();
					ASSERT(srcstoadd.GetCount() < sourcenum);
					bool bInserted = false;
					for (pos2 = srcstoadd.GetTailPosition();pos2 != 0;)
					{
						CUpDownClient * cur_srctoadd = srcstoadd.GetPrev(pos2);
						if (cur_srctoadd->GetAvailablePartCount() > pSource->GetAvailablePartCount())
						{
							srcstoadd.InsertAfter(pos2, pSource);
							bInserted = true;
							break;
						}
					}
					if (!bInserted)
						srcstoadd.AddHead(pSource);
				}
			}
#ifdef OLD_SOCKETS_ENABLED
	if (srcstoadd.IsEmpty() && (!g_eMuleApp.m_pServerConnect->IsConnected() || g_eMuleApp.m_pServerConnect->IsLowID()))
	{
		AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_SOURCELINKFAILED));
		return CString("");
	}
#endif //OLD_SOCKETS_ENABLED
	CString strExpiration;
	if (expirein == 0)
		strExpiration = "";
	else
	{
		COleDateTime expirationDate = COleDateTime::GetCurrentTime();
		COleDateTimeSpan timediff;
		timediff.SetDateTimeSpan(expirein, 0, 0, 0);
		expirationDate += timediff;
		strExpiration.Format(_T("@%02i%02i%02i"), (expirationDate.GetYear() % 100), expirationDate.GetMonth(), expirationDate.GetDay());
	}
	uint32 dwID = 0;
#ifdef OLD_SOCKETS_ENABLED

	dwID = g_eMuleApp.m_pServerConnect->GetClientID();
#endif //OLD_SOCKETS_ENABLED

	CString strLink;
	strLink.Format( _T("ed2k://|file|%s|%d|%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x|/|sources%s"),
	                g_eMuleApp.StripInvalidFilenameChars(GetFileName(), false),  // spaces to dots
	                GetFileSize(),
	                GetFileHash()[0], GetFileHash()[1], GetFileHash()[2], GetFileHash()[3], GetFileHash()[4], GetFileHash()[5], GetFileHash()[6], GetFileHash()[7],
	                GetFileHash()[8], GetFileHash()[9], GetFileHash()[10], GetFileHash()[11], GetFileHash()[12], GetFileHash()[13], GetFileHash()[14], GetFileHash()[15],
	                strExpiration );
	CString newsource;
	if (dwID >= 0x1000000)
	{
		newsource.Format(_T(",%i.%i.%i.%i:%i"), (uint8)dwID, (uint8)(dwID >> 8), (uint8)(dwID >> 16), (uint8)(dwID >> 24), g_eMuleApp.m_pGlobPrefs->GetPort());
		strLink.Append(newsource);
	}
	if (bLog)
	{
		AddDebugLogLine(false, _T("Generated ed2k link with %i sources for file '%s'"), (int)srcstoadd.GetCount(), GetFileName());
		while (!srcstoadd.IsEmpty())
		{
			CUpDownClient * pSource = srcstoadd.RemoveHead();
			AddDebugLogLine(false, RGB_LOG_DIMMED + _T("| User: %s - Available part count: %i"), pSource->GetUserName(), (int)pSource->GetAvailablePartCount());
			newsource.Format(_T(",%s:%i"), pSource->GetFullIP(), pSource->GetUserPort());
			strLink.Append(newsource);
		}
	}
	strLink.Append(_T("|/"));
	return strLink;

	EMULE_CATCH

	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	START - enkeyDEV(Ottavio84) -ChangeDir-
void CPartFile::SetAlternativeOutputDir(CString* path)
{
	EMULE_TRY

	path->TrimRight(_T("\\"));
	m_strAlternativePath = *path;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetOutputDir() returns the appropriate destination directory path for this file.
CString CPartFile::GetOutputDir()
{
	EMULE_TRY

	CString	strRetPath = g_eMuleApp.m_pGlobPrefs->GetIncomingDir();
	CString	strCatPath;

	if (GetCatID() != 0)
		strCatPath = CCat::GetCatByID(GetCatID())->GetPath();

//	If no alternative path defined
	if (m_strAlternativePath.IsEmpty())
	{
	//	... and a category dir is defined...
		if (strCatPath != "")
			return strCatPath; // Return the category dir
		else
			return strRetPath;	// Return the incoming dir.
	}

//	MOREVIT - Why all the heavy duty shell call nonsense? You could replace all of this
//	with a single call to PathFileExists(). Somebody just thought they were more
//	clever than they were methinks.

//	We're going to use desktop folder "static" functions , we need an object
//	of IShellFolder type and DeskTop always exists
	IShellFolder	*pDesktop	= NULL;
	ITEMIDLIST	*pList	= NULL;
	CString	strDir;

//	Try to create DeskTop folder object
	if (SHGetDesktopFolder(&pDesktop) == NOERROR)
	{
		IShellFolder	* pFolder = NULL;
		ULONG	uAttrib	= SFGAO_FOLDER ; //we're going to look for the "folder" attribute

		strDir = m_strAlternativePath;
	//	"ParseDisplayName" uses UNICODE so BSTR will go
	//	We have to remember to release it
		BSTR	pData	= strDir.AllocSysString();

		if (pData)
		{
		//	'm_strAlternativePath' is always created without "\" but this was the way
		//	the original code was, maybe it was for a reason :) , so I am going to leave
		//	first check without adding "\" as well ...

		//	So here we convert "display name" (path string) into ID list used by Shell
			pDesktop->ParseDisplayName(NULL, NULL, pData, NULL, &pList, &uAttrib);
		//	Release the BSTR
			::SysFreeString(pData);
		//	If list we got is null then no such object exists
			if (!pList)
			{
			//	Add "\" to get folder name
				strDir += _T("\\");
			//	Allocate new search BSTR
				pData = strDir.AllocSysString();
				if (pData)
				{
				//	We're going to look for "folder" attribute
					uAttrib	= SFGAO_FOLDER ;
				//	We convert "display name" (path string) into ID list used by Shell
					pDesktop->ParseDisplayName(NULL, NULL, pData, NULL, &pList, &uAttrib);
				//	Release the BSTR
					::SysFreeString(pData);
				}
			}
		//	If we found the object and it's "folder" type
			if (pList && (uAttrib & SFGAO_FOLDER))
			{
				LPMALLOC	ppMalloc;

				if (SHGetMalloc(&ppMalloc) == NOERROR)
				{
					ppMalloc->Free(pList);
					ppMalloc->Release();
				}
			//
			//	If the destination directory is already shared, just return it.
				for ( POSITION pos = g_eMuleApp.m_pGlobPrefs->m_sharedDirList.GetHeadPosition();
				      pos != 0; )
				{
					if (g_eMuleApp.m_pGlobPrefs->m_sharedDirList.GetNext(pos) == strDir)
					{
						return m_strAlternativePath;
					}
				}
			//
			//	Otherwise, add it to the shared directory list and share the directory
				g_eMuleApp.m_pGlobPrefs->m_sharedDirList.AddTail(strDir);
				g_eMuleApp.m_pSharedFilesList->Reload();
				return m_strAlternativePath;
			}
		//	Just in case the object is not folder we have to release memory
			else if (pList)
			{
				LPMALLOC	ppMalloc;

				if (SHGetMalloc(&ppMalloc) == NOERROR)
				{
					ppMalloc->Free(pList);
					ppMalloc->Release();
				}
			}
		}
		pDesktop->Release();
	}

//	Default return
	strRetPath.TrimRight(_T("\\"));
	return strRetPath;

	EMULE_CATCH

	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CPartFile::GetDownloadFileInfo()
{
	CString compl;
	compl.Format(_T("%s/%s"), CastItoXBytes(GetCompletedSize()), CastItoXBytes(GetFileSize()));

	CString lsc;
	if (lastseencomplete == NULL)
		lsc = GetResString(IDS_NEVER);
	else
		lsc = LocalizeLastSeenComplete();

	CString lastprogr;
	if (GetTransferred() == 0 /*SyruS won't work: GetFileDate()==NULL*/)
		lastprogr = GetResString(IDS_NEVER);
	else
		lastprogr = LocalizeLastDownTransfer();

	double availability = 0.0;
	if (GetPartCount() != 0)
	{
		availability = GetAvailablePartCount() * 100 / GetPartCount();
	}

	CString timeleft;
	if (GetStatus() != PS_COMPLETING && GetStatus() != PS_COMPLETE)
	{
		sint32 restTime = GetTimeRemaining();
		timeleft.Format(_T("%s"), CastSecondsToHM(restTime));
	}
	else
	{
		timeleft = _T("-");
	}

	CString sRet;
	CString strHash = EncodeBase16(GetFileHash(), 16);
//	BavarianSnail - added detailed source infos
	CString compl2;
	compl2.Format(_T("%s\\%s"), GetTempDir(), GetPartMetFileName());

	sRet.Format( GetResString(IDS_DL_FILENAME) + _T(": %s (%s)\n\n%s\n\n")
	             + GetResString(IDS_FD_HASH) + _T(" %s\n")
	             + GetResString(IDS_PARTINFOS)
	             + GetResString(IDS_PARTINFOS2) + _T("\n%s\n%s\n%s\n")
	             + GetResString(IDS_PARTINFOS3),
	             GetFileName(), CastItoXBytes(GetFileSize()),
	             GetResString(IDS_STATUS) + _T(": ") + GetPartfileStatus(),
	             strHash,
	             compl2, GetPartCount(), GetResString(IDS_AVAIL), GetAvailablePartCount(), availability,
	             static_cast<int>(GetPercentCompleted()), compl, GetTransferringSrcCount(),
	             GetResString(IDS_LASTSEENCOMPL) + _T(" ") + lsc,
	             GetResString(IDS_FD_LASTCHANGE) + _T(" ") + lastprogr,
				 GetResString(IDS_DLCOL_REMAININGTIME) + _T(": ") + timeleft,
	             GetOnQueueSrcCount(),
	             GetQueueFullSrcCount(),
	             GetNoNeededPartsSrcCount(),
	             GetSrcA4AFCount(),
	             GetConnectingSrcCount() );

	return sRet;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Improved Tooltips
CString CPartFile::GetDownloadFileInfo4Tooltips()
{
	CString compl;
	compl.Format(_T("%s/%s"), CastItoXBytes(GetCompletedSize()), CastItoXBytes(GetFileSize()));

	CString lsc;
	if (lastseencomplete == NULL)
		lsc = GetResString(IDS_NEVER);
	else
		lsc = LocalizeLastSeenComplete();

	CString lastprogr;
	if (GetTransferred() == 0)
		lastprogr = GetResString(IDS_NEVER);
	else
		lastprogr = LocalizeLastDownTransfer();

	double availability = 0.0;
	if (GetPartCount() != 0)
	{
		availability = GetAvailablePartCount() * 100 / GetPartCount();
	}

	CString sTransfSrcCount;
	if (GetTransferringSrcCount() > 0)
		sTransfSrcCount.Format(GetResString(IDS_TT_PARTINFOS3), GetTransferringSrcCount());

	CString strHash = EncodeBase16(GetFileHash(), 16);
	;
	CTime t(GetFileDate());

	CString sRet;
	sRet.Format( GetResString(IDS_TT_FILENAME)
	             + GetResString(IDS_TT_PARTINFOS2)
	             + GetResString(IDS_TT_STATUS)
	             + GetResString(IDS_TT_HASH)
	             + GetResString(IDS_TT_PARTINFOS)
	             + GetResString(IDS_TT_PARTINFOS4)
	             + GetResString(IDS_TT_LASTSEENCOMPL)
	             + GetResString(IDS_TT_LASTCHANGE)
	             + GetResString(IDS_TT_STATUS2),
	             GetFileName(), CastItoXBytes(GetFileSize()), GetFileSize(), GetResString(IDS_BYTES),
	             static_cast<int>(GetPercentCompleted()), compl, sTransfSrcCount,
	             GetPartfileStatus(),
	             strHash,
	             GetPartMetFileName(), GetPartCount(), GetResString(IDS_AVAIL), GetAvailablePartCount(), availability,
	             lsc, lastprogr,
	             GetOnQueueSrcCount(),
	             GetQueueFullSrcCount(),
	             GetNoNeededPartsSrcCount(),
	             GetSrcA4AFCount(),
	             GetConnectingSrcCount() );

	return sRet;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Ornis: Creating progressive presentation of the partfilestatuses - for webdisplay
CString CPartFile::GetProgressString(uint16 size)
{
	EMULE_TRY

	char crProgress = '0'; //green
	char crHave = '1';	// black
	char crPending = '2';	// yellow
	char crMissing = '3';  // red
	char crWaiting[6]; // differents blue
	crWaiting[0] = '4'; // blue few source
	crWaiting[1] = '5';
	crWaiting[2] = '6';
	crWaiting[3] = '7';
	crWaiting[4] = '8';
	crWaiting[5] = '9';

	CString my_ChunkBar = _T("");
	for (uint16 i = 0;i <= size + 1;i++)
		my_ChunkBar.AppendChar(crHave);	// one more for safety

	uint32 m_nFileSize = GetFileSize();
	double unit = static_cast<double>(size) / static_cast<double>(m_nFileSize);

	POSITION pos;

	if (GetStatus() == PS_COMPLETE || GetStatus() == PS_COMPLETING)
	{
		CharFillRange(&my_ChunkBar, 0, static_cast<double>(m_nFileSize) * unit, crProgress);
	}
	else

	//	red gaps
		for (pos = gaplist.GetHeadPosition();pos != 0;)
		{
			Gap_Struct * cur_gap = gaplist.GetNext(pos);
			bool gapdone = false;
			uint32 gapstart = cur_gap->m_dwStartOffset;
			uint32 gapend = cur_gap->m_dwEndOffset;
			for (uint32 i = 0; i < GetPartCount(); i++)
			{
				if (gapstart >= i * PARTSIZE && gapstart <= (i + 1) * PARTSIZE)
				{ // is in this part?
					if (gapend <= (i + 1) * PARTSIZE)
						gapdone = true;
					else
					{
						gapend = (i + 1) * PARTSIZE; // and next part
					}
				//	paint
					uint8 color;
					if (m_srcPartFrequencies.GetCount() >= (INT_PTR)i && m_srcPartFrequencies[static_cast<uint16>(i)])  // frequency?
						color = m_srcPartFrequencies[static_cast<uint16>(i)] < 10 ? crWaiting[m_srcPartFrequencies[static_cast<uint16>(i)] / 2] : crWaiting[5];
					else
						color = crMissing;

					CharFillRange(&my_ChunkBar, static_cast<double>(gapstart) * unit, static_cast<double>(gapend) * unit + 1, color);

					if (gapdone) // finished?
						break;
					else
					{
						gapstart = gapend;
						gapend = cur_gap->m_dwEndOffset;
					}
				}
			}
		}

//	yellow pending parts
	for (pos = requestedblocks_list.GetHeadPosition();pos != 0;)
	{
		Requested_Block_Struct * block = requestedblocks_list.GetNext(pos);
		CharFillRange(&my_ChunkBar, static_cast<double>(block->m_dwStartOffset + block->m_dwBytesTransferred) * unit, static_cast<double>(block->m_dwEndOffset) * unit, crPending);
	}

	return my_ChunkBar;

	EMULE_CATCH

	return _T("");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::CharFillRange(CString* buffer, double start, double dblEnd, char color)
{
	EMULE_TRY
	for (uint32 i = start;i <= dblEnd;i++)
		buffer->SetAt(i, color);
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//vcf:- eklmn: A4AF management
void CPartFile::DownloadAllA4AF()
{
	EMULE_TRY

//AddDebugLogLine(false, _T("Total %u A4AF sources before swap"), m_A4AFsrcList.GetCount());

//	Make a list of sources
	CArray<CUpDownClient*, CUpDownClient*> SourcesList;
	for (POSITION pos = m_A4AFsrcList.GetHeadPosition(); pos != 0;)
	{
		SourcesList.Add(m_A4AFsrcList.GetNext(pos));
	}

//	Actually handover
	for (int i = 0; i < SourcesList.GetCount(); i++)
		SourcesList[i]->SwapToAnotherFile(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//v- eklmn: download fix
void CPartFile::AddDownloadingSource(CUpDownClient* client)
{
	EMULE_TRY
	POSITION pos = m_downloadingSourceList.Find(client);
	if (pos == NULL)
	{
		m_downloadingSourceList.AddTail(client);
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//v- eklmn: download fix
void CPartFile::RemoveDownloadingSource(CUpDownClient* client)
{
	EMULE_TRY
	POSITION pos = m_downloadingSourceList.Find(client);
	if (pos != NULL)
	{
		m_downloadingSourceList.RemoveAt(pos);
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint8* CPartFile::MMCreatePartStatus()
{
	EMULE_TRY
//	create partstatus + info in mobilemule protocol specs
//	result needs to be deleted[] | slow, but not timecritical
	uint8 * result = new uint8[GetPartCount() + 1];
	for (int i = 0; i < GetPartCount(); i++)
	{
		result[i] = 0;
		if (IsComplete(i * PARTSIZE, ((i + 1) * PARTSIZE) - 1))
		{
			result[i] = 1;
			continue;
		}
		else
		{
			if (IsComplete(i * PARTSIZE + (0 * (PARTSIZE / 3)), ((i * PARTSIZE) + (1 * (PARTSIZE / 3))) - 1))
				result[i] += 2;
			if (IsComplete(i * PARTSIZE + (1 * (PARTSIZE / 3)), ((i * PARTSIZE) + (2 * (PARTSIZE / 3))) - 1))
				result[i] += 4;
			if (IsComplete(i * PARTSIZE + (2 * (PARTSIZE / 3)), ((i * PARTSIZE) + (3 * (PARTSIZE / 3))) - 1))
				result[i] += 8;
			uint8 freq = (uint8)m_srcPartFrequencies[i];
			if (freq > 44)
				freq = 44;
			freq = ceil(static_cast<double>(freq) / 3.0);
			freq <<= 4;
			result[i] += freq;
		}

	}
	return result;
	EMULE_CATCH
	return NULL;
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	#zegzav:pastcomment
void CPartFile::AddPastComment(CUpDownClient *pClient)
{
	EMULE_TRY
	if ((!pClient) || (pClient->m_pReqPartFile != this))
		return ;

	if ((pClient->GetFileComment().IsEmpty()) && (pClient->GetFileRating() == PF_RATING_NONE))
		return ;

	m_pastCommentList.Add(pClient->GetUserHash(), pClient->GetUserName(), GetFileName(), pClient->GetFileComment(), pClient->GetFileRating());
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	#zegzav:pastcomment
void CPartFile::RemovePastComment(CUpDownClient *pClient, bool bRestore)
{
	EMULE_TRY
	if (!pClient)
		return ;

	if (bRestore)
	{
		CString	strPastComment;
		EnumPartFileRating	ePastRating;

		if ( (m_pastCommentList.GetCommentRating(pClient->GetUserHash(), &strPastComment, &ePastRating))
		     && (pClient->GetFileComment().IsEmpty())
		     && (pClient->GetFileRating() == PF_RATING_NONE) )
		{
		//	Restore past comment
			pClient->SetFileComment(strPastComment);
			pClient->SetFileRating(ePastRating);
		}
	}

	m_pastCommentList.Remove(pClient->GetUserHash());
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
EnumCategories CPartFile::GetCatID() const
{
//	If catID is valid... (valid range is 1..CAT_PREDEFINED-1 but a
//	real check would have to scan the category map. This isn't the
//	best place to do that.
#ifndef VS2002 // In VS2002 the const cast operator causes an ambiguity
	if (static_cast<const _EnumCategories>(m_eCategoryID) >= CAT_PREDEFINED)
#else
	if (const_cast<CPartFile*>(this)->m_eCategoryID >= CAT_PREDEFINED)
#endif
		return CAT_NONE;
	else
		return m_eCategoryID;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetCatID(EnumCategories eCatID)
{
	m_eCategoryID = eCatID;

//	set new prio
	if (IsPartFile())
	{
		int	iPriority = 0;

		if (eCatID != CAT_NONE)
		{
			iPriority = CCat::GetCatByID(GetCatID())->GetPriority();
		}
		switch (iPriority)
		{
			case 0:
			{
				break;
			}
			case 1:
			{
				SetAutoPriority(false);
				SetPriority(PR_LOW);
				break;
			}
			case 2:
			{
				SetAutoPriority(false);
				SetPriority(PR_NORMAL);
				break;
			}
			case 3:
			{
				SetAutoPriority(false);
				SetPriority(PR_HIGH);
				break;
			}
			case 4:
			{
				SetAutoPriority(true);
				SetPriority(PR_HIGH);
				break;
			}
		}

		if (IsPartFile())
			SavePartFile();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CPartFile::SetStatus(EnumPartFileStatuses in)
{
	m_eStatus = in;

	EnumCategories	eCurCat = g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.GetCurTabCat();

	if (eCurCat >= CAT_PREDEFINED)
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.ChangeCategoryByID(eCurCat);
	else
		UpdateDisplayedInfo(true);

	if (g_eMuleApp.m_pGlobPrefs->ShowCatTabInfo())
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateCatTabTitles();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetRealFileSize() returns real file size on disk with respect to NTFS compression and/or NTFS sparse files
uint64 CPartFile::GetRealFileSize()
{
	return ::GetDiskFileSize(GetFilePath());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsStalled() performs obaldin's stall check
bool CPartFile::IsStalled() const
{
	bool	bIsStalled = false;

	CTimeSpan	ts = CTime::GetCurrentTime() - GetLastDownTransfer();

//	If WARN_PERIOD_OF_NO_PROGRESS days have gone by without a single download...
//	(??? or maybe sooner if there are no needed parts available?)
	if (ts.GetTotalHours() >= 24 * WARN_PERIOD_OF_NO_PROGRESS)
		bIsStalled = true;

	return bIsStalled;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	AllocateNeededSpace() is the preallocate function used to allocate the space needed by a file.
//	It uses a thread to let eMule still work and not drop sources while generating the file.
void CPartFile::AllocateNeededSpace()
{
	HANDLE	lTh;
	DWORD	lThID;
	m_bIsPreallocated = true;

	PAThreadParam *tp = new PAThreadParam;

	tp->cPartFile = &m_hPartFile;
	tp->iPosition = m_lFileSize - 1;

	lTh = CreateThread(NULL, 0L, AllocateNeededSpaceProc, tp, 0, &lThID);
	CloseHandle(lTh);

	SaveSettingsFile();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	AllocateNeededSpaceProc() is the thread created by AllocateNeededSpace().
//	It receives a pointer to 'm_hPartFile' and the filesize-1 (where it has to write the bogus data to allocate space.
DWORD WINAPI CPartFile::AllocateNeededSpaceProc(LPVOID lpParameter)
{
	PAThreadParam * tp = (PAThreadParam*)lpParameter;

	tp->cPartFile->Seek(tp->iPosition, CFile::begin);
	tp->cPartFile->Write("1", 1);
	tp->cPartFile->Flush();

	delete tp;

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	IsLastBlockComplete() returns status of last block of file (if downloaded the file was already allocated).
bool CPartFile::IsLastBlockComplete()
{
	long a = m_lFileSize % EMBLOCKSIZE;
	if (a == 0)
		a = EMBLOCKSIZE;

	return IsComplete(m_lFileSize - a, m_lFileSize - 1);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
void CPartFile::SetFileUpload (bool bActive)
{
	if (bActive)
	{
		InterlockedIncrement(&m_dwNumUploadingClients);
		ResetEvent(m_hCompletionEvent);
	}
	else
	{
		if (InterlockedDecrement(&m_dwNumUploadingClients) == 0)
			SetEvent(m_hCompletionEvent);
	}
}
*/
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
