//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "ServerList.h"
#include "emule.h"
#include "HttpDownloadDlg.h"
#include "safefile.h"
#include "otherfunctions.h"
#include <io.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CServerList::CServerList(CPreferences *in_prefs)
{
	m_iNumServers = 0;
	m_uServerListVersion = 0;
	m_iServerPos = 0;
	searchserverpos = 0;
	statserverpos = 0;
	m_pPrefs = in_prefs;
	udp_timer = 0;
	delservercount = 0;
	m_nLastSaved = ::GetTickCount();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CServerList::AutoUpdate()
{
	g_eMuleApp.m_pGlobPrefs->LoadServerlistAddresses() ; // Added by Tarod
	if (m_pPrefs->m_addressesList.IsEmpty())
	{
		MessageBox(0, GetResString(IDS_ERR_EMPTYADRESSESDAT), GetResString(IDS_ERR_EMPTYADRESSESDAT_TITLE), 64);
		return;
	}
	bool bDownloaded = false;
	CString servermet;
	CString oldservermet;
	CString strURLToDownload;
	CSingleLock AccessLock(&m_ServerMetMutex, TRUE);
	servermet.Format(_T("%sserver.met"), m_pPrefs->GetConfigDir());
	oldservermet.Format(_T("%sserver_met.old"), m_pPrefs->GetConfigDir());
	_tremove(oldservermet);
	_trename(servermet, oldservermet);

	POSITION Pos = m_pPrefs->m_addressesList.GetHeadPosition();
	while ((!bDownloaded) && (Pos != NULL))
	{
		CHttpDownloadDlg dlgDownload;
		strURLToDownload = m_pPrefs->m_addressesList.GetNext(Pos);
		dlgDownload.m_sURLToDownload = strURLToDownload.GetBuffer();
		dlgDownload.m_sFileToDownloadInto = servermet;
		if (dlgDownload.DoModal() == IDOK)
		{
			_tremove(oldservermet);
			bDownloaded = true;
		}
		else
			AddLogLine(true, RGB_LOG_WARNING + GetResString(IDS_ERR_FAILEDDOWNLOADMET), strURLToDownload.GetBuffer());
	}
	if (!bDownloaded)
		_trename(oldservermet, servermet);
}

bool CServerList::Init()
{
//	auto update the list by using an url
	if (m_pPrefs->AutoServerlist())
		AutoUpdate();
//	Load Metfile
	CString strPath = m_pPrefs->GetConfigDir() + CString(_T("server.met"));
	CSingleLock AccessLock(&m_ServerMetMutex, TRUE);
//	don't delete existing servers, because we might have added one by command line
	bool bRes = AddServerMetToList(strPath, false);
//	insert static servers from textfile
	strPath = m_pPrefs->GetConfigDir() + CString(_T("staticservers.dat"));
	AddServersFromTextFile(strPath);

//	FakeCheck
	if (m_pPrefs->IsUpdateFakeStartupEnabled())
		g_eMuleApp.m_pFakeCheck->DownloadFakeList();

	return bRes;
}

bool CServerList::AddServerMetToList(const CString& strFile, bool merge)
{
	// hide serverlist & stop redrawing until list will be filled
	g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.Hide();
	g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.SetRedraw(FALSE);

//	if (!merge)
//	{
//		g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.DeleteAllItems();
//		RemoveAllServers();
//	}
	CSafeFile servermet;
	try
	{
		if (!servermet.Open(strFile, CFile::modeRead | CFile::osSequentialScan))
		{
			if (!merge)
				AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_LOADSERVERMET));
			g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.SetRedraw(TRUE);
			g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.Show();
			return false;
		}
		servermet.Read(&m_uServerListVersion, 1);
		if (m_uServerListVersion != 0xE0 && m_uServerListVersion != MET_HEADER)
		{
			servermet.Close();
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_BADSERVERMETVERSION), m_uServerListVersion);
			g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.SetRedraw(TRUE);
			g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.Show();
			return false;
		}
		
		uint32 dwServerCount;
		servermet.Read(&dwServerCount, 4);

		ServerMet_Struct sCurSavingServer;
		uint32 dwAddedServerCount = 0;
		CString strListName;
		CServer* pNewServer = NULL;
		CServer* pExistingServer = NULL;
		
		for (uint32 j = 0; j != dwServerCount; j++)
		{
		//	get server
			servermet.Read(&sCurSavingServer, sizeof(ServerMet_Struct));
			try
			{
				pNewServer = new CServer(&sCurSavingServer);
			}
			catch (CMemoryException* error)
			{
				error->Delete();
				continue;
			}
			
		//add tags
			for (uint32 i = 0;i != sCurSavingServer.m_dwTagCount;i++)
				pNewServer->AddTagFromFile(servermet);
		//	set listname for server
			if (!pNewServer->GetListName())
			{
				strListName.Format(_T("Server %s"), pNewServer->GetAddress());
				pNewServer->SetListName(strListName);
			}
			
			if (!g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.AddServer(pNewServer, true, true))
			{
				pExistingServer = g_eMuleApp.m_pServerList->GetServerByAddress(pNewServer->GetAddress(), pNewServer->GetPort());
				if (pExistingServer)
				{
					pExistingServer->SetListName(pNewServer->GetListName());
					if (pNewServer->GetDescription())
						pExistingServer->SetDescription(pNewServer->GetDescription());
					g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pExistingServer);
				}
				safe_delete(pNewServer);
			}
			else
				dwAddedServerCount++;
		}

		if (!merge)
			AddLogLine(true, IDS_SERVERSFOUND, dwServerCount);
		else
			AddLogLine(true, IDS_SERVERSADDED, dwAddedServerCount, dwServerCount - dwAddedServerCount);
		servermet.Close();
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
		if (error->m_cause == CFileException::endOfFile)
			AddDebugLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_BADSERVERLIST));
		else
			AddDebugLogLine(true, RGB_LOG_ERROR + GetResString(IDS_ERR_FILEERROR_SERVERMET), GetErrorMessage(error));
		error->Delete();
	}
	g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.SetRedraw(TRUE);
	g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.Show();
	g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.ShowFilesCount();
	return true;
}

bool CServerList::AddServer(CServer* in_server)
{
	if (in_server == NULL)
		return false;

	if(in_server->GetDynIP() == "255.255.255.255")
		return false;

	if (g_eMuleApp.m_pGlobPrefs->FilterBadIPs())
		if (!IsGoodServerIP(in_server))
			return false;

	CServer* test_server = FindServerByAddress(*in_server);

	if (test_server)
	{
		test_server->ResetFailedCount();
		g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*test_server);
		return false;
	}

	m_serverList.AddTail(in_server);

	return true;
}

void CServerList::ServerStats()
{
#ifdef OLD_SOCKETS_ENABLED
	if (g_eMuleApp.m_pServerConnect->IsConnected() && m_serverList.GetCount() > 0)
	{
		CServer * ping_server = GetNextStatServer();
		CServer* test = ping_server;
		if (!ping_server)
			return;
		while (ping_server->GetLastPingedTime() != 0 && ((uint32)time(NULL) - ping_server->GetLastPingedTime()) < UDPSERVSTATREASKTIME)
		{
			ping_server = GetNextStatServer();
			if (ping_server == test)
				return;
		}
		if ( ping_server->GetFailedCount() >= g_eMuleApp.m_pGlobPrefs->GetDeadserverRetries() &&
		     g_eMuleApp.m_pGlobPrefs->DeadServer() &&
		     !ping_server->IsStaticMember() )
		{
			g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RemoveServer(ping_server);
			return;
		}
		Packet* packet = new Packet(OP_GLOBSERVSTATREQ, 4);
		srand((unsigned)time(NULL));
		uint32 c = 0x55AA0000 + (uint16)rand();
		ping_server->SetChallenge(c);
		memcpy2(packet->m_pcBuffer, &c, 4);
		ping_server->SetLastPinged(::GetTickCount());
		ping_server->SetLastPingedTime((uint32)time(NULL));
		ping_server->AddFailedCount();
		g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*ping_server);
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadServer(packet->m_dwSize);
		g_eMuleApp.m_pServerConnect->SendUDPPacket(packet, ping_server, true);
		ping_server->SetLastDescPingedCount(false);
		if (ping_server->GetLastDescPingedCount() < 2)
		{
			packet = new Packet(OP_SERVER_DESC_REQ);
			g_eMuleApp.m_pUploadQueue->AddUpDataOverheadServer(packet->m_dwSize);
			g_eMuleApp.m_pServerConnect->SendUDPPacket(packet, ping_server, true);
		}
		else
			ping_server->SetLastDescPingedCount(true);
	}
#endif //OLD_SOCKETS_ENABLED
}

/*filter Servers with invalid IP's / Port
0.*
10.*
172.16.0.0 - 172.31.255.255
192.168.0.0 - 192.168.255.255
127.*
*/
bool CServerList::IsGoodServerIP(CServer* in_server)
{
	if (!in_server->HasPublicAddress() && !in_server->HasDynIP())
		return false;

//	--- no4one : using filters from m_pIPFilter.dat for server filtering ---
//	netwolf: option for filtering servers listed in m_pIPFilter.dat and DynIP server
	if (g_eMuleApp.m_pGlobPrefs->FilterServersByIP())
		if (in_server->HasDynIP() || g_eMuleApp.m_pIPFilter->IsFiltered(in_server->GetIP()))
		{
			if (!g_eMuleApp.m_pGlobPrefs->GetCMNotLog())
				AddDebugLogLine(false, RGB_LOG_DIMMED + _T("Filtered Server: %s %s "), in_server->GetFullIP(), g_eMuleApp.m_pIPFilter->GetLastHit());

			if (!g_eMuleApp.m_pGlobPrefs->GetAddServersFromConn())
				//netwolf: remove filtered servers only if 'update from servers/clients' disabled (otherwise receive-remove loop)
				RemoveServer(in_server);
			return false;
		}
//	--- no4one end ---

	return true;
}

void CServerList::RemoveServer(CServer* out_server)
{
	POSITION remove_pos = m_serverList.Find(out_server);
	if (remove_pos)
	{
	//	reset UDP server
		if (g_eMuleApp.m_pDownloadQueue->pCurUDPServer == out_server)
			g_eMuleApp.m_pDownloadQueue->pCurUDPServer = GetNextServer(out_server);
	//	finally delete a server
		delete m_serverList.GetAt(remove_pos);
		m_serverList.RemoveAt(remove_pos);
		delservercount++;	//statistic value
	}
}

void CServerList::RemoveAllServers()
{
	POSITION pos1, pos2;
	for (pos1 = m_serverList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_serverList.GetNext(pos1);
		delete m_serverList.GetAt(pos2);
		m_serverList.RemoveAt(pos2);
		delservercount++;	//statistic value
	}
}

void CServerList::GetStatus (uint32 &total, uint32 &failed, uint32 &user, uint32 &file, uint32 &tuser, uint32 &tfile, double &occ)
{
	total = m_serverList.GetCount();
	failed = 0;
	user = 0;
	file = 0;
	tuser = 0;
	tfile = 0;
	occ = 0;
	uint32 m_dwMaxUsers = 0;
	uint32 tuserk = 0;

	CServer* curr;
	for (POSITION pos = m_serverList.GetHeadPosition(); pos != 0;m_serverList.GetNext(pos))
	{
		curr = (CServer*)m_serverList.GetAt(pos);
		if (curr->GetFailedCount())
			failed++;
		else
		{
			user += curr->GetNumUsers();
			file += curr->GetFiles();
		}
		tuser += curr->GetNumUsers();
		tfile += curr->GetFiles();

		if (curr->GetMaxUsers())
		{
			tuserk += curr->GetNumUsers(); // total users on servers with known maximum
			m_dwMaxUsers += curr->GetMaxUsers();
		}
	}
	if (m_dwMaxUsers > 0)
		occ = static_cast<double>(tuserk * 100.0) / m_dwMaxUsers;
}

void CServerList::GetUserFileStatus(uint32 &user, uint32 &file)
{
	user = 0;
	file = 0;
	CServer* curr;
	for (POSITION pos = m_serverList.GetHeadPosition(); pos != 0; m_serverList.GetNext(pos))
	{
		curr = (CServer*)m_serverList.GetAt(pos);
		if (!curr->GetFailedCount())
		{
			user += curr->GetNumUsers();
			file += curr->GetFiles();
		}
	}
}

CServerList::~CServerList()
{
	SaveServerMetToFile();
	POSITION pos1, pos2;
	for (pos1 = m_serverList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_serverList.GetNext(pos1);
		delete m_serverList.GetAt(pos2);
		m_serverList.RemoveAt(pos2);
	}
	if (udp_timer)
		KillTimer(0, udp_timer);
}

void CServerList::MoveServerDown(CServer* aServer)
{
	POSITION pos1, pos2;
	uint16 i = 0;
	for (pos1 = m_serverList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_serverList.GetNext(pos1);
		CServer* cur_server = m_serverList.GetAt(pos2);
		if (cur_server == aServer)
		{
			m_serverList.AddTail(cur_server);
			m_serverList.RemoveAt(pos2);
			return;
		}
		i++;
		if (i == m_serverList.GetCount())
			break;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Sort() reorders the server list by priority.
void CServerList::Sort()
{
	POSITION	pos1, pos2;
	uint16	i = 0;

//	For each server in the list...
	for (pos1 = m_serverList.GetHeadPosition(); (pos2 = pos1) != NULL;)
	{
		m_serverList.GetNext(pos1);

		CServer	*pServer = m_serverList.GetAt(pos2);

	//	If the server is high priority, stick it at the head of the list.
		if (pServer->GetPreferences() == CServer::SERVERPRIORITY_HIGH)
		{
			m_serverList.AddHead(pServer);
			m_serverList.RemoveAt(pos2);
		}
	//	If the server is low priority, stick it at the tail of the list.
		else if (pServer->GetPreferences() == CServer::SERVERPRIORITY_LOW)
		{
			m_serverList.AddTail(pServer);
			m_serverList.RemoveAt(pos2);
		}
		i++;
	//	If we're at the end of the list...
		if (i == m_serverList.GetCount())
			break;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetNextServer() returns actually pointed server & move pointer to the next one in the list or NULL
//  if the list pointer is past the end of the list. (Right now there is no usability checking).
CServer *CServerList::GetNextServer()
{
	CServer*		pNextServer = NULL;
	POSITION		posIndex;

	if (m_serverList.IsEmpty())
		return NULL;
	
	const uint32	dwServerCounter = static_cast<uint32>(m_serverList.GetCount());

	// If we're at the end or past the end of the list 
	// move pointer to begin of the list & return NULL as server
	if (m_iServerPos >= dwServerCounter)
	{
		m_iServerPos = 0;
		return NULL;
	}

	// find next server 
	while (pNextServer == NULL && m_iServerPos < dwServerCounter)
	{
		posIndex = m_serverList.FindIndex(m_iServerPos);
		// Check if search position is still valid (could be corrupted by server delete operation)
		if (posIndex == NULL)
		{
			posIndex = m_serverList.GetHeadPosition();
			m_iServerPos = 0;
		}

		// get actuall server
		pNextServer = m_serverList.GetAt(posIndex);
		// move the pointer to the next one
		m_iServerPos++;

	//	TODO: Add more option to filter bad server like min ping, min users etc
	}
	
	return pNextServer;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CServer* CServerList::GetNextSearchServer()
{
	CServer * nextserver = NULL;
	uint32 i = 0;
	while (!nextserver && i != m_serverList.GetCount())
	{
		POSITION posIndex = m_serverList.FindIndex(searchserverpos);
		if (posIndex == NULL)
		{	// check if search position is still valid (could be corrupted by server delete operation)
			posIndex = m_serverList.GetHeadPosition();
			searchserverpos = 0;
		}
		nextserver = m_serverList.GetAt(posIndex);
		searchserverpos++;
		i++;
		if (searchserverpos == m_serverList.GetCount())
			searchserverpos = 0;
	}
	return nextserver;
}

CServer* CServerList::GetNextStatServer()
{
	CServer * nextserver = NULL;
	uint32 i = 0;
	while (!nextserver && i != m_serverList.GetCount())
	{
		POSITION posIndex = m_serverList.FindIndex(statserverpos);
		if (posIndex == NULL)	// check if search position is still valid (could be corrupted by server delete operation)
		{
			posIndex = m_serverList.GetHeadPosition();
			statserverpos = 0;
		}
		nextserver = m_serverList.GetAt(posIndex);
		statserverpos++;
		i++;
		if (statserverpos == m_serverList.GetCount())
			statserverpos = 0;
	}
	return nextserver;
}

bool CServerList::BroadCastPacket(Packet* packet)
{ // unused atm . but might be useful later
	if (udp_timer)
		return false;
	udp_timer = SetTimer(0, 4322, UDPSEARCHSPEED, CServerList::UDPTimerProc);
	broadcastpos = m_serverList.GetHeadPosition();
	broadcastpacket = packet;
	return true;
}

void CServerList::SendNextPacket()
{
#ifdef OLD_SOCKETS_ENABLED
	if (g_eMuleApp.m_pListenSocket->TooManySockets())
	{
		KillTimer(0, udp_timer);
		udp_timer = 0;
		delete broadcastpacket;
		return;
	}
#endif //OLD_SOCKETS_ENABLED

	if (broadcastpos)
	{
		CServer * cur_server = m_serverList.GetAt(broadcastpos);
#ifdef OLD_SOCKETS_ENABLED

		if (cur_server != g_eMuleApp.m_pServerConnect->GetCurrentServer())
			g_eMuleApp.m_pServerConnect->SendUDPPacket(broadcastpacket, cur_server, false);
#endif //OLD_SOCKETS_ENABLED

		m_serverList.GetNext(broadcastpos);
	}
	else
	{
		KillTimer(0, udp_timer);
		udp_timer = 0;
		delete broadcastpacket;

	}
}

void CServerList::CancelUDPBroadcast()
{
	if (udp_timer)
	{
		KillTimer(0, udp_timer);
		udp_timer = 0;
		delete broadcastpacket;
	}
}

void CALLBACK CServerList::UDPTimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
	EMULE_TRY
	g_eMuleApp.m_pServerList->SendNextPacket();
	EMULE_CATCH
}

CServer* CServerList::GetNextServer(CServer* lastserver)
{
	if (m_serverList.IsEmpty())
		return 0;
	if (!lastserver)
		return m_serverList.GetHead();
	POSITION pos = m_serverList.Find(lastserver);
	if (!pos)
	{
		TRACE("Error: CServerList::GetNextServer");
		return m_serverList.GetHead();
	}
	m_serverList.GetNext(pos);
	if (!pos)
		return NULL;
	else
		return m_serverList.GetAt(pos);
}

CServer* CServerList::FindServerByAddress(const CServer& in_server)
{
	for (POSITION pos = m_serverList.GetHeadPosition();pos != 0;m_serverList.GetNext(pos))
	{
		CServer * s = m_serverList.GetAt(pos); // i_a: small speed optimization
		if (s->HasSameAddress(in_server))
			return s;
	}
	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetServerByAddress() searches the server list for a server with stringified IP address 'strAddress'
//	and port 'nPort'. If none is found, NULL is returned.
CServer *CServerList::GetServerByAddress(const CString& strAddress, uint16 nPort)
{
//
//	Perform a linear search through the server list for a Server with the specified IP and port.
	for (POSITION pos = m_serverList.GetHeadPosition(); pos != 0; m_serverList.GetNext(pos))
	{
		CServer	* pServer = m_serverList.GetAt(pos);

		if ((nPort == pServer->GetPort() || nPort == 0) && pServer->GetAddress() == strAddress)
			return pServer;
	}
	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CServer* CServerList::GetServerByIP(uint32 nIP, uint16 port)
{
	for (POSITION pos = m_serverList.GetHeadPosition();pos != 0;)
	{
		CServer* s = m_serverList.GetNext(pos);
		if (s->GetIP() == nIP && (port == 0 || s->GetPort() == port))
			return s;
	}
	return NULL;
}

bool CServerList::SaveServerMetToFile()
{
	CSingleLock AccessLock(&m_ServerMetMutex, TRUE);
	CString strNewServerMet;
	strNewServerMet.Format(_T("%sserver.met.new"), m_pPrefs->GetConfigDir());
//	AddLogLine( false, "%s", strNewServerMet );
	try
	{
		FILE* servermet = _tfopen(strNewServerMet, _T("wb"));
		if (!servermet)
		{
			AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_SAVESERVERMET));
			return false;
		}
		m_uServerListVersion = 0xE0;
		fwrite(&m_uServerListVersion, 1, 1, servermet);
		
		uint32 dwServerCount = m_serverList.GetCount();
		fwrite(&dwServerCount, 4, 1, servermet);
		
		ServerMet_Struct sCurSavingServer;
		CServer* pNextServer;

		for (uint32 j = 0; j != dwServerCount; j++)
		{
			pNextServer = GetServerAt(j);
			sCurSavingServer.m_dwIP = pNextServer->GetIP();
			sCurSavingServer.m_uPort = pNextServer->GetPort();

			uint16 uTagcount = 10;
			
			if (!pNextServer->GetListName().IsEmpty())
				uTagcount++;
			if (!pNextServer->GetDynIP().IsEmpty())
				uTagcount++;
			if (!pNextServer->GetDescription().IsEmpty())
				uTagcount++;
			if (!pNextServer->GetVersion().IsEmpty())
				uTagcount++;
			
			sCurSavingServer.m_dwTagCount = uTagcount;
			fwrite(&sCurSavingServer, 1, sizeof(ServerMet_Struct), servermet);

			if (!pNextServer->GetListName().IsEmpty())
			{
				CServerTag servername(ST_SERVERNAME, pNextServer->GetListName());
				servername.WriteToFile(servermet);
			}
			if (!pNextServer->GetDynIP().IsEmpty())
			{
				CServerTag serverdynip(ST_DYNIP, pNextServer->GetDynIP());
				serverdynip.WriteToFile(servermet);
			}
			if (!pNextServer->GetDescription().IsEmpty())
			{
				CServerTag serverdesc(ST_DESCRIPTION, pNextServer->GetDescription());
				serverdesc.WriteToFile(servermet);
			}
			CServerTag serverfail(ST_FAIL, pNextServer->GetFailedCount());
			serverfail.WriteToFile(servermet);
		//	Modified by Tarod [txh Hein] - save server preferences using eDonkey correct values
			CServerTag serverpref(ST_PREFERENCE, pNextServer->eMule2ed2k(pNextServer->GetPreferences()));
			serverpref.WriteToFile(servermet);
			CServerTag serveruser(_T("users"), pNextServer->GetNumUsers());
			serveruser.WriteToFile(servermet);
			CServerTag serverfiles(_T("files"), pNextServer->GetFiles());
			serverfiles.WriteToFile(servermet);
			CServerTag serverping(ST_PING, pNextServer->GetPing());
			serverping.WriteToFile(servermet);
			CServerTag serverlastp(ST_LASTPING, pNextServer->GetLastPingedTime());
			serverlastp.WriteToFile(servermet);
			CServerTag servermaxusers(ST_MAXUSERS, pNextServer->GetMaxUsers());
			servermaxusers.WriteToFile(servermet);
			CServerTag serversoftfiles(ST_SOFTFILES, pNextServer->GetSoftMaxFiles());
			serversoftfiles.WriteToFile(servermet);
			CServerTag serverhardfiles(ST_HARDFILES, pNextServer->GetHardMaxFiles());
			serverhardfiles.WriteToFile(servermet);
			if (!pNextServer->GetVersion().IsEmpty())
			{
				CServerTag serverversion(ST_VERSION, pNextServer->GetVersion());
				serverversion.WriteToFile(servermet);
			}
			CServerTag serverUDPFlags(ST_UDPFLAGS, pNextServer->GetUDPFlags() );
			serverUDPFlags.WriteToFile(servermet);
		}
	//	eklmn: by shutdown force OS flush data direct to disk [SlugFiller]
		if (!g_eMuleApp.m_pdlgEmule->IsRunning())
			_commit(_fileno(servermet));
		fclose(servermet);
		
		CString strCurServerMet;
		CString strOldServerMet;
		strCurServerMet.Format(_T("%sserver.met"), m_pPrefs->GetConfigDir());
		strOldServerMet.Format(_T("%sserver_met.old"), m_pPrefs->GetConfigDir());
		_tremove(strOldServerMet);
		_trename(strCurServerMet, strOldServerMet);
		_trename(strNewServerMet, strCurServerMet);
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
		AddLogLine(false, RGB_LOG_ERROR + GetResString(IDS_ERR_SAVESERVERMET2));
		error->Delete();
		return false;
	}
	m_nLastSaved = ::GetTickCount();
	return true;
}

void CServerList::AddServersFromTextFile(CString strFilename)
{
	CString strLine;
	CStdioFile f;
	if (!f.Open(strFilename, CFile::modeRead | CFile::typeText))
		return;
	while (f.ReadString(strLine))
	{
	//	format is host:port,priority,Name
		if (strLine.GetLength() < 5)
			continue;
		if (strLine.GetAt(0) == '#' || strLine.GetAt(0) == '/')
			continue;

	//	fetch host
		int pos = strLine.Find(':');
		if (pos == -1)
		{
			pos = strLine.Find(','); 
			if (pos == -1)
				continue;
		}
		CString strHost = strLine.Left(pos);
		strLine = strLine.Mid(pos + 1);

	//	fetch  port
		pos = strLine.Find(',');
		if (pos == -1)
			continue;
		CString strPort = strLine.Left(pos);
		strLine = strLine.Mid(pos + 1);

	//	Barry - fetch priority
		pos = strLine.Find(',');
		int priority = PR_HIGH;
		if (pos == 1)
		{
			CString strPriority = strLine.Left(pos);
			try
			{
				priority = _tstoi(strPriority.GetBuffer(0));
				if ((priority < 0) || (priority > 2))
					priority = PR_HIGH;
			}
			catch (...)
			{}
			strLine = strLine.Mid(pos + 1);
		}

	//	fetch name
		CString strName = strLine;
		strName.Replace(_T("\r"), _T(""));
		strName.Replace(_T("\n"), _T(""));

	//	create server object and add it to the list
		CServer* nsrv = new CServer(_tstoi(strPort), strHost.GetBuffer());
		nsrv->SetListName(strName.GetBuffer());
		nsrv->SetIsStaticMember(true);
	//	Barry - Was always high
		nsrv->SetPreference((CServer::EnumServerPriority)priority);
		if (!g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.AddServer(nsrv, true))
		{
			delete nsrv;
			CServer* srvexisting = GetServerByAddress(strHost.GetBuffer(), _tstoi(strPort));
			if (srvexisting)
			{
				srvexisting->SetListName(strName.GetBuffer());
				srvexisting->SetIsStaticMember(true);
			//	Barry - Was always high
				srvexisting->SetPreference((CServer::EnumServerPriority)priority);
				if (g_eMuleApp.m_pdlgEmule->m_wndServer)
					g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*srvexisting);
			}

		}
	}

	f.Close();
}

void CServerList::Process()
{
	if (::GetTickCount() - m_nLastSaved > MIN2MS(17))
		this->SaveServerMetToFile();
}
