//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "SharedFileList.h"
#include "knownfilelist.h"
#include "packets.h"
#include "otherfunctions.h"
#include <time.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CSharedFileList::CSharedFileList(CPreferences *pPrefs, CServerConnect *pServerConnect, CKnownFileList *pKnownFileList)
{
	m_pPrefs = pPrefs;
	m_pServerConnect = pServerConnect;
	m_pKnownFileList = pKnownFileList;
	m_pOutput = NULL;

	CSingleLock		sLock(&m_mutexList,true);

	m_mapSharedFiles.InitHashTable(1024);

	sLock.Unlock();

	FindSharedFiles();
	m_pLanCast = new CLanCast(pPrefs);
	m_bDelayReload = false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CSharedFileList::~CSharedFileList()
{
	delete m_pLanCast;
	while (!m_waitingForHashList.IsEmpty())
	{
		delete m_waitingForHashList.RemoveTail();
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	NextLANBroadcast() broadcasts the next hash for LANCast
void CSharedFileList::NextLANBroadcast()
{
	if (g_eMuleApp.m_app_state != g_eMuleApp.APP_STATE_RUNNING)
		return;

	static POSITION		pos = 0;
	static CCKey		bufKey;

	CKnownFile		   *pKnownFile = NULL;

	CSingleLock		sLock(&m_mutexList,true);

	if (!m_mapSharedFiles.IsEmpty())
	{
		if (pos == 0)
		{
			pos = m_mapSharedFiles.GetStartPosition();
		}

		if (pos != 0)
		{
			m_mapSharedFiles.GetNextAssoc(pos,bufKey,pKnownFile);
			m_pLanCast->BroadcastHash(pKnownFile);
		}
	}
	sLock.Unlock();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	FindSharedFiles() adds all the files in the shared directories list to the known file list and the
//		shared files map. As a side-effect (naughty) it sets the 'm_iFilesToBeHashed' in the eMule Dialog
//		to the number of files in the hash waiting list.
void CSharedFileList::FindSharedFiles()
{
	BOOL			bEmpty;
	CSingleLock		sLock1(&m_mutexList,true); // list thread safe

	bEmpty	=	m_mapSharedFiles.IsEmpty();

	sLock1.Unlock();

	if (!bEmpty)
	{
		CCKey			bufKey;
		CKnownFile	   *pKnownFile = NULL;

		sLock1.Lock();

		POSITION		pos = m_mapSharedFiles.GetStartPosition();

		while (pos)
		{
			m_mapSharedFiles.GetNextAssoc(pos, bufKey, pKnownFile);
			pKnownFile->SetSharedFile(false);
		}
		m_mapSharedFiles.RemoveAll();

		sLock1.Unlock();

		g_eMuleApp.m_pDownloadQueue->AddPartFilesToShare();
	}
//	If the Incoming directory isn't shared yet...
	if (!m_pPrefs->m_sharedDirList.Find(m_pPrefs->GetIncomingDir()) && !m_pPrefs->m_sharedDirList.Find(m_pPrefs->GetIncomingDir()+'\\'))
		AddFilesFromDirectory(m_pPrefs->GetIncomingDir().GetBuffer());

	for (POSITION pos = m_pPrefs->m_sharedDirList.GetHeadPosition(); pos != NULL; m_pPrefs->m_sharedDirList.GetNext(pos))
		AddFilesFromDirectory(m_pPrefs->m_sharedDirList.GetAt(pos).GetBuffer());

	sLock1.Lock();

	if (m_waitingForHashList.IsEmpty())
	{
		AddLogLine(false, IDS_SHAREDFOUND,m_mapSharedFiles.GetCount());
	}
	else
	{
		g_eMuleApp.m_pdlgEmule->m_iFilesToBeHashed = m_waitingForHashList.GetCount();
		AddLogLine(false, IDS_SHAREDFOUNDHASHING,m_mapSharedFiles.GetCount(), m_waitingForHashList.GetCount());
	}
	sLock1.Unlock();
	BuildSharedVDirForList();
	HashNextFile();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	FindUnknownFile() checks to see if the spec'd file is already in the waiting for hash list.
bool CSharedFileList::FindUnknownFile(CString strDirectory, CString strFileName)
{
	bool		bFound = false;

	for (POSITION pos = m_waitingForHashList.GetHeadPosition(); pos != 0; m_waitingForHashList.GetNext(pos))
	{
		UnknownFile_Struct		*pUnknownFile = reinterpret_cast<UnknownFile_Struct*>(m_waitingForHashList.GetAt(pos));

		if (pUnknownFile->m_strDirectory == strDirectory && pUnknownFile->m_strFileName == strFileName)
		{
			bFound = true;
			break;
		}
	}

	return bFound;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	AddFilesFromDirectory() adds all the files in the spec'd directory to the known files list and the
//		shared files map.
void CSharedFileList::AddFilesFromDirectory(const CString &strDirectory)
{
	CString		strExtension;
	CFileFind	ff;
	CString		strSearchPath;

	strSearchPath.Format(_T("%s\\*"),strDirectory);

	bool		bEnd = !ff.FindFile(strSearchPath,0);

	if (bEnd)
		return;

	while (!bEnd)
	{
		bEnd = !ff.FindNextFile();
		if (ff.IsDirectory() || ff.IsDots() || ff.IsSystem() || ff.IsTemporary() || ff.GetLength() >= 0xFFFFFFFF)
			continue;
	//	Ultras: prevent to add the zero lenght files
		if (!ff.GetLength())
			continue;

		CTime		dwLastWriteTime;

		ff.GetLastWriteTime(dwLastWriteTime);
		strExtension = ff.GetFileName().Right(ff.GetFileName().GetLength() - ff.GetFileName().ReverseFind('.')).MakeLower();
	//	If the file doesn't have the extension of a type we don't want to share...
		if (!(strExtension==".part" || strExtension==".met" || strExtension==".settings" || strExtension==".bak" || strExtension==".txtsrc" || strExtension==".stats" || strExtension==".dir" || strExtension==".dll" || strExtension==".sys" || strExtension==".drv" || strExtension==".tmp" || strExtension==".log" || strExtension==".torrent" || strExtension==".getright" || strExtension==".jc!"))
		{
			uint32			time_tLocalTime = mktime(dwLastWriteTime.GetLocalTm());
			CKnownFile	   *pKnownFile = m_pKnownFileList->FindKnownFile(ff.GetFileName().GetBuffer(),time_tLocalTime,static_cast<uint32>(ff.GetLength()));

		//	If the file is already known...
			if (pKnownFile != NULL)
			{
				pKnownFile->SetPath(strDirectory);
				pKnownFile->SetFilePath(ff.GetFilePath());

				CSingleLock		sLock(&m_mutexList,true);

				pKnownFile->SetSharedFile(true);
				m_mapSharedFiles.SetAt(CCKey(pKnownFile->GetFileHash()),pKnownFile);
				sLock.Unlock();
			}
		//	If the file is not in the known list...
			else 
			{
			//	Start a thread to hash the new file
				UnknownFile_Struct	*pUnknownFile = NULL;
				try 
				{
				//	If the spec'd file isn't already in the list (avoid duplications due to overlapping
				//		share paths)
					if (!FindUnknownFile(strDirectory,ff.GetFileName()))
					{
						pUnknownFile = new UnknownFile_Struct;

						pUnknownFile->m_strDirectory = strDirectory;
						pUnknownFile->m_strFileName = ff.GetFileName();
						m_waitingForHashList.AddTail(pUnknownFile);
					}
				}
				catch(CException *err)
				{ 
					g_eMuleApp.m_pdlgEmule->AddBugReport(__FUNCTION__,__FILE__, __LINE__,GetErrorMessage(err));
					err->Delete();
					if (pUnknownFile)
						safe_delete(pUnknownFile);
				}
			}
		}
	}
	ff.Close();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	SafeAddKnownFile() shares the KnownFile 'pNewKnownFile'. If 'bOnlyAdd' is true, the new file
//		is added to the shared files map but not offered to the connected server.
void CSharedFileList::SafeAddKnownFile(CKnownFile *pNewKnownFile, bool bOnlyAdd)
{
//	Tell 'pNewKnownFile' that it's being shared now
	pNewKnownFile->SetSharedFile(true);
//
//	TODO: Check if the file is already known - only with another date
//
	CSingleLock		sLock(&m_mutexList,true);

//	Add 'pNewKnownFile' to the shared files map
	m_mapSharedFiles.SetAt(CCKey(pNewKnownFile->GetFileHash()),pNewKnownFile);

	sLock.Unlock();

//	If we're just to add the file and not advertise it, return
	if (bOnlyAdd)
		return;
//	Update the SharedFiles window if it's open
	if (m_pOutput != NULL)
		m_pOutput->ShowFile(pNewKnownFile);
	HashNextFile();
//	Offer new file to server
#ifdef OLD_SOCKETS_ENABLED
	if (!m_pServerConnect->IsConnected())
		return;
#else

#endif //OLD_SOCKETS_ENABLED
	pNewKnownFile->SetOfferedToServer(true);

#ifdef OLD_SOCKETS_ENABLED
	CMemFile		packetStream(100);

	uint32			dwFileCount = 1;

	packetStream.Write(&dwFileCount,4);
	WriteToOfferedFilePacket(*pNewKnownFile,packetStream);

	Packet		*pPacket = new Packet(&packetStream);

	pPacket->m_eOpcode = OP_OFFERFILES;	
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadServer(pPacket->m_dwSize);
	m_pServerConnect->SendPacket(pPacket,true);
#endif //OLD_SOCKETS_ENABLED
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	RemoveFile() removes the first occurrence of 'pKnownFile' in the shared file list.
void CSharedFileList::RemoveFile(CKnownFile *pKnownFile)
{
	pKnownFile->SetSharedFile(false);
	m_pOutput->RemoveFile(pKnownFile);

	CSingleLock		sLock(&m_mutexList,true);

	m_mapSharedFiles.RemoveKey(CCKey(pKnownFile->GetFileHash()));
	m_pKnownFileList->RemoveFile(pKnownFile);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSharedFileList::Reload(bool bSendToServer /*=true*/)
{
	if(bSendToServer && m_pServerConnect->IsConnected())
		AddDebugLogLine(false, _T("Reloading and sending to server shared files"));

//	If there are no shared files waiting to be hashed...
	if (m_waitingForHashList.IsEmpty())
	{
	    FindSharedFiles();
	    if (m_pOutput != NULL)
		    m_pOutput->ShowFileList(this);
	    if (bSendToServer && m_pServerConnect->IsConnected())
		    SendListToServer();
    }
	else
	{
		//eklmn: clear the waiting hash correct way
		while (!m_waitingForHashList.IsEmpty())
		{
			delete m_waitingForHashList.RemoveTail();
		}
		m_bDelayReload = true;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSharedFileList::SetOutputCtrl(CSharedFilesCtrl *pSharedFilesCtrl)
{
	m_pOutput = pSharedFilesCtrl;
	m_pOutput->ShowFileList(this);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSharedFileList::SendListToServer()
{
#ifdef OLD_SOCKETS_ENABLED
	CSingleLock		sLock(&m_mutexList,true);

	CServer*		pCurServer = m_pServerConnect->GetCurrentServer();

	if (m_mapSharedFiles.IsEmpty() || !pCurServer)
		return;
//
//	Construct the byte stream for the OFFERFILES packet.
//
	CMemFile		packetStream;
	CCKey			bufKey;
	CKnownFile	   *pKnownFile;
	uint32			dwNumFiles;

//	If the connected server has a soft file limit and we have fewer files than the limit shared...
	if ( pCurServer->GetSoftMaxFiles() != 0
		&& static_cast<uint32>(m_mapSharedFiles.GetCount()) <= pCurServer->GetSoftMaxFiles() ) 
	{
		dwNumFiles = m_mapSharedFiles.GetCount();
		packetStream.Write(&dwNumFiles,sizeof(DWORD));			// <count:DWORD> Shared file count

		for (POSITION pos = m_mapSharedFiles.GetStartPosition(); pos != 0; )
		{
			m_mapSharedFiles.GetNextAssoc(pos,bufKey,pKnownFile);
			WriteToOfferedFilePacket(*pKnownFile,packetStream);	// (offered file info)[count]
			pKnownFile->SetOfferedToServer(true);
		}
	}
	else
	{
	//	Send a subset
		POSITION								pos, pos2;
		CTypedPtrList<CPtrList, CKnownFile*>	sortedList;
		bool									bAdded = false;

	//	For each shared file...
		for (pos = m_mapSharedFiles.GetStartPosition(); pos != 0;)
		{
			m_mapSharedFiles.GetNextAssoc(pos, bufKey, pKnownFile);

			bAdded = false;
		//
		//	Perform an insertion sort into 'sortedList'.
			for (pos2 = sortedList.GetHeadPosition(); pos2 != 0 && !bAdded; sortedList.GetNext(pos2))
			{
				uint8		nPrioPos2 = GetRealPrio(sortedList.GetAt(pos2)->GetPriority());
				uint8		nPrioFile = GetRealPrio(pKnownFile->GetPriority());

			//	If the file we're inserting has a higher priority than the one from the sorted list
			//		 or it's the same priority but it's file name comes before the other...
				if ( (nPrioPos2 < nPrioFile)
				  || ( (nPrioPos2 == nPrioFile)
				    && (sortedList.GetAt(pos2)->GetFileName().CompareNoCase(pKnownFile->GetFileName()) > 0) ) )
				{
				//	Insert it before the current one in the sorted list.
					sortedList.InsertBefore(pos2,pKnownFile);
					bAdded = true;
				}
			}
		//	If we didn't find a file to add this one before, add it to the end of the list.
			if (!bAdded)
				sortedList.AddTail(pKnownFile);
		}
	//
	//	Start writing the byte stream
		dwNumFiles = pCurServer->GetSoftMaxFiles();
	//	If we don't know the soft limit, we send all our files
		if (dwNumFiles == 0)
			dwNumFiles = sortedList.GetCount();
		packetStream.Write(&dwNumFiles,sizeof(DWORD));				// <count:DWORD> Shared file count

		uint32		dwCount = 0;

		pos = sortedList.GetHeadPosition();
		while (pos != 0 && dwCount < dwNumFiles)
		{
			dwCount++;
			sortedList.GetAt(pos)->SetOfferedToServer(true);
			WriteToOfferedFilePacket(*sortedList.GetNext(pos),packetStream);	// (offered file info)[count]
		}
	//
	//	Clear the "offered to server" flag on all the files we couldn't send
		while (pos != 0 && dwCount<static_cast<uint32>(sortedList.GetCount()))
		{
			dwCount++;
			sortedList.GetNext(pos)->SetOfferedToServer(false);
		}
		sortedList.RemoveAll();
	}
	sLock.Unlock();
	UpdateOfferedItems();

	Packet		*pPacket = new Packet(&packetStream);

	pPacket->m_eOpcode = OP_OFFERFILES;	

	// compress packet
	//   - this kind of data is highly compressable (N * (1 MD4 and at least 3 string meta data tags and 1 integer meta data tag))
	//   - the min. amount of data needed for one published file is ~100 bytes
	//   - this function is called once when connecting to a server and when a file becomes shareable - so, it's called rarely.
	//   - if the compressed size is still >= the original size, we send the uncompressed packet
	// therefor we always try to compress the packet
	if (pCurServer->GetTCPFlags() & SRV_TCPFLG_COMPRESSION)
	{
		UINT uUncomprSize = pPacket->m_dwSize;
		pPacket->PackPacket();
	}
	
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadServer(pPacket->m_dwSize);

	m_pServerConnect->SendPacket(pPacket,true);
#endif //OLD_SOCKETS_ENABLED
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	UpdateOfferedItems() causes the shared file list window to be updated.
void CSharedFileList::UpdateOfferedItems()
{
	if (m_pOutput != NULL)
	{
		m_pOutput->PostMessage(WM_SFL_UPDATEOFFEREDITEMS,0,0);
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSharedFileList::WriteToOfferedFilePacket(CKnownFile &sharedFile,CMemFile &packetStream)
{
#ifdef OLD_SOCKETS_ENABLED
	packetStream.Write(sharedFile.GetFileHash(),16);		// <HASH> file hash

	// if connected and highID, set client ID and port
	uint32 dwClientID = 0;
	uint16 uClientPort = 0;
	if(m_pServerConnect->IsConnected() && !m_pServerConnect->IsLowID())
	{
		dwClientID = m_pServerConnect->GetClientID();
		uClientPort = g_eMuleApp.m_pGlobPrefs->GetPort();
	}
	packetStream.Write(&dwClientID, 4);
	packetStream.Write(&uClientPort, 2);

	//write a number of Tags
	uint32		dwTagCount = 3;
	packetStream.Write(&dwTagCount,4);	// <Tag_set> {<TAGCOUNT:DWORD> # tags determined by file type

	CFileTag	nameTag(FT_FILENAME,sharedFile.GetFileName());
	nameTag.WriteToFile(packetStream);					// (FT_FILENAME:string) file name

	CFileTag	sizeTag(FT_FILESIZE,sharedFile.GetFileSize());
	sizeTag.WriteToFile(packetStream);					// (FT_FILESIZE:int) file size (bytes) }

	CFileTag	typeTag(FT_FILETYPE, sharedFile.GetFileType());
	typeTag.WriteToFile(packetStream);
#endif //OLD_SOCKETS_ENABLED
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetDatasize() returns the size of the largest shared file in 'qwLargest' and returns
//		the total size of all shared files as its result.
uint64 CSharedFileList::GetDatasize(uint64 &qwLargest)
{
	qwLargest = 0;

	uint64			qwFilesTotalBytes;

	qwFilesTotalBytes = 0;

	CCKey			bufKey;
	CKnownFile	   *pKnownFile;
	CSingleLock		sLock(&m_mutexList,true);

	for (POSITION pos = m_mapSharedFiles.GetStartPosition(); pos != NULL;)
	{
		m_mapSharedFiles.GetNextAssoc(pos,bufKey,pKnownFile);
		qwFilesTotalBytes += pKnownFile->GetFileSize();
	//	If this file is bigger than all the others...well duh.
		if (pKnownFile->GetFileSize() > qwLargest)
			qwLargest = pKnownFile->GetFileSize();
	}

	return qwFilesTotalBytes;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	GetFileByID() returns the Known File with hash 'strFileHash' or NULL if there is none.
CKnownFile *CSharedFileList::GetFileByID(const uchar *strFileHash)
{
	CKnownFile		*pResultFile;
	CCKey			tKey(strFileHash);
	CSingleLock		sLock(&m_mutexList,true);

	if (m_mapSharedFiles.Lookup(tKey,pResultFile))
		return pResultFile;
	else
		return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
short CSharedFileList::GetFilePriorityByID(uchar *pFileHash)
{
	CKnownFile		*pKnownFile = GetFileByID(pFileHash);

	if (pKnownFile != NULL)
		return pKnownFile->GetPriority();
	else
		return -10;	// file doesn't exist
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSharedFileList::HashNextFile()
{
//	If we're waiting to reload the shared files list (after clearing hash list)...
	if(m_bDelayReload == true)
	{
	//	Do reload now
		m_bDelayReload = false;
		Reload();
		return;
	}
//	If there're no files to hash, just return
	if (m_waitingForHashList.IsEmpty())
		return;

	UnknownFile_Struct	   *pUnknownFile = m_waitingForHashList.RemoveHead();

	g_eMuleApp.m_fileHashControl.AddToHash(this, pUnknownFile->m_strDirectory, pUnknownFile->m_strFileName); 	
	delete pUnknownFile;	
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSharedFileList::UpdateItem(CKnownFile *pKnownFile,bool bResort/*=true*/)
{
	if (m_pOutput != NULL)
	{
		m_pOutput->PostMessage( WM_SFL_UPDATEITEM,
							    static_cast<uint32>(bResort),
							    reinterpret_cast<LPARAM>(pKnownFile) );
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define INCOMPLETE_DIR		_T("!Incomplete Files")
#define SHARED_DIR			_T("!Shared Files")

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CSharedFileList::AddSharedVDirForList(bool bIncomplete, const CString &strNewDir)
{
	EMULE_TRY
	// clean dir
	CString strDir(strNewDir);
	PathRemoveBackslash(strDir.GetBuffer());
	if (strDir.IsEmpty())
		return false;

	// check already added
	POSITION pos= m_SharedVDirForList.GetStartPosition();
	while (pos)
	{
		CString strCurKey;
		CString strCurDir;
		m_SharedVDirForList.GetNextAssoc(pos, strCurKey, strCurDir);
		if (strCurDir.CompareNoCase(strDir) == 0)
			return false;
	}

	// set key
	CString strKey, strWantedKey;
	if (bIncomplete)
	{
		strWantedKey= INCOMPLETE_DIR;	// default name used by Hybrid for temp dir
	}
	else
	{
		strWantedKey= PathFindFileName((LPCTSTR)strDir);	// short name without complete path
		if ((strWantedKey.IsEmpty()) || ((strWantedKey.GetLength() > 1) && (strWantedKey[1] == ':')))
			strWantedKey= SHARED_DIR;
	}

	CString strDir0, strDir1;
	if ((m_SharedVDirForList.Lookup(strWantedKey, strDir0)) || (m_SharedVDirForList.Lookup(strWantedKey + _T(" (1)"), strDir1)))
	{
		// rename key 0
		if (!strDir0.IsEmpty())
		{
			m_SharedVDirForList[strWantedKey + _T(" (1)")]= strDir0;
			m_SharedVDirForList.RemoveKey(strWantedKey);
		}

		// build unique key
		int n= 2;
		do
		{
			strKey.Format(_T("%s (%d)"), (LPCTSTR)strWantedKey, n++);
		} while (m_SharedVDirForList.Lookup(strKey, strDir0));
	}
	else
		strKey= strWantedKey;
	m_SharedVDirForList[strKey]= strDir;

	return true;

	EMULE_CATCH
	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CSharedFileList::BuildSharedVDirForList()
{
	if (!m_pPrefs)
		return;

	POSITION pos;
	m_SharedVDirForList.RemoveAll();

	// temp dirs
	AddSharedVDirForList(true, m_pPrefs->GetTempDir());
	pos= m_pPrefs->m_tempDirList.GetHeadPosition();
	while (pos)
		AddSharedVDirForList(true, m_pPrefs->m_tempDirList.GetNext(pos));

	// incoming dir
	AddSharedVDirForList(false, m_pPrefs->GetIncomingDir());

	// shared dirs
    pos = m_pPrefs->m_sharedDirList.GetHeadPosition();
    while (pos)
		AddSharedVDirForList(false, m_pPrefs->m_sharedDirList.GetNext(pos));
	
	// category dirs
	for (int i= 0; i < CCat::GetNumCats(); i++)
		AddSharedVDirForList(false, CCat::GetCatPathByIndex(i));
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

IMPLEMENT_DYNCREATE(CHashFileThread, CWinThread)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CHashFileThread::CHashFileThread()
{
	m_pOwner = NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	SetValues() sets the thread parameters
void CHashFileThread::SetValues( CSharedFileList *pOwner, LPCTSTR in_directory, LPCTSTR in_filename,
							    CPartFile *in_pPartFile)
{
	 m_pOwner = pOwner;
	 m_strDirectory = in_directory;
	 m_strFileName = in_filename;
	 m_pPartFile = in_pPartFile;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int CHashFileThread::Run()
{
//	If there is neither a SharedFileList owner or a PartFile owner or if there's no filename...
	if (!(m_pOwner || m_pPartFile) || m_strFileName.IsEmpty())
		AfxEndThread(0,true);

	CSingleLock sLock1(&(g_eMuleApp.m_mutexHashing)); // only one filehash at a time

	sLock1.Lock();

	CKnownFile		*pNewKnownFile = new CKnownFile();

//	Try to create a hashed KnownFile for the spec'd file. If successful...
	if (pNewKnownFile->CreateFromFile(m_strDirectory,m_strFileName))
	{
	//	Notify the PartFile (via a message to the eMule window) that the file it was
	//		completing has finished hashing.
		if (g_eMuleApp.m_pdlgEmule)
			PostMessage( g_eMuleApp.m_pdlgEmule->m_hWnd,
						 TM_FINISHEDHASHING,
						 (m_pOwner ? 0 : (WPARAM)m_pPartFile),
						 (LPARAM)pNewKnownFile);
	}
	else
	{
		delete pNewKnownFile;
	}

	sLock1.Unlock();
	AfxEndThread(0,true);
	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
