//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#pragma once
#include "types.h"
#include "opcodes.h"
#include "updownclient.h"
#include "preferences.h"
#include "WebServer.h" // kuchin
#include "Loggable.h"

class CUploadQueue : public CLoggable{
	friend class CWebServer; // kuchin
public:
	CUploadQueue(CPreferences* in_prefs);
	~CUploadQueue();
	void	Process();
	void	AddClientToQueue(CUpDownClient* client,bool bIgnoreTimelimit = false);
	bool	RemoveFromUploadQueue(CUpDownClient* client, EnumEndTransferSession eReason, bool updatewindow = true);
	bool	RemoveFromWaitingQueue(CUpDownClient* client,bool updatewindow = true);
	bool	IsOnUploadQueue(CUpDownClient* client)	{return GetWaitingClient(client);}
	//bool	IsDownloading(CUpDownClient* client)	{return GetDownloadingClient(client);}
	// eklmn: optimization: search client with extra function?! Why?!
	bool	IsDownloading(CUpDownClient* client)	{return uploadinglist.Find(client);}
	bool IsFileUploading (CPartFile* check_file);		//eklmn: Complete Bug
	uint32	GetDataRateShown();		//Cax2 - more immediate data rate feedback.
	uint32	GetDataRate();
	bool    CheckForTimeOver(CUpDownClient* client);
	int		GetWaitingUserCount()					{return waitinglist.GetCount();}
	int		GetUploadQueueLength()					{return uploadinglist.GetCount();}
	int		GetActiveUploadsCount()					{return m_MaxActiveClients;}
	void	UpdateBanCount();
	CUpDownClient*	GetWaitingClientByIP_UDP(uint32 dwIP, uint16 dwPort);
	CUpDownClient*	GetNextClient(CUpDownClient* update);

	HANDLE m_hSaveSTimer;
	BOOL m_bExiting;

	void	SCHShift1UploadCheck();
	void	SCHShift2UploadCheck();
	
	void	DeleteAll();
	uint16	GetWaitingPosition(CUpDownClient* client);
	void	SetBanCount(uint32 in_count)			{m_iBannedCount = in_count;}
	uint32	GetBanCount()							{return m_iBannedCount;}
	// v: eklmn: extended UL statistic
	uint32	GetSuccessfulUpCount()					{return m_iULSessionSuccessful;}
	uint32	GetULFullChunkCount()					{return m_iULSessionSuccessfulFullChunk;}	// Should that be full up-chuck count?
	uint32	GetULPartChunkCount()					{uint32 part_chunk = 0; 
														  for (int i =0;i<ETS_TERMINATOR;i++) 
															part_chunk += m_iULSessionSuccessfulPartChunk[i];
														  return part_chunk;
														}
	uint32	GetULPartChunkSubCount(EnumEndTransferSession index)		{return m_iULSessionSuccessfulPartChunk[index];}
	uint32	GetFailedUpCount()						{uint32 failed = 0; 
														  for (int i =0;i<ETS_TERMINATOR;i++)
															failed += m_iULSessionFailed[i];
														  return failed;
														}
	uint32	GetFailedSubCount(EnumEndTransferSession index)		{return m_iULSessionFailed[index];}
	// ^: eklmn: extended UL statistic
	uint32	GetAverageUpTime();
	void	FindSourcesForFileById(CTypedPtrList<CPtrList, CUpDownClient*>* srclist, const uchar* filehash);
	void	AddUpDataOverheadSourceExchange(uint32 data, uint32 dummy=0)	{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadSourceExchange += data;
															  m_nUpDataOverheadSourceExchangePackets++;}
	void	AddUpDataOverheadFileRequest(uint32 data, uint32 dummy=0)		{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadFileRequest += data;
															  m_nUpDataOverheadFileRequestPackets++;}
	void	AddUpDataOverheadServer(uint32 data, uint32 dummy=0)			{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadServer += data;
															  m_nUpDataOverheadServerPackets++;}
	void	AddUpDataOverheadOther(uint32 data, uint32 dummy=0)				{ m_nUpDataRateMSOverhead += data;
															  m_nUpDataOverheadOther += data;
															  m_nUpDataOverheadOtherPackets++;}
	uint32	GetUpDataRateOverhead()						{return m_nUpDataRateOverhead;}
	uint64	GetUpDataOverheadSourceExchange()			{return m_nUpDataOverheadSourceExchange;}
	uint64	GetUpDataOverheadFileRequest()				{return m_nUpDataOverheadFileRequest;}
	uint64	GetUpDataOverheadServer()					{return m_nUpDataOverheadServer;}
	uint64	GetUpDataOverheadOther()					{return m_nUpDataOverheadOther;}
	uint64	GetUpDataOverheadSourceExchangePackets()	{return m_nUpDataOverheadSourceExchangePackets;}
	uint64	GetUpDataOverheadFileRequestPackets()		{return m_nUpDataOverheadFileRequestPackets;}
	uint64	GetUpDataOverheadServerPackets()			{return m_nUpDataOverheadServerPackets;}
	uint64	GetUpDataOverheadOtherPackets()				{return m_nUpDataOverheadOtherPackets;}
	void	CompUpDataRateOverhead();
protected:
	void	RemoveFromWaitingQueue(POSITION pos, bool updatewindow);
	POSITION	GetWaitingClient(CUpDownClient* client);
	POSITION	GetWaitingClientByID(CUpDownClient* client);
	//POSITION	GetDownloadingClient(CUpDownClient* client);
	bool		CanAcceptNewClient(uint32 numberOfUploads);
	void		AddUpNextClient(CUpDownClient* directadd = NULL);
private:
	std::list<uint64> m_averageDataRateList;		//Cax2 - memory optimization
	std::list<DWORD> m_averageTickList;	//Cax2 - memory optimization
	CList<int,int> activeClients_list;
	CTypedPtrList<CPtrList, CUpDownClient*> waitinglist;
	CTypedPtrList<CPtrList, CUpDownClient*> uploadinglist;

	uint32	m_iDataRate;   //data rate 
	uint32	m_iDataRateAvg; //datarage average (since progstart) *unused*
	sint32	m_iEstMaxDataRate; // estimated max data rate	

//	DWORD	lasttimechecked; // last time that we checked to remove a client from upload queue
	CPreferences* m_pAppPrefs;
	UINT_PTR m_hTimer;
	uint32	m_iBannedCount;
	uint32	m_iULSessionSuccessful;
	uint32	m_iULSessionSuccessfulFullChunk;
	uint32	m_iULSessionSuccessfulPartChunk[ETS_TERMINATOR];	
	uint32	m_iULSessionFailed[ETS_TERMINATOR];

	uint32	m_iTotalUploadTime;
    sint32  m_iLeftOverBandwidth;
    uint32  m_MaxActiveClients;
	uint32	m_nUpDataRateOverhead;
	uint32	m_nUpDataRateMSOverhead;
	uint64	m_nUpDataOverheadSourceExchange;
	uint64	m_nUpDataOverheadFileRequest;
	uint64	m_nUpDataOverheadServer;
	uint64	m_nUpDataOverheadOther;
	uint64	m_nUpDataOverheadSourceExchangePackets;
	uint64	m_nUpDataOverheadFileRequestPackets;
	uint64	m_nUpDataOverheadServerPackets;
	uint64	m_nUpDataOverheadOtherPackets;
	CList<int,int>	m_AvarageUDRO_list;

    uint32  m_guessedMaxBandwidth;
	uint32  m_guessedMaxLANBandwidth; // LANCAST (moosetea) the max LAN data rate

    DWORD   m_lastGaveDataTick;
    DWORD   m_lastCalculatedDataRateTick;
//	uint32	sumavgdata; // enkeyDEV(Ottavio84) -Accurate speed measurement-
//	uint32	sendtimestamp;	// enkeyDEV(Ottavio84) -Upload rate is made independent from the inaccurate WM_TIMER-
	CWinThread *m_pSaveThread;
};
